# Copyright (c) 2011-2012  The Board of Trustees of The Leland Stanford Junior University
# Copyright (c) 2013  Barnstormer Softworks, Ltd.

import traceback
import hashlib
from types import NoneType

from flask import request

from foam.flowvisor import Connection as FV
from foam.core.json import jsonify, jsonValidate, JSONValidationError
from foam.api.jsonrpc import Dispatcher, route
from foam import types
import foam.task
import foam.lib
import foam.geni.lib
from foam.geni.db import GeniDB
from foam.core.configdb import ConfigDB

class AdminAPIv1(Dispatcher):
  def __init__ (self, app):
    super(AdminAPIv1, self).__init__("Admin v1", app.logger, app)
    self._log.info("Loaded")

  def validate (self, rjson, types):
    return jsonValidate(rjson, types, self._log)

  @route('/core/admin/get-fv-slice-name', methods=["POST"])
  def getFVSliceName (self):
    if not request.json:
      return
    try:
      self.validate(request.json, [("slice_urn", (unicode,str))])
      name = GeniDB.getFlowvisorSliceName(request.json["slice_urn"])
      return jsonify({"name" : name})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/list-slivers', methods=["POST", "GET"])
  def listSlivers (self):
    try:
      deleted = False
      status = 0

      if request.method == "POST":
        if not request.json:
          return ""
        if request.json.has_key("deleted"):
          deleted = request.json["deleted"]
        if request.json.has_key("status"):
          st = request.json["status"].lower()
          if st == "approved":
            status = True
          elif st == "rejected":
            status = False
          elif st == "pending":
            status = None

      slivers = GeniDB.getSliverList(deleted, status)
      return jsonify({"slivers" : slivers})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg = traceback.format_exc())

  @route('/core/admin/approve-sliver', methods=["POST"])
  def approveSliver (self):
    if not request.json:
      return
    return foam.geni.lib.approveSliver(request, self._log)

  @route('/core/admin/reject-sliver', methods=["POST"])
  def rejectSliver (self):
    if not request.json:
      return
    try:
      self.validate(request.json, [("sliver_urn", (unicode,str))])
      slice_name = GeniDB.getFlowvisorSliceName(sliver_urn=request.json["sliver_urn"])
      sobj = GeniDB.getSliverObj(request.json["sliver_urn"])

      data = GeniDB.getSliverData(sobj.getURN(), True)

      GeniDB.setSliverStatus(request.json["sliver_urn"], False)
      if FV.sliceExists(slice_name):
        FV.deleteSlice(slice_name)

      foam.task.emailRejectSliver(data)

      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/disable-sliver', methods=["POST"])
  def disableSliver (self):
    if not request.json:
      return
    try:
      self.validate(request.json, [("sliver_urn", (unicode,str))])
      slice_name = GeniDB.getFlowvisorSliceName(sliver_urn=request.json["sliver_urn"])
      sobj = GeniDB.getSliverObj(request.json["sliver_urn"])

      data = GeniDB.getSliverData(sobj.getURN(), True)

      GeniDB.setSliverStatus(request.json["sliver_urn"], None)
      GeniDB.commit()

      if FV.sliceExists(slice_name):
        FV.deleteSlice(slice_name)

      foam.task.emailDisableSliver(data)

      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())


  @route('/core/admin/get-config', methods=["POST"])
  def getConfig (self):
    if not request.json:
      return

    try:
      objs = self.validate(request.json, [("key", (unicode,str))])
      u = ConfigDB.getUser(request.environ["USER"])
      # Don't look here - stupidity to get around the fact that we don't
      # have output processors
      if objs["key"] == "geni.max-lease":
        val = ConfigDB.getConfigItemByKey("geni.max-lease").getValue(u)
        return jsonify({"value" : str(val)})
      else:
        return jsonify({"value" : ConfigDB.getConfigItemByKey(request.json["key"]).getValue(u)})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except TypeError, e:
      return jsonify(None, 3, msg = "Unknown key (%s)" % (objs["key"]))
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg = traceback.format_exc())

  @route('/core/admin/get-config-keys', methods=["GET"])
  def getConfigKeys (self):
    try:
      citems = ConfigDB.getConfigItems()
      return jsonify(citems)
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg = traceback.format_exc())

  @route('/core/admin/set-config', methods=["POST"])
  def setConfig (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("key", (unicode,str)), ("value", (dict, int, unicode, str, NoneType))])
      u = ConfigDB.getUser(request.environ["USER"])
      key = request.json["key"]
      ConfigDB.getConfigItemByKey(request.json["key"]).write(request.json["value"], u)
      return jsonify({"status" : "success"})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())


  @route('/core/admin/get-sliver-flowspace', methods=["POST"])
  def getSliverFlowspace (self):
    if not request.json:
      return
    try:
      self.validate(request.json, [("sliver_urn", (unicode,str))])
      sobj = GeniDB.getSliverObj(request.json["sliver_urn"])
      return jsonify({"flowspace" : sobj.generateFlowEntries()})
    except JSONValidationError, e:
      return jsonify(e.__json__())
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/get-version', methods=["GET"])
  def jsonGetVersion (self):
    try:
      import foam.version
      return jsonify({"version" : foam.version.VERSION})
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/set-location', methods=["POST"])
  def setLocation (self):
    if not request.json:
      return
    try:
      obj = self.validate(request.json, [("lat", float), ("long", float), ("country", (unicode,str))])
      ConfigDB.getConfigItemByKey("geni.location-data").write(obj)
      return jsonify({"status" : "success"})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/add-datapath', methods=["POST"])
  def addDatapath (self):
    if not request.json:
      return

    try:
      objs = self.validate(request.json, [("dpid", (foam.geni.lib.jsonDPID))])
      GeniDB.addDatapath(str(objs["dpid"]))
      return jsonify({"status" : "success"})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/remove-datapath', methods=["POST"])
  def removeDatapath (self):
    if not request.json:
      return

    try:
      objs = self.validate(request.json, [("dpid", (foam.geni.lib.jsonDPID))])
      GeniDB.removeDatapath(str(objs["dpid"]))
      return jsonify({"status" : "success"})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/list-datapaths', methods=["GET"])
  def listDatapaths (self):
    try:
      devices = FV.getDeviceList()
      db_devices = GeniDB.getDeviceSet()
      GeniDB.refreshDevices(devices)

      for dpid in devices:
        db_devices.add(dpid)

      return jsonify(list(db_devices))
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())
    

  @route('/core/admin/set-trigger', methods=["POST"])
  def setTrigger (self):
    if not request.json:
      return
    try:
      obj = self.validate(request.json, [("type", types.TriggerType), ("start", (str,unicode)),
                              ("end", (str,unicode)), ("event", types.EventType),
                              ("action", None)])
      GeniDB.addTrigger(obj["type"], obj["start"], obj["end"], obj["event"], obj["action"])
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/set-sliver-expiration', methods=["POST"])
  def adminSetSliverExpiration (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("datetime", types.DateTime), ("urn", types.SliverURN)])
      GeniDB.updateSliverExpiration(objs["urn"], objs["datetime"])
      return jsonify({"status" : "success"})
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/core/admin/rebuild-flowvisor-cache', methods=["GET"])
  def adminRebuildFlowvisorCache (self):
    try:
      FV.rebuildCache()
      return jsonify(None)
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())


def setup (app):
  api = AdminAPIv1(app)
  return api

