# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import logging
import zlib
import base64
import xmlrpclib
from xml.parsers.expat import ExpatError

from flaskext.xmlrpc import XMLRPCHandler
from flask import request

import foam.task
import foam.lib
import foam.api.xmlrpc
import foam.version
import foam.geni.approval
from foam.creds import CredVerifier, Certificate
from foam.core.configdb import ConfigDB
from foam.core.log import KeyAdapter
from foam.geni.db import GeniDB, UnknownSlice, UnknownNode
from foam.core.exception import CoreException
from foam.geni import version as gapiver
from foam.geni.resources import UnknownSliceError

import foam.geni.lib
from foam.geni.codes import GENI_ERROR_CODE

from .awsevent import Publisher as AWSEvent

import sfa

class AMAPIv2(foam.api.xmlrpc.Dispatcher):
  """Implementation of GENI AM API v2"""
  def __init__ (self, log):
    super(AMAPIv2, self).__init__("GAPIv2", log)
    self._actionLog = KeyAdapter("v2", logging.getLogger('gapi-actions'))
    self._auto_priority = ConfigDB.getConfigItemByKey("geni.auto-sliver-priority").getValue()

  def recordAction (self, action, credentials = [], urn = None):
    cred_ids = []

    self._actionLog.info("Sliver: %s  Action: %s" % (urn, action))

    for cred in credentials:
      self._actionLog.info("Credential: %s" % (cred))

  def buildPropertyList(self, geni_code, value=None, output=None):
    """Return XML-RPC struct (aka property list)"""
    result = {}
    result["code"] = {'geni_code': geni_code}
    # Non-zero geni_code implies error: output is required, value is optional
    if geni_code:
      result["output"] = output
      if value:
        result["value"] = value
    # Zero geni_code implies success: value is required, output is optional
    else:
      result["value"] = value
    return result

  def pub_GetVersion (self, options=None):
    """Return the version of the GENI AM API and RSpecs supported

    Get static version and configuration information about this aggregate.

    """
    self.recordAction("getversion")
    d = gapiver.Registry.getGAPIData(2, request.environ['SERVER_ADDR'], request.environ['SERVER_PORT'])

    result = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=d)

    # geni_api is duplicated under value to allow v1 clients to determine the AM API version
    result["geni_api"] = 2

    return result

  def pub_ListResources (self, credentials, options):
    """Return information about available resources or resources allocated to a slice

    List the resources at this aggregate in an RSpec: may be all resources,
    only those available for reservation, or only those already reserved for the given slice.

    """
    try:
      CredVerifier.checkValid(credentials, [])

      # Parse options
      compressed = options.get("geni_compressed", False)
      urn = options.get("geni_slice_urn", None)
      spec_version = options.get("geni_rspec_version")
      supported_spec = {'version': '3', 'type': 'GENI'}
      return_request = options.get("foam_get_request", False)
      if spec_version:
        if spec_version != supported_spec:
          msg = "RSpec type/version not supported"
          propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADVERSION, output=msg)
          return propertyList
      else:
        msg = "Required option geni_rspec_version missing"
        propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADARGS, output=msg)
        return propertyList

      cert = Certificate(request.environ['CLIENT_RAW_CERT'])
      if urn:
        CredVerifier.checkValid(credentials, "getsliceresources", urn)
        self.recordAction("listresources", credentials, urn)
        AWSEvent.publishEvent("listresources", options, target_urn = urn, user_urn = cert.getURN(),
                              req_ip = request.environ["REMOTE_ADDR"])
# TODO: Need to fail properly if there's no slice here for any resource manager
        if return_request:
          # TODO: slice_urn now
          rspec = GeniDB.getRspec(sliver_urn)
        else:
          rspec = foam.geni.lib.getManifest(urn)
      else:
        self.recordAction("listresources", credentials)
        AWSEvent.publishEvent("gapi2.listresources", options, target_urn = None, user_urn = cert.getURN(),
                              req_ip = request.environ["REMOTE_ADDR"])
        rspec = foam.geni.lib.getAdvertisement()
      if compressed:
        zrspec = zlib.compress(rspec)
        rspec = base64.b64encode(zrspec)

      propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=rspec)

    except ExpatError:
      msg = "Error parsing credential strings"
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADARGS, output=msg)
      self._log.error(msg)
    except UnknownSliceError as x:
      # Raised by GeniDB.getSliverURN()
      msg = "Attempt to list resources on sliver for unknown slice %s" % (urn)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      x.log(self._log, msg, logging.INFO)
    except xmlrpclib.Fault as x:
      # Something thrown via GCF, we'll presume it was something related to credentials
      msg = "GCF credential check failure: <%s>" % (x)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.info(msg)
      self._log.debug(x, exc_info=True)
    except AttributeError as x:
      # New GCF problem with user creds that have no gid_caller, probably
      msg = "GCF credential check failure: <%s>" % (x)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.info(msg)
      self._log.debug(x, exc_info=True)
    except CoreException as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <ListResources> failed!")
    except Exception as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception(msg)
    finally:
      return propertyList


  def pub_CreateSliver (self, slice_urn, credentials, rspec, users, options):
    """Allocate resources to a slice

    Reserve the resources described in the given RSpec for the given slice, returning a manifest RSpec of what has been reserved.

    """
    
    self._log.info(str(users))
    user_info = {"_api" : 2}
    try:
      if CredVerifier.checkValid(credentials, "createsliver"):
        self.recordAction("createsliver", credentials, slice_urn)
        try:
          cert = Certificate(request.environ['CLIENT_RAW_CERT'])
          AWSEvent.publishEvent("gapi2.createsliver", options, target_urn = slice_urn, user_urn = cert.getURN(),
                                req_ip = request.environ["REMOTE_ADDR"])
          user_info["urn"] = cert.getURN()
          user_info["email"] = cert.getEmailAddress()
          self._log.debug("Parsed user cert with URN (%(urn)s) and email (%(email)s)" % user_info)
        except Exception as e:
          self._log.exception("UNFILTERED EXCEPTION")
          user_info["urn"] = None
          user_info["email"] = None

        user_info["login_users"] = users
        manifest = foam.geni.lib.createSliver(slice_urn, credentials, rspec, user_info)

        propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=manifest)

    except foam.geni.lib.RspecParseError as e:
      msg = str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADARGS, output=msg)
      e.log(self._log, msg, logging.INFO)
    except foam.geni.lib.RspecValidationError as e:
      msg = str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADARGS, output=msg)
      e.log(self._log, msg, logging.INFO)
    except foam.geni.lib.DuplicateSliver as ds:
      msg = "Attempt to create multiple slivers for slice [%s]" % (ds.slice_urn)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ALREADYEXISTS, output=msg)
      ds.log(self._log, msg, logging.INFO)
    except foam.geni.lib.UnknownComponentManagerID as ucm:
      msg = "Component Manager ID specified in %s does not match this aggregate." % (ucm.cid)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      ucm.log(self._log, msg, logging.INFO)
    except (foam.geni.lib.UnmanagedComponent, UnknownNode) as uc:
      msg = "DPID in component %s is unknown to this aggregate." % (uc.cid)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      uc.log(self._log, msg, logging.INFO)
    except CoreException as e:
      if e._unexpected:
        self._log.exception("Call to known method <CreateSliver> failed!")
        AWSEvent.publishError(e, self._log)
      else:
        self._log.error(str(e))
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    except Exception as e:
      AWSEvent.publishError(e)
      msg = "Exception"
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception(msg)
    finally:
      return propertyList

  def pub_DeleteSliver (self, slice_urn, credentials, options):
    """Delete a sliver

    Stop all the slice's resources and remove the reservation.
    Returns True or False indicating whether it did this successfully.

    """
    from ..geni.resources import Registry

    # TODO: This needs to conform to the return required by the API

    user_info = {}
    try:
      if CredVerifier.checkValid(credentials, "deletesliver", slice_urn):
        self.recordAction("deletesliver", credentials, slice_urn)
        cert = Certificate(request.environ['CLIENT_RAW_CERT'])
        AWSEvent.publishEvent("gapi2.deletesliver", options, target_urn = slice_urn, user_urn = cert.getURN(),
                              req_ip = request.environ["REMOTE_ADDR"])
        user_info["urn"] = cert.getURN()
        user_info["email"] = cert.getEmailAddress()
        Registry.deleteSliversForSlice(slice_urn, user_info)

        propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=True)

    except UnknownSliceError as e:
      msg = "Attempt to delete unknown sliver for slice URN %s" % (slice_urn)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.SEARCHFAILED, output=msg)
      e.log(self._log, msg, logging.INFO)
    except CoreException as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <DeleteSliver> failed!")
    except Exception as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception("Exception")
    finally:
      return propertyList

  def pub_SliverStatus (self, slice_urn, credentials, options):
    """Returns the status of the reservation for this slice at this aggregate"""
    from ..geni.resources import Registry

    try:
      if CredVerifier.checkValid(credentials, "sliverstatus", slice_urn):
        self.recordAction("sliverstatus", credentials, slice_urn)
        cert = Certificate(request.environ['CLIENT_RAW_CERT'])
        AWSEvent.publishEvent("gapi2.sliverstatus", options, target_urn = slice_urn, user_urn = cert.getURN(),
                              req_ip = request.environ["REMOTE_ADDR"])
        slivers = Registry.getSlivers(slice_urn)
        if not slivers:
          raise UnknownSlice(slice_urn)
        result = Registry.sliverStatus(slivers, options)
        result["geni_urn"] = slice_urn
        result["geni_status"] = "ready"
        result["geni_resources"] = []
        for sliver in slivers:
          result["geni_resources"].append({"geni_urn" : sliver, "geni_status" : "ready", "geni_error" : ""})
        propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=result)
    except UnknownSlice as e:
      msg = "Attempt to get status on unknown sliver for slice %s" % (slice_urn)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.SEARCHFAILED, output=msg)
      e.log(self._log, msg, logging.INFO)
    except CoreException as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <SliverStatus> failed!")
    except Exception as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception(msg)
    finally:
      return propertyList

  def pub_RenewSliver (self, slice_urn, credentials, exptime, options):
    """Renew the reservation for resources in this slice"""
    from ..geni.resources import Registry

    try:
      if CredVerifier.checkValid(credentials, "renewsliver", slice_urn):
        self.recordAction("renewsliver", credentials, slice_urn)
        cert = Certificate(request.environ['CLIENT_RAW_CERT'])
        AWSEvent.publishEvent("gapi2.renewsliver", {"exptime" : exptime, "options" : options},
                              target_urn = slice_urn, user_urn = cert.getURN(),
                              req_ip = request.environ["REMOTE_ADDR"])

        creds = CredVerifier.fromStrings(credentials, "renewsliver", slice_urn)
        Registry.renewSliversForSlice(slice_urn, exptime, creds)
#        foam.task.emailRenewSliver(data)

        propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=True)
    except foam.lib.BadSliverExpiration as e:
      msg = "Bad expiration request: %s" % (e.msg)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, msg, logging.INFO)
    except CoreException as e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <RenewSliver> failed!")
    except Exception:
      AWSEvent.publishError(e)
      msg = "Exception"
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception("Exception")
    finally:
      return propertyList

  def pub_Shutdown (self, slice_urn, credentials, options):
    """Perform an emergency shutdown of the resources in a slice at this aggregate"""
    if CredVerifier.checkValid(credentials, "shutdown", slice_urn):
      self.recordAction("shutdown", credentials, slice_urn)
      foam.lib.shutdown(slice_urn)
      return self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=True)
    return self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=False)

def setup (app):
  endpoint = "/foam/gapi/2"
  foam.geni.version.Registry.addVersion(2, endpoint)
  gapi2 = XMLRPCHandler('gapi2')
  gapi2.connect(app, endpoint)
  gapi2.register_instance(AMAPIv2(app.logger))
  app.logger.info("[GAPIv2] Loaded.")
