# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime
import logging
import json

from flaskext.xmlrpc import XMLRPCHandler, Fault
from flask import request

import foam.api.xmlrpc
from foam.geni import version as gapiver
from foam.core.log import KeyAdapter
from foam.core.exception import CoreException
from foam.creds import CredVerifier, Certificate
from foam.geni.codes import GENI_ERROR_CODE

from .awsevent import Publisher as AWSEvent

class AMAPIv3(foam.api.xmlrpc.Dispatcher):
  def __init__ (self, log):
    super(AMAPIv3, self).__init__("GAPIv3", log)
    self._actionLog = KeyAdapter("v3", logging.getLogger('gapi-actions'))

  def recordAction (self, action, credentials = [], urn = None, data = None):
    cred_ids = []

    self._actionLog.info("Sliver: %s  Action: %s" % (urn, action))
    for cred in credentials:
      self._actionLog.info("Credential: %s" % (cred))

  def buildPropertyList(self, geni_code, value=None, output=None):
    """Return XML-RPC struct (aka property list)"""
    result = {}
    result["code"] = {'geni_code': geni_code}
    # Non-zero geni_code implies error: output is required, value is optional
    if geni_code:
      result["output"] = output
      if value:
        result["value"] = value
    # Zero geni_code implies success: value is required, output is optional
    else:
      result["value"] = value
    return result

  def pub_GetVersion (self, options=None):
    self.recordAction("getversion")
    d = gapiver.Registry.getGAPIData(3, request.environ['SERVER_ADDR'], request.environ['SERVER_PORT'])

    result = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=d)

    # geni_api is duplicated under value to allow v1 clients to determine the AM API version
    result["geni_api"] = 3
    return result

  def pub_PerformOperationalAction (self, urns, credentials, action, options):
    from ..geni.resources import Registry

    self._log.info("[%s] URNs: <%s>, Options: %s" % (action, ",".join(urns), json.dumps(options)))

    try:
      #TODO: Should refactor this into a utility
      creds = []
      for cred in credentials:
        if cred["geni_type"] != "geni_sfa":
          # We don't support ABAC, go away
          return None

        creds.append(cred["geni_value"])

      for urn in urns:
        # At the moment you need delete privileges to perform actions
        CredVerifier.checkValid(creds, "deletesliver", Registry.getSliceURN(urn))

      cert = Certificate(request.environ['CLIENT_RAW_CERT'])
      AWSEvent.publishEvent("gapi3.%s" % (action), options, target_urn = urns[0], user_urn = cert.getURN(),
                            req_ip = request.environ["REMOTE_ADDR"])

      out = Registry.performOperationalAction(urns, action, options, cert.getURN())
      result = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=out)
    except CoreException, e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % (str(e))
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <PerformOperationalAction> failed!")
    except Exception, e:
      AWSEvent.publishError(e)
      self._log.exception("Exception")
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    finally:
      result["geni_api"] = 3
      return result

  def pub_PerformAggregateAction (self, action, options):
    from ..geni.resources import Registry

    self._log.info("[%s] Options: %s" % (action, json.dumps(options)))

    try:
      cert = Certificate(request.environ['CLIENT_RAW_CERT'])

      AWSEvent.publishEvent(action, options, target_urn = "None", user_urn = cert.getURN(),
                            req_ip = request.environ["REMOTE_ADDR"])

      out = Registry.performAggregateAction(action, options, cert.getURN())
      result = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=out)
    except CoreException, e:
      AWSEvent.publishError(e)
      msg = "Exception: %s" % (str(e))
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <PerformAggregateAction> failed!")
    except Exception, e:
      AWSEvent.publishError(e)
      self._log.exception("Exception")
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    finally:
      result["geni_api"] = 3
      return result

  def pub_Describe (self, urns, credentials, options):
    from ..geni.resources import Registry

    self._log.info("[%s] URNs: <%s>, Options: %s" % ("Describe", ",".join(urns), json.dumps(options)))

    try:
      #TODO: Should refactor this into a utility
      creds = []
      for cred in credentials:
        if cred["geni_type"] != "geni_sfa":
          # We don't support ABAC, go away
          return None

        creds.append(cred["geni_value"])

      for urn in urns:
        CredVerifier.checkValid(creds, "getsliceresources", Registry.getSliceURN(urn))

      # If you gave us more than one slice right now, well, too bad.
      cert = Certificate(request.environ['CLIENT_RAW_CERT'])
      AWSEvent.publishEvent("gapi3.describe", options, target_urn = urns[0], user_urn = cert.getURN(),
                            req_ip = request.environ["REMOTE_ADDR"])

      rspec = Registry.generateManifest(Registry.getSliceURN(urns[0]))
      result = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=rspec)
    except CoreException, e:
      msg = "Exception: %s" % (str(e))
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <Describe> failed!")
    except Exception, e:
      self._log.exception("Exception")
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    finally:
      result["geni_api"] = 3
      return result

  def pub_Allocate (self, slice_urn, credentials, rspec, options):
    from ..geni.resources import Registry, InsufficientGENIPrivilegesError

    self._log.info("[%s] URNs: <%s>, Options: %s" % ("Allocate", slice_urn, json.dumps(options)))

    try:
      #TODO: Should refactor this into a utility
      creds = []
      for cred in credentials:
        if cred["geni_type"] != "geni_sfa":
          # We don't support ABAC, go away
          return None

        creds.append(cred["geni_value"])

      user_info = {"_api" : 3}
      if CredVerifier.checkValid(creds, "createsliver"):
        self.recordAction("allocate", credentials, slice_urn)
        try:
          cert = Certificate(request.environ['CLIENT_RAW_CERT'])
          AWSEvent.publishEvent("gapi3.allocate", options, target_urn = slice_urn, user_urn = cert.getURN(),
                                req_ip = request.environ["REMOTE_ADDR"])
          user_info["urn"] = cert.getURN()
          user_info["email"] = cert.getEmailAddress()
          self._log.debug("Parsed user cert with URN (%(urn)s) and email (%(email)s)" % user_info)
        except Exception as e:
          self._log.exception("UNFILTERED EXCEPTION")
          user_info["urn"] = None
          user_info["email"] = None

        # Allocate has no users
        user_info["login_users"] = []

        manifest = Registry.allocateSlivers(slice_urn, creds, rspec, user_info)
        propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=manifest)
      else:
        raise InsufficientGENIPrivilegesError("createsliver")
    except foam.geni.lib.RspecParseError as e:
      msg = str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADARGS, output=msg)
      e.log(self._log, msg, logging.INFO)
    except foam.geni.lib.RspecValidationError as e:
      msg = str(e)
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.BADARGS, output=msg)
      e.log(self._log, msg, logging.INFO)
    except CoreException as e:
      if e._unexpected:
        self._log.exception("Call to known method <Allocate> failed!")
        AWSEvent.publishError(e, self._log)
      else:
        self._log.error(str(e))
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    except Exception as e:
      AWSEvent.publishError(e)
      msg = "Exception"
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception(msg)
    finally:
      return propertyList

  def pub_Provision (self, urns, credentials, options):
    from ..geni.resources import Registry, InsufficientGENIPrivilegesError

    self._log.info("[%s] URNs: <%s>, Options: %s" % ("Provision", ", ".join(urns), json.dumps(options)))

    try:
      #TODO: Should refactor this into a utility
      creds = []
      for cred in credentials:
        if cred["geni_type"] != "geni_sfa":
          # We don't support ABAC, go away
          return None

        creds.append(cred["geni_value"])

      user_info = {"_api" : 3}
      if CredVerifier.checkValid(creds, "createsliver"):
        self.recordAction("provision", credentials, urns[0])
        try:
          cert = Certificate(request.environ['CLIENT_RAW_CERT'])
          AWSEvent.publishEvent("gapi3.provision", options, target_urn = urns[0], user_urn = cert.getURN(),
                                req_ip = request.environ["REMOTE_ADDR"])
          user_info["urn"] = cert.getURN()
          user_info["email"] = cert.getEmailAddress()
          self._log.debug("Parsed user cert with URN (%(urn)s) and email (%(email)s)" % user_info)
        except Exception as e:
          self._log.exception("UNFILTERED EXCEPTION")
          user_info["urn"] = None
          user_info["email"] = None

        user_info["login_users"] = options["geni_users"]
        manifest = Registry.provisionSlivers(urns[0], creds, {"user-info" : user_info})
        propertyList = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=manifest)
      else:
        raise InsufficientGENIPrivilegesError("createsliver")
    except CoreException as e:
      if e._unexpected:
        self._log.exception("Call to known method <Provision> failed!")
        AWSEvent.publishError(e, self._log)
      else:
        self._log.error(str(e))
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    except Exception as e:
      AWSEvent.publishError(e)
      msg = "Exception"
      propertyList = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      self._log.exception(msg)
    finally:
      return propertyList

  def pub_Delete (self, urns, credentials, options):
    from ..geni.resources import Registry

    self._log.info("[%s] URNs: <%s>, Options: %s" % ("Delete", ",".join(urns), json.dumps(options)))

    user_info = {}
    try:
      #TODO: Should refactor this into a utility
      creds = []
      for cred in credentials:
        if cred["geni_type"] != "geni_sfa":
          # We don't support ABAC, go away
          return None

        creds.append(cred["geni_value"])

      # If you send a slice URN, we're bagging everything and just deleting every sliver
      sent_slice_urn = None
      for urn in urns:
        if Registry.isSliceURN(urn):
          sent_slice_urn = urn
        CredVerifier.checkValid(creds, "deletesliver", Registry.getSliceURN(urn))
      if sent_slice_urn:
        urns = [sent_slice_urn]

      # If you gave us more than one slice right now, well, too bad.
      cert = Certificate(request.environ['CLIENT_RAW_CERT'])
      user_info["urn"] = cert.getURN()
      user_info["email"] = cert.getEmailAddress()
      for urn in urns:
        AWSEvent.publishEvent("gapi3.delete", options, target_urn = urn, user_urn = cert.getURN(),
                              req_ip = request.environ["REMOTE_ADDR"])

      if sent_slice_urn:
        res = Registry.deleteSliversForSlice(urns[0], user_info)
      else:
        res = Registry.deleteSlivers(Registry.getSliceURN(urns[0]), urns, user_info)

      dinfo_list = []
      for rinfo in res:
        if not rinfo:
          continue
        if rinfo[1]:
          obj = {"geni_sliver_urn" : rinfo[0], "geni_allocation_status" : "geni_unallocated",
                 "geni_expires" : rinfo[1].strftime("%Y-%m-%dT%H:%M:%S.00Z")}
        else:
          obj = {"geni_sliver_urn" : rinfo[0], "geni_allocation_status" : "geni_unallocated",
                 "geni_expires" : datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S.00Z"),
                 "geni_error" : rinfo[2]}
        dinfo_list.append(obj)

      result = self.buildPropertyList(GENI_ERROR_CODE.SUCCESS, value=dinfo_list)
    except CoreException, e:
      msg = "Exception: %s" % (str(e))
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=msg)
      e.log(self._log, "Call to known method <Delete> failed!")
    except Exception, e:
      self._log.exception("Exception")
      result = self.buildPropertyList(GENI_ERROR_CODE.ERROR, output=str(e))
    finally:
      result["geni_api"] = 3
      return result

def setup (app):
  endpoint = "/foam/gapi/3"
  gapiver.Registry.addVersion(3, endpoint)
  gapi3 = XMLRPCHandler('gapi3')
  gapi3.connect(app, endpoint)
  gapi3.register_instance(AMAPIv3(app.logger))
  app.logger.info("[GAPIv3] Loaded.")
