# Copyright (c) 2012-2013  Barnstormer Softworks, Ltd.

import traceback
import datetime

from flask import request

from foam.config import FOAM_ROOT
from foam.api.jsonrpc import Dispatcher, route
from foam.core.json import jsonify, jsonValidate, JSONValidationError, OPTIONAL
from foam.core.configdb import ConfigDB, ConfigItem, UnknownConfigKey
from foam.geni.topology import (Attachment, TopoDB, jsonAttachmentType, GENIHost, GENIOpenFlow,
                                GENINonOpenFlow, Other, ProtoGENIHost)
from foam.types import UUID
import foam.geni.lib

class GENIAPIv1(Dispatcher):
  def __init__ (self, app):
    super(GENIAPIv1, self).__init__("GENI v1", app.logger, app)
    self._log.info("Loaded")
    from foam.geni.approval import AppData
    self.appdata = AppData

  def validate (self, rjson, types):
    return jsonValidate(rjson, types, self._log)

  @route('/geni/approval/get-macs', methods=['GET'])
  def getMACs (self):
    r = {}
    for vlan in self.appdata.iterSharedVLANs():
      r[vlan.vid] = [x for x in vlan.iterMACs()]
    return jsonify(r)

  @route('/geni/approval/get-ethertypes', methods=['GET'])
  def getEthertypes (self):
    r = {}
    for vlan in self.appdata.iterSharedVLANs():
      r[vlan.vid] = [x for x in vlan.iterEthertypes()]
    return jsonify(r)

  @route('/geni/approval/get-subnets', methods=['GET'])
  def getSubnets (self):
    r = {}
    for vlan in self.appdata.iterSharedVLANs():
      r[vlan.vid] = [x for x in vlan.iterSubnets()]
    return jsonify(r)

  @route('/geni/approval/get-reserved-vlans', methods=['GET'])
  def getReservedVLANs (self):
    return jsonify([x for x in self.appdata.iterReservedVLANs()])

  @route('/geni/approval/rebuild-db', methods=["GET"])
  def rebuildDB (self):
    self.appdata.rebuild()
    return jsonify(None)

  @route('/geni/approval/add-user-urn', methods=["POST"])
  def addUserURN (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("urn", (unicode, str))])
      u = ConfigDB.getUser(request.environ["USER"])
      self.appdata.addUserURN(objs["urn"], u)
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/remove-user-urn', methods=["POST"])
  def removeUserURN (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("urn", (unicode, str))])
      u = ConfigDB.getUser(request.environ["USER"])
      self.appdata.removeUserURN(objs["urn"], u)
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/list-user-urns', methods=["GET"])
  def listUserURNs (self):
    try:
      u = ConfigDB.getUser(request.environ["USER"])
      return jsonify(list(ConfigDB.getConfigItemByKey("geni.approval.user-urns").getValue(u)))
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg = traceback.format_exc())

  @route('/geni/approval/add-shared-vlan', methods=["POST"])
  def addSharedVLAN (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("vlan-id", (int))])
      u = ConfigDB.getUser(request.environ["USER"])
      self.appdata.addSharedVLAN(objs["vlan-id"], u)
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/remove-shared-vlan', methods=["POST"])
  def removeSharedVLAN (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("vlan-id", (int))])
      u = ConfigDB.getUser(request.environ["USER"])
      self.appdata.removeSharedVLAN(objs["vlan-id"], u)
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/list-shared-vlans', methods=["GET"])
  def listSharedVLANs (self):
    try:
      u = ConfigDB.getUser(request.environ["USER"])
      return jsonify(list(ConfigDB.getConfigItemByKey("geni.approval.shared-vlans").getValue(u)))
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg = traceback.format_exc())
    
  @route('/geni/approval/add-reserved-vlan', methods=["POST"])
  def addReservedVLAN (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("vlan-id", (int))])
      u = ConfigDB.getUser(request.environ["USER"])
      self.appdata.addAdminVLAN(objs["vlan-id"], u)
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/remove-reserved-vlan', methods=["POST"])
  def removeReservedVLAN (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("vlan-id", (int))])
      u = ConfigDB.getUser(request.environ["USER"])
      self.appdata.removeAdminVLAN(objs["vlan-id"], u)
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/list-reserved-vlans', methods=["GET"])
  def listReservedVLANs (self):
    try:
      u = ConfigDB.getUser(request.environ["USER"])
      return jsonify(list(ConfigDB.getConfigItemByKey("geni.approval.admin-vlans").getValue(u)))
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg = traceback.format_exc())

  @route('/geni/approval/create-port-group', methods=["POST"])
  def createPortGroup (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("name", (unicode, str)),
                                          ("desc", (unicode, str))])
      u = ConfigDB.getUser(request.environ["USER"])
      pg = self.appdata.createPortGroup(objs["name"], objs["desc"])
      return jsonify(str(pg.uuid))
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/remove-port-group', methods=["POST"])
  def removePortGroup (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("group-id", (unicode, str))])
      self.appdata.removePortGroup(objs["group-id"])
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/add-port', methods=["POST"])
  def addPortToGroup (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("group-id", (unicode, str)),
                                          ("dpid", (unicode, str)),
                                          ("port-num", (int))])
      pg = self.appdata.getPortGroup(objs["group-id"])
      pg.addPort(objs["dpid"], objs["port-num"])
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/remove-port', methods=["POST"])
  def removePortFromGroup (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("group-id", (unicode, str)),
                                          ("dpid", (unicode, str)),
                                          ("port-num", (int))])
      pg = self.appdata.getPortGroup(objs["group-id"])
      pg.removePort(objs["dpid"], objs["port-num"])
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/list-port-groups', methods=["GET"])
  def listPortGroups (self):
    try:
      groups = self.appdata.getPortGroups()
      l = [str(x) for x in groups.keys()]
      return jsonify(l)
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/approval/show-port-group', methods=["POST"])
  def showPortGroup (self):
    if not request.json:
      return
    try:
      objs = self.validate(request.json, [("group-id", (unicode, str))])
      pg = self.appdata.getPortGroup(objs["group-id"])
      return jsonify(pg)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/topology/nuke-attachments', methods=["GET"])
  def nukeAttachments (self):
    for attachment in TopoDB.listAttachments():
      TopoDB.removeAttachment(attachment.uuid)
    return jsonify(None)
    
  @route('/geni/topology/create-attachment', methods=["POST"])
  def createAttachment (self):
    from foam.geni.lib import jsonDPID
    if not request.json:
      return
    try:
      o = self.validate(request.json, [
          ("type", (jsonAttachmentType)),
          ("local-dpid" , (jsonDPID)),
          ("local-port" , (str, unicode)),
          ("remote-port" , (str, unicode), OPTIONAL),
          ("remote-component-id" , (str, unicode), OPTIONAL),
          ("remote-hostname", (str, unicode), OPTIONAL),
          ("desc" , (str, unicode))])
      if o["type"] == "geni-host": 
        a = GENIHost()
      elif o["type"] == "geni-of":
        a = GENIOpenFlow()
      elif o["type"] == "geni-nof":
        a = GENINonOpenFlow()
      elif o["type"] == "pg-host":
        a = ProtoGENIHost()
        a.remote_component_id = o["remote-component-id"]
      elif o["type"] == "other":
        a = Other()
        a.remote_hostname = o["remote-hostname"]
        a.remote_port = o["remote-port"]

      if o["type"].startswith("geni-"):
        a.remote_component_id = o["remote-component-id"]
        a.remote_port = o["remote-port"]

      a.dpid = o["local-dpid"]
      a.local_port = o["local-port"]
      a.desc = o["desc"]
      u = TopoDB.addAttachment(o["local-dpid"], a)
      return jsonify(u)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/topology/remove-attachment', methods=["POST"])
  def removeAttachment (self):
    if not request.json:
      return
    try:
      o = self.validate(request.json, [("uuid", (UUID))])
      TopoDB.removeAttachment(o["uuid"])
      return jsonify(None)
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/topology/list-attachments', methods=["GET"])
  def listAttachments (self):
    try:
      return jsonify(TopoDB.listAttachments())
    except JSONValidationError, e:
      jd = e.__json__()
      return jsonify(jd, code = 1, msg = jd["exception"])
    except Exception, e:
      self._log.exception("Exception")
      return jsonify(None, code = 2, msg  = traceback.format_exc())

  @route('/geni/gapi/getversion/<version>', methods=["GET"])
  def gapiGetVersion (self, version):
    from foam.geni import version as gapiver
    data = gapiver.Registry.getGAPIData(version, "<host>", "<port>")
    return jsonify(data)

  @route('/geni/gapi/listresources', methods=["GET"])
  def gapiListResources (self):
    data = foam.geni.lib.getAdvertisement()
    return jsonify(data)

def doUpgrade ():
  try:
    ver = ConfigDB.getConfigItemByKey("geni.data-version").getValue()
  except UnknownConfigKey, e:
    item = ConfigItem().setKey("geni.data-version").setValue(0)
    ConfigDB.installConfigItem(item)
    ver = ConfigDB.getConfigItemByKey("geni.data-version").getValue()

  if ver == 0:
    ConfigDB.getConfigItemByKey("geni.openflow.analysis-engine").write(True)
    ConfigDB.getConfigItemByKey("site.attachments").write({})
    ConfigDB.getConfigItemByKey("geni.data-version").write(1)

def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("geni.report-foam-version").setValue(True)
                .setDesc("Include the FOAM version in GetVersion calls")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.report-flowvisor-version").setValue(True)
                .setDesc("Include the FlowVisor version in GetVersion calls")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.report-site-info").setValue(True)
                .setDesc("Include site info in GetVersion calls")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.require-sliver-email").setValue(True)
                .setDesc("Require email address from user credential or request rspec")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.auto-sliver-priority").setValue(2000)
                .setDesc("Priority for auto-approved slivers")
                .setUpdateFuncName("foam.core.configdb.coerceInt"))

  citems.append(ConfigItem().setKey("geni.max-lease").setValue(datetime.timedelta(weeks=2))
                .setDesc("Max sliver lease time")
                .setUpdateFuncName("foam.geni.lib.updateMaxLease"))
  citems.append(ConfigItem().setKey("geni.cert-dir")
                .setValue("%s/etc/gcf-ca-certs" % (FOAM_ROOT))
                .setDesc("Directory for trusted SSL root certificates."))
  citems.append(ConfigItem().setKey("geni.site-tag").setValue(None)
                .setDesc("Site tag for GENI aggregates (used in URNs)")
                .setUpdateFuncName("foam.geni.lib.siteTagChange"))

  citems.append(ConfigItem().setKey("geni.approval.approve-on-creation").setValue(0)
                .setDesc("Approve slivers on creation.")
                .setUpdateFuncName("foam.geni.approval.setMode"))
  citems.append(ConfigItem().setKey("geni.approval.report-status").setValue(True)
                .setDesc("Report auto-approval status in GetVersion")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.openflow.analysis-engine").setValue(True)
                .setDesc("Run analysis engine for openflow slivers")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.openflow.portgroups").setValue(None)
                .setDesc("Internal Portgroups Storage")
                .setUpdateFuncName("foam.geni.approval.updatePortGroups"))
  citems.append(ConfigItem().setKey("geni.approval.user-urns").setValue(None)
                .setDesc("Internal User URN Storage")
                .setUpdateFuncName("foam.geni.approval.updateUserURNs"))
  citems.append(ConfigItem().setKey("geni.approval.admin-vlans").setValue(None)
                .setDesc("Internal storage for admin reserved VLANs")
                .setUpdateFuncName("foam.geni.approval.updateVLANs"))
  citems.append(ConfigItem().setKey("geni.approval.shared-vlans").setValue(None)
                .setDesc("Internal storage for shared VLANs")
                .setUpdateFuncName("foam.geni.approval.updateVLANs"))
  citems.append(ConfigItem().setKey("geni.approval.check-vlans").setValue(True)
                .setDesc("Try to auto-approve flowspaces with VLANs requested.")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  [ConfigDB.installConfigItem(x) for x in citems]

  api = GENIAPIv1(app)

  doUpgrade()

  return api
