# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

from __future__ import absolute_import

import base64
from cStringIO import StringIO
import datetime
import logging
import sys

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
import dateutil
from lxml import etree as ET

from ..core.configdb import ConfigDB, ConfigItem
from ..creds import CredVerifier
from .lib import RspecParseError
from ..core.log import KeyAdapter
from ..core import json
from ..core.exception import CoreException
from ..lib import _asUTC

#########
### NOTHING in the registry is persisted across restarts!
### You MUST re-register all slivers on restart
#########

class Namespace(object):
  def __init__ (self, prefix, name, location = None):
    self.prefix = prefix
    self.name = name
    self.location = location

  def __hash__ (self):
    return hash(self.name)

  def __repr__ (self):
    return self.name

class SkipHandler(Exception): pass # Really this should inherit from like "CoreFlowControl"

class BadSliverExpiration(CoreException):
  def __init__ (self, msg):
    super(BadSliverExpiration, self).__init__()
    self.msg = msg
  def __str__ (self):
    return self.msg

class UnsupportedActionError(CoreException):
  def __init__ (self, action):
    self.action = action
  def __str__ (self):
    return "Requested action (%s) not supported at this aggregate" % (self.action)

class UnsupportedResourceActionError(CoreException):
  def __init__ (self, action, urn):
    self.action = action
    self.urn = urn
  def __str__ (self):
    return "Requested action (%s) is not supported by resource (%s)" % (self.action, self.urn)

class UnknownSliceError(CoreException):
  def __init__ (self, urn, expected = True):
    super(UnknownSliceError, self).__init__()
    self.urn = urn
    self._unexpected = not expected
  def __str__ (self):
    return "Slice %s is not known at this aggregate" % (self.urn)

class UnknownSliverError(CoreException):
  def __init__ (self, urn):
    super(UnknownSliver, self).__init__()
    self.urn = urn
  def __str__ (self):
    return "Referenced sliver (%s) is not known to this aggregate." % (self.urn)

class LockedSliversError(CoreException):
  def __init__ (self, urn):
    super(LockedSliversError, self).__init__()
    self.urn = urn
  def __str__ (self):
    return "Slivers for slice (%s) are locked and will be removed when all locking users have deleted them." % (self.urn)

class InsufficientGENIPrivilegesError(CoreException):
  def __init__ (self, privilege):
    super(InsufficientGENIPrivilegesError, self).__init__()
    self.privilege = privilege
  def __str__ (self):
    return "Required privilege (%s) not present in supplied credentials." % (self.privilege)

class DecryptionNotSupportedError(CoreException):
  def __str__ (self):
    return "Private data decryption is not supported by this aggregate"



XSNS = Namespace("xsi", "http://www.w3.org/2001/XMLSchema-instance")
GENINS = Namespace("geni", "http://www.geni.net/resources/rspec/3", "http://www.geni.net/resources/rspec/3/request.xsd")
SITEINFONS = Namespace("siteinfo", "http://www.protogeni.net/resources/rspec/ext/site-info/1")

class _Registry(object):
  def __init__ (self):
    self._log = KeyAdapter("Registry", logging.getLogger('foam'))
    self._ad_callbacks = []
    self._allocate_callbacks = []
    self._provision_callbacks = []
    self._manifest_callbacks = []
    self._status_callbacks = []
    self._renew_callbacks = []
    self._delete_callbacks = []
    self._delete_sliver_callbacks = []
    self._poa_callbacks = {}
    self._paa_callbacks = {}
    self._ad_namespaces = set()
    self._ad_loclist = []
    self._ampkey = None

    self._mfest_namespaces = set()
    self._mfest_loclist = []

    # Keys are Slice URNs, values are lists of sliver URNs for that slice
    self._sliceResources = {}

    # { slice_urn : [ user_urn, ...], ... }
    self._del_lock_data = {}

    self.addAdNamespace(XSNS)
    self.addManifestNamespace(XSNS)

  def setup (self):
    self._sign = ConfigDB.getConfigItemByKey("geni.sign.enable").getValue()
    self._key_path = ConfigDB.getConfigItemByKey("geni.sign.key-path").getValue()
    self._key_data = None
    self._cert_path = ConfigDB.getConfigItemByKey("geni.sign.cert-path").getValue()
    self._cert_data = None
    self._del_lock_data = ConfigDB.getConfigItemByKey("geni.delete-lock-data").getValue()
    for k,v in self._del_lock_data.iteritems():
      self._del_lock_data[k] = set(v)

    pkey_path = ConfigDB.getConfigItemByKey("geni.rspec-pkey-path").getValue()
    if pkey_path:
      with open(pkey_path, "rb") as kf:
        self._ampkey = serialization.load_pem_private_key(kf.read(), password=None, backend=default_backend())

  @property
  def key_data (self):
    if not self._key_data:
      if self._key_path:
        self._key_data = open(self._key_path, "rb").read()
    return self._key_data

  @property
  def cert_data (self):
    if not self._cert_data:
      if self._cert_path:
        self._cert_data = open(self._cert_path, "rb").read()
    return self._cert_data
      
  def _signXMLDocument (self, element):
    # FIXME: This is totally broken now that xmldsig doesn't work
    from signxml import xmldsig

    try:
      xmldsig(element).sign(key = self.key_data, cert = self.cert_data)
    except Exception,e:
      self._log.debug(e.err_code)

  def decrypt (self, data):
    if not self._ampkey:
      raise DecryptionNotSupportedError()

    ciphertext = base64.b64decode(data)

    from cryptography.hazmat.primitives.asymmetric import padding
    from cryptography.hazmat.primitives import hashes
    return self._ampkey.decrypt(ciphertext, padding.OAEP(padding.MGF1(hashes.SHA1()), hashes.SHA1(), None))

  def registerSliver (self, slice_urn, sliver_urn):
    self._log.info("Registering sliver %s for slice %s" % (sliver_urn, slice_urn))
    self._sliceResources.setdefault(slice_urn, set()).add(sliver_urn)

  def unregisterSliver (self, slice_urn, sliver_urn):
    self._log.info("Unregistering sliver %s for slice %s" % (sliver_urn, slice_urn))
    try:
      self._sliceResources[slice_urn].remove(sliver_urn)
    except ValueError:
      pass # It wasn't here.  No problem!

  def getSlivers (self, slice_urn):
    try:
      return list(self._sliceResources[slice_urn])
    except KeyError:
      return []
      
  def getSliceURN (self, sliver_urn):
    if sliver_urn in self._sliceResources.iterkeys():
      # This is a slice URN, just return it
      return sliver_urn

    # TODO: Should reverse index to make this more efficient
    for k,v in self._sliceResources.iteritems():
      if sliver_urn in v:
        return k
    return None

  def isSliceURN (self, urn):
    return urn in self._sliceResources.iterkeys()

  def addAdNamespace (self, ns):
    self._log.info("Adding Advertisement namespace [%s] [%s]" % (ns.name, ns.location))
    if ns in self._ad_namespaces:
      return
    self._ad_namespaces.add(ns)

    if ns.location is not None:
      self._ad_loclist.append(ns.name)
      self._ad_loclist.append(ns.location)

  def addManifestNamespace(self, ns):
    self._log.info("Adding Manifest namespace [%s] [%s]" % (ns.name, ns.location))
    if ns in self._mfest_namespaces:
      return
    self._mfest_namespaces.add(ns)

    if ns.location is not None:
      self._mfest_loclist.append(ns.name)
      self._mfest_loclist.append(ns.location)

  def deleteSliversForSlice (self, slice_urn, user_info):
    if slice_urn in self._del_lock_data:
      users = self._del_lock_data[slice_urn]
      users.discard(user_info["urn"])
      if users:
        raise LockedSliversError(slice_urn)
      else:
        del self._del_lock_data[slice_urn]

    # TODO: for 'False' accumulate errors
    res = []
    for cb in self._delete_callbacks:
      res.append(cb(slice_urn, user_info))
    if False in res:
      self._log.info("Failed to delete slivers for slice %s" % (slice_urn))
    if None in res:
      self._log.info("No slivers for slice %s" % (slice_urn))
      raise UnknownSliceError(slice_urn)
    return []

  def deleteSlivers (self, slice_urn, sliver_urns, user_info):
    if slice_urn in self._del_lock_data:
      users = self._del_lock_data[slice_urn]
      users.discard(user_info["urn"])
      if users:
        raise LockedSliversError(slice_urn)
      else:
        del self._del_lock_data[slice_urn]

    res = []
    for cb in self._delete_sliver_callbacks:
      res.extend(cb(slice_urn, sliver_urns, user_info))
    if not res:
      self._log.info("No slivers for slice %s" % (slice_urn))
      raise UnknownSliceError(slice_urn)
    return res

  def renewSliversForSlice (self, slice_urn, newexp, creds):
    reqexp = dateutil.parser.parse(str(newexp))
    reqexp = _asUTC(reqexp)

    max_expiration = _asUTC(datetime.datetime.utcnow()) + ConfigDB.getConfigItemByKey("geni.max-lease").getValue()

    if reqexp > max_expiration:
      raise BadSliverExpiration(
        "The requested expiration date (%s) is past the allowed maximum expiration (%s)." %
          (reqexp, max_expiration))

    valid_exp = False
    for cred in creds:
      credexp = _asUTC(cred.expiration)
      if reqexp > credexp:
        continue
      else:
        valid_exp = True
        break

    if not valid_exp:
      raise BadSliverExpiration(
          "No credential found whose expiration is greater than or equal to the requested sliver expiration (%s)" %
            (reqexp))

    res = set()
    for cb in self._renew_callbacks:
      res.add(cb(slice_urn, reqexp))
    if False in res:
      self._log.info("Failed to renew slivers for slice %s" % (slice_urn))
    if True in res:
      return
    if None in res:
      self._log.info("No slivers for slice %s" % (slice_urn))
      raise UnknownSliceError(slice_urn)

  def generate (self):
    NSMAP = {}
    for namespace in self._ad_namespaces:
      NSMAP[namespace.prefix] = namespace.name
    NSMAP[GENINS.prefix] = GENINS.name
    NSMAP[SITEINFONS.prefix] = SITEINFONS.name

    rspec = ET.Element("{%s}rspec" % (GENINS.name), nsmap = NSMAP)
    rspec.attrib["{%s}schemaLocation" % (XSNS.name)] = " ".join(self._ad_loclist)
    rspec.attrib["type"] = "advertisement"

    for cb in self._ad_callbacks:
      cb(rspec)

    # Default GENI stuff
    locdata = ConfigDB.getConfigItemByKey("geni.location-data").getValue()
    if locdata:
      sie = ET.SubElement(rspec, "{%s}site_info" % (SITEINFONS))
      loce = ET.SubElement(sie, "{%s}location" % (SITEINFONS))
      loce.attrib["country"] = locdata["country"]
      loce.attrib["longitude"] = str(locdata["long"])
      loce.attrib["latitude"] = str(locdata["lat"])

    if self._sign:
      self._signXMLDocument(rspec)

    xml = StringIO()
    ET.ElementTree(rspec).write(xml)
    return xml.getvalue()

  def generateManifest (self, slice_urn):
    NSMAP = {}
    for namespace in self._mfest_namespaces:
      NSMAP[namespace.prefix] = namespace.name
    NSMAP[GENINS.prefix] = GENINS.name

    manifest = ET.Element("{%s}rspec" % (GENINS.name), nsmap = NSMAP)
    manifest.attrib["{%s}schemaLocation" % (XSNS.name)] = " ".join(self._ad_loclist)
    manifest.attrib["type"] = "manifest"

    res = set()
    for cb in self._manifest_callbacks:
      res.add(cb(slice_urn, manifest))
    if False in res:
      self._log.info("Errors while generating manifest for slice %s" % (slice_urn))
    if None in res:
      self._log.info("No slivers for slice %s" % (slice_urn))
      raise UnknownSliceError(slice_urn)

    if self._sign:
      self._signXMLDocument(manifest)

    xml = StringIO()
    xml.write(ET.tostring(manifest, pretty_print=True))
    return xml.getvalue()

  def allocateSlivers (self, slice_urn, credentials, rspec, user_info):
    creds = CredVerifier.fromStrings(credentials, "createsliver", slice_urn)
    try:
      s = StringIO(rspec)
      rspec_dom = ET.parse(s)
    except Exception, exc:
      raise RspecParseError(slice_urn, str(exc))

    NSMAP = {}
    for namespace in self._mfest_namespaces:
      NSMAP[namespace.prefix] = namespace.name
    NSMAP[GENINS.prefix] = GENINS.name

    manifest = ET.Element("{%s}rspec" % (GENINS.name), nsmap = NSMAP)
    manifest.attrib["{%s}schemaLocation" % (XSNS.name)] = " ".join(self._ad_loclist)
    manifest.attrib["type"] = "manifest"

    for cb in self._allocate_callbacks:
      cb(slice_urn, creds, rspec_dom, user_info, manifest)

    if self._sign:
      self._signXMLDocument(manifest)

    xml = StringIO()
    xml.write(ET.tostring(manifest, pretty_print=True))
    return xml.getvalue()

  def provisionSlivers (self, slice_urn, credentials, options):
    creds = CredVerifier.fromStrings(credentials, "createsliver", slice_urn)
    failed = False
    failed_exc = None
    for cb in self._provision_callbacks:
      try:
        cb(slice_urn, creds, options)
      except Exception as e:
        failed = True
        failed_exc = sys.exc_info()
        break

    if failed:
      user_info = {"urn" : "PROVISIONFAILURE", "email" : "PROVISIONFAILURE"}
#      for cb in self._delete_callbacks:
#        try:
#          cb(slice_urn, user_info)
#        except Exception as e:
#          self._log.exception("Cleanup exception AFTER failed provision", e)
      raise failed_exc[0], failed_exc[1], failed_exc[2]

    return self.generateManifest(slice_urn)

  def sliverStatus (self, urns, options):
    r = {}
    for cb in self._status_callbacks:
      cb(r, urns, options)
    return r

  def performOperationalAction (self, urns, action, options, user_urn):
    try:
      cblist = self._poa_callbacks[action]
    except KeyError, e:
      raise UnsupportedActionError(action)
    outlist = []
    for cb in cblist:
      try:
        outlist.append(cb(action, urns, options, user_urn))
      except SkipHandler:
        continue
    # Nasty hack for backwards compatibility
    if len(outlist) == 1:
      return outlist[0]
    return outlist

  def performAggregateAction (self, action, options, user_urn):
    try:
      cb = self._paa_callbacks[action]
    except KeyError, e:
      raise UnsupportedActionError(action)
    return cb(action, options, user_urn)

  def addAdvertisementCallback (self, cb):
    self._ad_callbacks.append(cb)

  def addAllocateCallback (self, cb):
    self._allocate_callbacks.append(cb)

  def addProvisionCallback (self, cb):
    self._provision_callbacks.append(cb)

  def addManifestCallback (self, cb):
    self._manifest_callbacks.append(cb)

  def addDeleteCallback (self, cb):
    self._delete_callbacks.append(cb)

  def addDeleteSliverCallback (self, cb):
    self._delete_sliver_callbacks.append(cb)

  def addStatusCallback (self, cb):
    self._status_callbacks.append(cb)

  def addRenewCallback (self, cb):
    self._renew_callbacks.append(cb)

  def addPOACallback (self, action, cb):
    if action in self._poa_callbacks and cb in self._poa_callbacks[action]:
      return

    if hasattr(cb, "_realinfo"):
      self._log.info("Adding POA [%s] callback [%s:%s]" % (action, cb._realinfo[0], cb._realinfo[1]))
    else:
      self._log.info("Adding POA [%s] callback [%s:%s]" % (action, cb.__module__, cb.__name__))

    self._poa_callbacks.setdefault(action, set()).add(cb)

  def addPAACallback (self, action, cb):
    self._log.debug("Adding PAA callback for [%s]" % (action))
    self._paa_callbacks[action] = cb

  ### GLOBAL POAS ###
 
  def poaSetDeleteLock (self, action, urns, options, user_urn):
    for urn in urns:
      self._del_lock_data.setdefault(urn, set()).add(user_urn)
    ConfigDB.getConfigItemByKey("geni.delete-lock-data").write(self._del_lock_data)
    return True

  def poaGetDeleteLocks (self, action, urns, options, user_urn):
    d = {}
    for urn in urns:
      d[urn] = list(self._del_lock_data.setdefault(urn, set()))
    return d


Registry = _Registry()
Registry.addPOACallback("geni:set-delete-lock", Registry.poaSetDeleteLock)
Registry.addPOACallback("geni:get-delete-locks", Registry.poaGetDeleteLocks)

def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("geni.sign.enable").setValue(False)
                .setDesc("Sign returned manifests and advertisements")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("geni.sign.key-path").setValue(None)
                .setDesc("Key path for signing documents"))
  citems.append(ConfigItem().setKey("geni.sign.cert-path").setValue(None)
                .setDesc("Certificate path for signing documents"))
  citems.append(ConfigItem().setKey("geni.location-data").setValue({})
                .setDesc("Location data for aggregate manager"))
  citems.append(ConfigItem().setKey("geni.delete-lock-data").setValue({})
                .setDesc("Storage for sliver delete lock metadata"))
  citems.append(ConfigItem().setKey("geni.rspec-pkey-path").setValue(None)
                .setDesc("Path for private key for request rspec data the AM needs to consume"))
  [ConfigDB.installConfigItem(x) for x in citems]

  Registry.setup()

