# Copyright (c) 2012-2013 Barnstormer Softworks

import logging

from foam.core.exception import CoreException
from foam.core.log import KeyAdapter
from foam.core.configdb import ConfigDB, ConfigItem
from foam.core.json import JSONInput
from foam.types import UUID
from foam.openflow.types import DPID

class IllegalAttachmentsValue(CoreException):
  def __init__ (self, newval):
    self.val = newval
  def __str__ (self):
    return "Tried to set attachments config to illegal value: %s" % (self.val)

class Attachment(object):
  def __init__ (self):
    self.uuid = UUID()
    self.dpid = None
    self.local_port = None
    self.remote_port = None
    self.desc = None

  def __json__ (self):
    d = {"uuid" : self.uuid, "local-dpid" : self.dpid, "local-port" : self.local_port,
        "remote-port" : self.remote_port, "description" : self.desc}
    return d

class GENIAttachment(Attachment):
  def __init__ (self):
    super(GENIAttachment, self).__init__()
    self.remote_component_id = None
    self.type = "geni-attachment"

  def __json__ (self):
    base = super(GENIAttachment, self).__json__()
    d = {"remote-component-id" : self.remote_component_id, "type" : self.type}
    base.update(d)
    return base

class GENIHost(GENIAttachment):
  def __init__ (self):
    super(GENIHost, self).__init__()
    self.type = "geni-host"

class ProtoGENIHost(GENIAttachment):
  def __init__ (self):
    super (ProtoGENIHost, self).__init__()
    self.type = "pg-host"

class GENIOpenFlow(GENIAttachment):
  def __init__ (self):
    super(GENIOpenFlow, self).__init__()
    self.type = "geni-of"

class GENINonOpenFlow(GENIAttachment):
  def __init__ (self):
    super(GENINonOpenFlow, self).__init__()
    self.type = "geni-nof"

class Other(Attachment):
  def __init__ (self):
    super(Other, self).__init__()
    self.type = "other"
    self.remote_hostname = None

  def __json__ (self):
    base = super(Other, self).__json__()
    d = {"remote-hostname" : self.remote_hostname}
    base.update(d)
    return base

class _TopoDB(object):
  def __init__ (self):
    self._log = KeyAdapter("GENI-TopoDB", logging.getLogger("foam"))
    self._attachments = ConfigDB.getConfigItemByKey("site.attachments").getValue()
    self._attachmentsByDPID = {}
    for k,v in self._attachments.iteritems():
      self._attachmentsByDPID.setdefault(v.dpid, {}).setdefault(v.local_port, []).append(v)

  def addAttachment (self, dpid, attachment):
    attachment.dpid = dpid
    self._attachments[attachment.uuid] = attachment
    self._attachmentsByDPID.setdefault(dpid, {}).setdefault(attachment.local_port, []).append(attachment)
    self.writeAttachments()

  def removeAttachment (self, uuid):
    a = self._attachments[uuid]
    del self._attachments[uuid]
    try:
      ports = self._attachmentsByDPID[a.dpid][a.local_port]
      del ports[ports.index(a)]
    except Exception, e:
      pass
    self.writeAttachments()

  def writeAttachments (self):
    ConfigDB.getConfigItemByKey("site.attachments").write(self._attachments)

  def listAttachments (self):
    return self._attachments.values()

  def getDPIDAttachments (self, dpid):
    try:
      d = DPID(dpid)
      return self._attachmentsByDPID[d]
    except KeyError, e:
      return {}

  def getAttachment (self, aid):
    return self._attachments[aid]

def updateAttachments (key, newval):
  if type(newval) is not dict:
    raise IllegalAttachmentsValue(newval)
  return newval

TopoDB = None

class jsonAttachmentType(JSONInput):
  TYPES = set(["geni-host", "geni-of", "geni-nof", "other", "pg-host"])

  @staticmethod
  def validate (obj):
    if not isinstance(obj, (str, unicode)):
      return False
    if str(obj) in jsonAttachmentType.TYPES:
      return True
    return False

  @staticmethod
  def construct (obj):
    return str(obj)

def setup(app):
  global TopoDB

  citems = []
  citems.append(ConfigItem().setKey("site.attachments").setValue({})
                .setDesc("")
                .setUpdateFuncName("foam.geni.topology.updateAttachments"))
  [ConfigDB.installConfigItem(x) for x in citems]

  TopoDB = _TopoDB()


