# Copyright (c) 2013-2017  Barnstormer Softworks, Ltd.

import copy
import os.path
import os

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from foam.core.configdb import ConfigDB
import foam.version
from foam.flowvisor import Connection as FV

class _Registry(object):
  def __init__ (self):
    self.supported_versions = {}
    self._callbacks = []
    self._static = {}
    self._init_static()
    self._init_trust_roots()

  def _init_static (self):
    self._static = {
         "geni_request_rspec_versions" : [
            { 'extensions': [ 'http://www.geni.net/resources/rspec/ext/openflow/3', ],
               'namespace': 'http://www.geni.net/resources/rspec/3',
               'schema': 'http://www.geni.net/resources/rspec/3/request.xsd',
               'type': 'GENI',
               'version': '3'}
            ],
         "geni_ad_rspec_versions" : [
            { 'extensions': [ 'http://www.geni.net/resources/rspec/ext/openflow/3',
                              'http://geni.bssoftworks.com/rspec/ext/topo/1', ],
              'namespace': 'http://www.geni.net/resources/rspec/3',
              'schema': 'http://www.geni.net/resources/rspec/3/ad.xsd',
              'type': 'GENI',
              'version': '3'}
            ],
         "geni_api_versions" : { },
         }

  def _init_trust_roots (self):
    trinfo = []
    cdir = ConfigDB.getConfigItemByKey("geni.cert-dir").getValue()
    for name in os.listdir(cdir):
      try:
        data = {}
        cert = x509.load_pem_x509_certificate(open("%s/%s" % (cdir, name), "rb").read(), default_backend())
        cnlist = cert.issuer.get_attributes_for_oid(NameOID.COMMON_NAME)
        data["CN"] = cnlist[0].value
        data["Serial"] = str(cert.serial)
        data["Expires"] = cert.not_valid_after.strftime("%Y-%m-%d")
        urns = []
        for ext in cert.extensions:
          if ext.oid == x509.SubjectAlternativeName.oid:
            urns.extend(ext.value.get_values_for_type(x509.UniformResourceIdentifier))
        data["SubjAltNameURIs"] = urns
        trinfo.append(data)
      except ValueError:
        continue
    self._static["trusted-roots"] = trinfo

  def addStaticInfo (self, info):
    self._static.update(info)

  def addVersion (self, ver, endpoint):
    self.supported_versions[ver] = endpoint

  def addCallback (self, cb):
    self._callbacks.append(cb)

  def getGAPIData (self, ver, addr, port):
    d = copy.copy(self._static)
    d["geni_api"] = ver

    for (v,e) in self.supported_versions.iteritems():
      url = "https://%s:%s%s" % (addr, port, e)
      d["geni_api_versions"][str(v)] = url

    if ConfigDB.getConfigItemByKey("geni.report-foam-version").getValue():
      d["foam_version"] = foam.version.VERSION

    if ConfigDB.getConfigItemByKey("geni.report-site-info").getValue():
      d["site_info"] = self.generateSiteInfo()

    pkey_path = ConfigDB.getConfigItemByKey("geni.rspec-pkey-path").getValue()
    if pkey_path:
      with open(pkey_path, "rb") as kf:
        pkey = serialization.load_pem_private_key(kf.read(), password=None, backend=default_backend())
      pub = pkey.public_key()
      d["request.pubkey"] = pub.public_bytes(serialization.Encoding.PEM,
                                          serialization.PublicFormat.SubjectPublicKeyInfo)
      d["request.padding"] = "OAEP"
      d["request.hash"] = "sha1"

    for cb in self._callbacks:
      d.update(cb(ver))

    return d
    
  def generateSiteInfo (self):
    """Generate site's admin, location and description info"""
    dmap = [("site.admin.name", "admin-name"),
            ("site.admin.email", "admin-email"),
            ("site.admin.phone", "admin-phone"),
            ("site.location.address", "org-address"),
            ("site.location.organization", "org-name"),
            ("site.description", "description")]

    sinfo = {}
    for ckey, vkey in dmap:
      val = ConfigDB.getConfigItemByKey(ckey).getValue()
      if val is not None:
        sinfo[vkey] = val

    return sinfo
    
Registry = _Registry()
