# Copyright (c) 2011-2012  The Board of Trustees of The Leland Stanford Junior University
# Copyright (c) 2012  Barnstormer Softworks, Ltd.

from foam.core.exception import CoreException

class MergeDPIDMismatch(CoreException):
  def __init__ (self, a, b):
    super(MergeDPIDMismatch, self).__init__()
    self.a = a
    self.b = b
  def __str__ (self):
    return "Attempt to merge mismatched datapaths: %s, %s" % (self.a, self.b)

class Port(object):
  def __init__ (self):
    self.name = None
    self.num = None
    self.__dpid = None

  @property
  def dpid (self):
    return self.__dpid

  @dpid.setter
  def dpid (self, value):
    self.__dpid = value.lower()

  def __str__ (self):
    if self.num is None or self.name is None:
      return "[%s] %s" % (str(self.num), str(self.name))
    return "[%d] %s" % (self.num, self.name)

  def __hash__ (self):
    return hash("%s-%d" % (self.dpid, self.num))

  def __json__ (self):
    return { "num" : self.num, "name" : self.name, "dpid" : self.dpid }

  def __cmp__ (self, other):
    if self.dpid == other.dpid:
      if self.num < other.num:
        return -1
      elif self.num == other.num:
        return 0
      elif self.num > other.num:
        return 1
    elif self.dpid < other.dpid:
      return -1
    elif self.dpid > other.dpid:
      return 1


class DPID(object):
  MAX = (2 ** 64) - 1

  class OutOfRangeError(CoreException):
    def __init__ (self, val):
      super(OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%d) out of range of valid DPIDs" % (self.val)

  class InputTypeError(CoreException):
    def __init__ (self, val):
      super(InputTypeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  def __init__ (self, val):
    self._dpid = None

    if isinstance(val, (unicode, str)):
      val = int(val.replace(":", ""), 16)

    if isinstance(val, (int, long)):
      if val < DPID.MAX and val >= 0:
        self._dpid = val
      else:
        raise DPID.OutOfRangeError(val)
    else:
      raise DPID.InputTypeError(val)

  def __eq__ (self, other):
    return self._dpid == other._dpid

  def __hash__ (self):
    return self._dpid

  def __str__ (self):
    s = self.hexstr()
    return ":".join(["%s%s" % (s[x], s[x+1]) for x in xrange(0,15,2)])

  def __repr__ (self):
    return str(self)

  def __json__ (self):
    return str(self)
  
  def hexstr (self):
    return "%016x" % (self._dpid) 

  @property
  def int (self):
    return self._dpid


class EthernetMAC (object):
  MAX = (2 ** 48) - 1

  class OutOfRangeError(CoreException):
    def __init__ (self, val):
      super(OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%d) out of range of valid Ethernet Addresses" % (self.val)

  class InputTypeError(CoreException):
    def __init__ (self, val):
      super(InputTypeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  def __init__ (self, val):
    self._mac = None

    if isinstance(val, (unicode, str)):
      val = val.replace(":", "")
      val = val.replace("-", "")
      val = int(val, 16)

    if isinstance(val, (long, int)):
      if val < EthernetMAC.MAX and val >= 0:
        self._mac = val
      else:
        raise EthernetMAC.OutOfRangeError(val)
    else:
      raise EthernetMAC.InputTypeError(val)

  def __eq__ (self, other):
    return self._mac == other._mac

  def __hash__ (self):
    return self._mac

  def __str__ (self):
    s = self.hexstr()
    return ":".join(["%s%s" % (s[x], s[x+1]) for x in range(0,11,2)])

  def __json__ (self):
    return str(self)

  def __repr__ (self):
    return str(self)

  def hexstr (self):
    return "%012x" % (self._mac)

class Datapath(object):
  def __init__ (self):
    self.dpid = None
    self.ports = set()

  def __str__ (self):
    return "[%s] %s\n" % (self.dpid,
        "    ".join(["%s" % (str(x)) for x in self.ports]))

  def __json__ (self):
    return {"dpid" : self.dpid,
            "ports" : [x.__json__() for x in self.ports]}

  #TODO: Should this return a new datapath?
  def merge (self, other):
    if self.dpid != other.dpid:
      raise MergeDPIDMismatch(self.dpid, other.dpid)

    if (len(other.ports) == 0) or (len(self.ports) == 0):
      self.ports = set()
      return

    for p in other.ports:
      self.ports.add(p)

  def addPort (self, port):
    self.ports.add(port)


class Controller(object):
  TYPES = set(["primary", "backup", "monitor"])

  def __init__ (self):
    self.type = None
    self.url = None

  def __str__ (self):
    return "%s: %s" % (self.type, self.url)

  def __json__ (self):
    return {"type" : self.type, "url" : self.url}
