# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import traceback

from flask import request

from . import data
from . import geni
from . import hostdriver
from .drivers import docker
from .drivers.docker import CM
from .drivers import dropbox
from .drivers import dockerhgvols
from .drivers import dockerdropboxvols
from . import ports as Ports
from ..api.jsonrpc import Dispatcher, route
from ..core.json import jsonify, jsonify_ephemeral, jsonValidate, JSONValidationError
from ..core.configdb import ConfigDB
from .drivers.ovsdb import OVSDBInfoDriver

class VTSAdminAPIv1(Dispatcher):
  def __init__ (self, app):
    super(VTSAdminAPIv1, self).__init__("VTS Admin v1", app.logger, app)
    self._log.info("Loaded")
    self._data = data.DB

  def validate (self, rjson, types):
    return jsonValidate(rjson, types, self._log)

  @route('/core/admin/vts/image', methods=['POST'])
  def addImage (self):
    if not request.json:
      return
    o = self.validate(request.json, [("name", (unicode, str))])
    self._data.addImage(o["name"])
    return jsonify(None)

  @route('/core/admin/vts/images', methods=['GET'])
  def getImages (self):
    return jsonify(self._data.images)

  @route('/core/admin/vts/image-data', methods=['GET'])
  def getImageData (self):
    outobjs = {}
    for desc in CM.imagedescs:
      obj = {}
    return jsonify([])

  @route('/core/admin/vts/pgvlan', methods=['POST'])
  def addPGVlan (self):
    if not request.json:
      return
    o = self.validate(request.json, [("name", (unicode, str)),
                                     ("vid", (int))])
    self._data.addPGVLAN(o["name"], o["vid"])
    return jsonify(None)

  @route('/core/admin/vts/pgvlans', methods=['GET'])
  def getPGVlans (self):
    return jsonify(self._data.vlans)

  ##### TODO
  ### These two methods are hacks for the fact that local circuit planes
  ### aren't a real thing right now
  @route('/core/admin/vts/pg-cmid/<cmid>', methods=['POST'])
  def setPGCMID (self, cmid):
    self._data.setPGComponentManager(cmid)
    return jsonify(None)

  @route('/core/admin/vts/pg-cmid', methods=['GET'])
  def getPGCMID (self):
    return jsonify(self._data._pgcmid)

  @route('/core/admin/vts/status', methods=['GET', 'POST'])
  def getpostStatus (self):
    if request.method == "POST":
      if not request.json:
        return
    elif request.method == "GET":
      pass

  @route('/core/admin/vts/vf/sslvpn/local-ip', methods=['POST'])
  def postSSLVPNLocalIP (self):
    if not request.json:
      return
    ip = request.json
    ConfigDB.getConfigItemByKey("vts.sslvpn.local-ip").write(ip)
    return jsonify(None)

  @route('/core/admin/vts/slivers', methods=['GET'])
  def getSlivers (self):
    slist = []
    for sliver in self._data.slivers:
      pgcircuits = 0
      for dp in sliver.datapaths:
        for port in dp.ports.itervalues():
          if port.type == "pglocal":
            pgcircuits += 1
      
      hgvols = 0
      dbxvols = 0
      for ct in sliver.containers:
        hgvols += len(ct._hgvols)
        dbxvols += len(ct._dbxvols)

      slist.append({"urn" : sliver.urn, "slice-urn" : sliver.slice_urn, "user-urn" : sliver.user_urn,
                    "datapath-count" : len(sliver.datapaths), "container-count" : len(sliver.containers),
                    "uuid" : sliver.uuid, "expiration" : sliver.expstr, "pg-circuits" : pgcircuits,
                    "hg-vols" : hgvols, "dropbox-vols" : dbxvols, "provisioned" : sliver._provisioned})
    return jsonify(slist)

  @route('/core/admin/vts/slice/<slice_urn>', methods=['DELETE'])
  def deleteSlice (self, slice_urn):
    # Delete all slivers for a slice
   data.DB.deleteSlice(slice_urn)
   return jsonify(None)

  @route('/core/admin/vts/dropbox/<sliver_uuid>', methods=['DELETE'])
  def deleteDropboxPath (self, sliver_uuid):
    # Clean up dropbox paths for slivers that might no longer be here, or are abusive
    DBVM = dockerdropboxvols.DropboxVolumeManager
    DBVM.destroyTree(sliver_uuid)
    return jsonify(None)

  @route('/core/admin/vts/slices', methods=['GET'])
  def getSlices (self):
    return jsonify(self._data.slices)

  @route('/core/admin/vts/sliver/<sliver_urn>/datapaths', methods=['GET'])
  def getSliverDatapaths (self, sliver_urn):
    sliver = self._data.getSliver(sliver_urn)
    return jsonify_ephemeral(sliver.datapaths)

  @route('/core/admin/vts/sliver/<sliver_urn>/containers', methods=['GET'])
  def getSliverContainers (self, sliver_urn):
    sliver = self._data.getSliver(sliver_urn)
    return jsonify_ephemeral(sliver.containers)

  @route('/core/admin/vts/sliver/<sliver_urn>/deallocate', methods=['GET'])
  def deallocateSliver (self, sliver_urn):
    self._data.deallocateSliver(sliver_urn)
    return jsonify(None)


  @route('/core/admin/vts/sliver/<sliver_urn>/datapath/<dpname>/port/<client_id>', methods=['PUT', 'DELETE'])
  def putdeleteDatapathPort (self, sliver_urn, dpname, client_id):
    if request.method == "DELETE":
      sliver = self._data.getSliver(sliver_urn)
      sliver.removeDatapathPort(dpname, client_id)
      return jsonify(None)
    elif request.method == "PUT":
      if not request.json:
        return
      d = request.json
      sliver = self._data.getSliver(sliver_urn)
      if d["type"] == "pg-local":
        p = Ports.PGLocal(None)
        p.client_id = client_id
        p.circuit = d["shared-lan"]
        self._data.allocateLocalCircuit(p.circuit)
        sliver.addDatapathPort(dpname, p)
      # TODO: Other types
      return jsonify(None)

  @route('/core/admin/vts/sliver/<sliver_urn>/request-rspec', methods=["GET"])
  def getRequestRspec (self, sliver_urn):
    sliver = self._data.getSliver(sliver_urn)
    return jsonify(sliver.rspec)

  @route('/core/admin/vts/sliver/<sliver_urn>/manifest-rspec', methods=["GET"])
  def getManifestRspec (self, sliver_urn):
    from ..geni import resources
    mfest = resources.Registry.generateManifest(sliver_urn)
    return jsonify(mfest)

  @route('/core/admin/vts/slice/<urn>', methods=['GET'])
  def getSliver (self, urn):
    return jsonify(self._data.getSliversBySlice(urn))

  @route('/core/admin/vts/target-bridge', methods=["POST"])
  def postTargetBridge (self):
    if not request.json:
      return
    o = self.validate(request.json, [("name", (unicode, str)),
                                     ("brname", (unicode, str))])
    self._data.addTargetBridge(o["name"], o["brname"])
    return jsonify(None)

  @route('/core/admin/vts/target-bridges', methods=["GET"])
  def getTargetBridges (self):
    return jsonify(self._data.target_bridges)

  @route('/core/admin/vts/circuitplane/<typ>', methods=["PUT", "DELETE"])
  def putCircuitPlane (self, typ):
    if request.method == "DELETE":
      self._data.removeCircuitPlane(typ) # <typ> is really the label here
      return jsonify(None)
    elif request.method == "PUT":
      if typ == 'ipv4':
        o = self.validate(request.json, data.CircuitPlanes.IPv4.VALFIELDS)
        cp = data.CircuitPlanes.IPv4.jconstruct(o)
        self._data.addCircuitPlane(cp)
        return jsonify(None)
      return jsonify(None, -1, "Unknown type")

  @route('/core/admin/vts/ovs/bridges', methods=["GET"])
  def getBridges (self):
    ovsd = OVSDBInfoDriver(self._log)
    return jsonify(ovsd.getBridges())

  @route('/core/admin/vts/ovs/bridge/<name>', methods=["DELETE"])
  def deleteBridge (self, name):
    ovsd = OVSDBInfoDriver(self._log)
    return jsonify(ovsd.deleteBridge(name))

  @route('/core/admin/vts/container/<name>/nat', methods=["PUT"])
  def addContainerNAT (self, name):
    if not request.json:
      return
    o = self.validate(request.json, [("ip-address", (unicode, str)),
                                     ("gateway", (unicode, str))])
    driver = docker.DockerDriver(self._log.info)
    return jsonify(driver.setupNAT(data.PNR, name, o["ip-address"], o["gateway"]))

  @route('/core/admin/vts/sliver/<sliver_urn>/gapi-owner-lock', methods=["PUT"])
  def putGAPIOwnerLock (self, sliver_urn):
    if not request.json:
      return
    sliver = self._data.getSliver(sliver_urn)
    obj = self.validate(request.json, [("lock-enable", (list)),
                                       ("lock-disable", (list))])
    sliver.gapi_lock.enable(obj["lock-enable"])
    sliver.gapi_lock.disable(obj["lock-disable"])
    return jsonify(None)


def setup (app):
  dropbox.setup(app)
  docker.setup(app)
  hostdriver.setup(app)
  data.setup(app)
  geni.setup(app)
  dockerhgvols.setup(app)
  dockerdropboxvols.setup(app)
  api = VTSAdminAPIv1(app)
  return api
