# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime
import hashlib
import random
import uuid
import string

import IPy

from ..core.allocation import Allocation
from ..core.configdb import ConfigDB, ConfigItem
from ..core.exception import CoreException
from ..core.log import KeyAdapter
from ..core import json
from ..core import policy
from .drivers.docker import CM, DockerDriver, RawHostVolume
from .drivers.dstat import Driver as StatDriver
from .drivers.dockerhgvols import VolumeInfo as HGVolumeInfo
from .drivers.dockerhgvols import HGVolumeManager
from .drivers.dockerdropboxvols import VolumeInfo as DropboxVolumeInfo
from .drivers.dockerdropboxvols import DropboxVolumeManager
from .drivers import ovs as OVS
from .functions import FunctionFactory
from ..geni.version import Registry as APIRegistry
from .hostdriver import HostDriver
from ..types.UUID import UUID
from . import ports as Ports
from ..openflow.types import DPID
from .vol_manager import CentralVolumeManager as CVM

MAXROUNDS = 500 # Number of attempts a registry will make to try to find an unused value

class UnknownPortError(CoreException):
  def __init__ (self, client_id):
    super(UnknownPortError, self).__init__()
    self.cid = client_id
  def __str__ (self):
    return "Port (%s) not found in sliver." % (self.cid)

class InsufficientLocalCircuitsError(CoreException):
  def __str__ (self):
    return "Not enough local circuits available to satisfy the request."

class IllegalLocalTunnelError(CoreException):
  def __str__ (self):
    return "Attempt to create WAN tunnel with two local endpoints"

class KeyInUseError(CoreException):
  def __init__ (self, endpoint, key):
    super(KeyInUseError, self).__init__()
    self.endpoint = endpoint
    self.key = key
  def __str__ (self):
    return "Key %d already in use for endpoint %s" % (self.key, self.endpoint)

class IllegalLocalKeyError(CoreException):
  def __init__ (self, endpoint, key):
    super(IllegalLocalKeyError, self).__init__()
    self.endpoint = endpoint
    self.key = key
  def __str__ (self):
    return "Key %d is not valid for endpoint %s" % (self.key, self.endpoint)

class DuplicateDatapathNameError(CoreException):
  def __init__ (self, name):
    super(DuplicateDatapathNameError, self).__init__()
    self.name = name
  def __str__ (self):
    return "Datapath name (%s) is already in use." % (self.name)

class DuplicatePatchNameError(CoreException):
  def __init__ (self, name):
    super(DuplicatePatchNameError, self).__init__()
    self.name = name
  def __str__ (self):
    return "Patch name (%s) is already in use." % (self.name)

class DuplicateMirrorNameError(CoreException):
  def __init__ (self, name):
    super(DuplicateMirrorNameError, self).__init__()
    self.name = name
  def __str__ (self):
    return "Mirror name (%s) is already in use." % (self.name)

class DuplicateUsernameError(CoreException):
  def __init__ (self, name):
    super(DuplicateUsernameError, self).__init__()
    self.uname = name
  def __str__ (self):
    return "Username (%s) is already in use." % (self.uname)

class DuplicateDPIDError(CoreException):
  def __init__ (self, dpid):
    super(DuplicateDPIDError, self).__init__()
    self.dpid = dpid
  def __str__ (self):
    return "Datapath ID (%s) is already in use." % (self.dpid)

class UnknownResourceError(CoreException):
  def __init__ (self, clientid, expected = True):
    super(UnknownResourceError, self).__init__()
    self.client_id = clientid
    self._unexpected = not expected
  def __str__ (self):
    return "Resource (%s) is unknown." % (self.client_id)

class UnknownDatapathError(UnknownResourceError):
  def __str__ (self):
    return "Datapath name (%s) is not present in this slice." % (self.client_id)

class UnknownContainerError(UnknownResourceError):
  def __str__ (self):
    return "Container name (%s) is not present in this slice." % (self.client_id)

class UnknownInstanceError(UnknownResourceError):
  def __init__ (self, clientid, image):
    super(UnknownInstanceError, self).__init__(clientid)
    self.image = image
  def __str__ (self):
    return "No instance of (%s) found with name (%s)" % (self.image, self.client_id)

class UnknownImageAttributeError(CoreException):
  def __init__ (self, attrname, client_id = None):
    super(UnknownImageAttributeError, self).__init__()
    self.client_id = client_id
    self.attr = attrname
  def __str__ (self):
    return "Trying to set unknown attribute (%s) for image bound to client-id (%s)" % (self.attr, self.client_id)

class RequiredAttributeError(CoreException):
  def __init__ (self, attrname, client_id):
    super(RequiredAttributeError, self).__init__()
    self.client_id = client_id
    self.attr = attrname
  def __str__ (self):
    return "Attribute (%s) for image bound to client-id (%s) is required, but unspecified" % (self.attr, self.client_id)


class _DatapathNameRegistry(object):
  def __init__ (self):
    self._names = set()

  def insertName (self, name):
    # Only for loading slivers from the database to seed the registry
    # use .getName() to get new names
    if name in self._names:
      raise DuplicateDatapathNameError(name)
    self._names.add(name)

  def getName (self):
    while True:
      name = str(int(hashlib.sha1(str(UUID())).hexdigest(), 16) % (10**13))
      if name not in self._names:
        self.insertName(name)
        return name

  def removeName (self, name):
    self._names.discard(name)


class _DPIDRegistry(object):
  def __init__ (self):
    self._dpids = set()

  def insertDPID (self, dpid):
    if isinstance(dpid, DPID):
      dpid = dpid.int
    if dpid in self._dpids:
      raise DuplicateDPIDError(dpid)
    self._dpids.add(dpid)

  def getDPID (self):
    while True:
      dpid = hex(uuid.uuid4().int & (1<<64)-1)[2:-1]
      if dpid not in self._dpids:
        self.insertDPID(dpid)
        return DPID(dpid)

  def removeDPID (self, dpid):
    self._dpids.discard(dpid)


class _UsernameRegistry(object):
  def __init__ (self):
    self._unames = set()

  def insertUsername (self, uname):
    if uname in self._unames:
      raise DuplicateUsernameError(uname)
    self._unames.add(uname)

  def removeUsername (self, uname):
    self._unames.discard(uname)

  def getUsername (self):
    while True:
      uname = random.choice(string.ascii_lowercase) + ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(10))
      if uname not in self._unames:
        self.insertUsername(uname)
        return uname


class _PatchNameRegistry(object):
  def __init__ (self):
    self._names = set()

  def insertName (self, name):
    # Only for loading slivers from the database to seed the registry
    # use .getName() to get new names
    if name in self._names:
      raise DuplicatePatchNameError(name)
    self._names.add(name)

  def getName (self):
    while True:
      name = "v%s" % (str(int(hashlib.sha1(str(UUID())).hexdigest(), 16) % (10**13)))
      if name not in self._names:
        self.insertName(name)
        return name

  def removeName (self, name):
    self._names.discard(name)


class _MirrorNameRegistry(object):
  # NOTE: We believe that the names of mirrors and sflow and netflow instances
  # in OVS are in separate namespaces (as they go in different DB tables).  If
  # this turns out not to be true we'll have to change them to be a single registry.
  def __init__ (self):
    self._names = set()

  def insertName (self, name):
    # Only for loading slivers from the database to seed the registry
    # use .getName() to get new names
    if name in self._names:
      raise DuplicateMirrorNameError(name)
    self._names.add(name)
  
  def getName (self):
    for i in xrange(MAXROUNDS):
      name = "mr%06d" % (random.randint(0,999999))
      if name not in self._names:
        self.insertName(name)
        return name

  def removeName (self, name):
    self._names.discard(name)


class _SFlowIndexRegistry(object):
  # TODO: This is...erm...horrible?
  # We don't keep track of the used one, just the highest value
  # This is mostly ok, unless one sliver lives long enough for
  # all values to roll over, and then we won't catch a confict
  def __init__ (self):
    self._counters = {}

  def insertCounter (self, ctr):
    key = ctr[0]
    idx = int(ctr[1:])
    try:
      val = self._counters[key]
      if idx > val:
        self._counters[key] = idx
    except KeyError:
      self._counters[key] = idx

  def getIndex (self):
    # Create indices in the form of [a-z][0-999]
    for xchr in xrange(97,123):
      try:
        val = self._counters[chr(xchr)]
        if val < 999:
          idx = val + 1
          self._counters[chr(xchr)] = idx
          return "%s%d" % (chr(xchr), idx)
      except KeyError:
        idx = 0
        self._counters[chr(xchr)] = idx
        return "%s%d" % (chr(xchr), idx)


class CircuitPlanes(object):
  # TODO: We need some better way to maintain keys for various tunnel types
  # IF the plane doesn't support GRE/VXLAN/whatever, we shouldn't keep a list
  # of allocated keys for those tunnel types (or maybe we should, who knows)
  class IPv4(object):
    TYPE = "ipv4"
    VALFIELDS = [("label", (unicode, str)),
                 ("endpoint", (unicode, str)),
                 ("encoded", (bool)),
                 ("supported-types", (list)),
                 ("mtu", (int))]

    def __init__ (self):
      self.label = None
      self.endpoint = None
      self.encoded = True
      self.supported_types = set()
      self.mtu = 1500
      # {endpoint_int : set(keys)}
      self._grekeys = {}

    def addTunnelType (self, typ):
      self.supported_types.add(typ)

    def allocateGREKey (self, remote_endpoint):
      rei = remote_endpoint.int()
      lei = self.endpoint.int()
      if (rei == lei):
        raise IllegalLocalTunnelError()
      # TODO: We need to lock this thing probably
      allocated = self._grekeys.setdefault(rei, set())
      val = random.randrange(1, 2**31, 2)
      if rei < lei:
        while (val % 2) or (val in allocated):
          val = random.randrange(1, 2**31, 1)
      else:
        while val in allocated:
          val = random.randrange(1, 2**31, 2)
      self._grekeys.setdefault(rei, set()).add(val)
      return val

    def reserveGREKey (self, remote_endpoint, key):
      rei = remote_endpoint.int()
      lei = self.endpoint.int()
      if (rei < lei) and ((key % 2) == 0): # These should be odd
        raise IllegalLocalKeyError(remote_endpoint, key)
      elif (rei > lei) and not ((key % 2) == 0): # These should be even
        raise IllegalLocalKeyError(remote_endpoint, key)

      if key in self._grekeys.setdefault(rei, set()):
        raise KeyInUseError(remote_endpoint, key)
      self._grekeys.setdefault(rei, set()).add(key)

    def releaseGREKey (self, remote_endpoint, key):
      self._grekeys.setdefault(remote_endpoint.int(), set()).discard(key)

    def __json__ (self):
      return {"type" : "ipv4",
              "label" : self.label,
              "endpoint" : str(self.endpoint),
              "encoded" : self.encoded,
              "supported-types" : list(self.supported_types),
              "mtu" : self.mtu,
              "gre-keys" : self._grekeys}

    @staticmethod
    def jvalidate (obj):
      if not isinstance(obj, dict):
        return False
      if len(obj) < 5:
        return False
      return True

    @staticmethod
    def jconstruct (obj):
      cp = CircuitPlanes.IPv4()
      cp.label = obj["label"]
      cp.endpoint = IPy.IP(obj["endpoint"])
      cp.encoded = obj["encoded"]
      cp.supported_types = set(obj["supported-types"])
      cp.mtu = obj["mtu"]
      try:
        for k,v in obj["gre-keys"]:
          cp._grekeys[k] = set(v)
      except KeyError:
        pass # Initial creation
      return cp


class XRangeProxy(object):
  def __init__ (self, container):
    self._data = container
  def __iter__ (self):
    for obj in self._data:
      yield obj
  def __len__ (self):
    return len(self._data)
  def __getitem__ (self):
    return self._data[idx]


class SFlowInfo(object):
  def __init__ (self):
    self.agent_intf = "eth0"  #TODO: needs to be pulled from the host config for control interface
    self.collector_ip = None
    self.collector_port = 6343
    self.header_bytes = 128
    self.sampling_n = 64
    self.polling_secs = 5
    self.bridge_id = None
    self.idx = SFIR.getIndex()

  def __getitem__ (self, attr):
    return getattr(self, attr)

  def __json__ (self):
    return {"agent-intf" : self.agent_intf,
            "collector-ip" : self.collector_ip,
            "collector-port" : self.collector_port,
            "header-bytes" : self.header_bytes,
            "sampling-n" : self.sampling_n,
            "polling-secs" : self.polling_secs,
            "bridge-id" : self.bridge_id,
            "idx" : self.idx}

  @staticmethod
  def jconstruct (obj):
    s = SFlowInfo()
    s.agent_intf = obj["agent-intf"]
    s.collector_ip = obj["collector-ip"]
    s.collector_port = obj["collector-port"]
    s.header_bytes = obj["header-bytes"]
    s.sampling_n = obj["sampling-n"]
    s.polling_secs = obj["polling-secs"]
    s.bridge_id = obj["bridge-id"]
    s.idx = obj["idx"]
    SFIR.insertCounter(s.idx)
    return s


class NetFlowInfo(object):
  def __init__ (self):
    self._idx = None
    self.collector_ip = None
    self.collector_port = 6343
    self.timeout = 20
    self.bridge_id = None
    
  @property
  def idx (self):
    if not self._idx:
      self._idx = SFIR.getIndex()
    return self._idx

  def __getitem__ (self, attr):
    return getattr(self, attr)

  def __json__ (self):
    return {"collector-ip" : self.collector_ip,
            "collector-port" : self.collector_port,
            "bridge-id" : self.bridge_id,
            "timeout" : self.timeout,
            "idx" : self.idx}

  @staticmethod
  def jconstruct (obj):
    s = NetFlowInfo()
    s.collector_ip = obj["collector-ip"]
    s.collector_port = obj["collector-port"]
    s.bridge_id = obj["bridge-id"]
    s.timeout = obj["timeout"]
    s.idx = obj["idx"]
    SFIR.insertCounter(s.idx)
    return s


# TODO: Target bridges need a type, but right now we encapsulate all OVS here
class TargetBridge(object):
  def __init__ (self, brname):
    self.id = brname
    self.mirror = None

  def setOFPortNums (self):
    return

  def generatePortClientID (self):
    return "Nope"

  def addPort (self, port):
    return

  def connectWire (self, rport, log, sname = None):
    hd = HostDriver(log.info)

    newp = Ports.Internal(self)
    newp.client_id = self.generatePortClientID()
    rport.target = newp.client_id

    if rport.vlan:
      newp.vlan = rport.vlan

    newp.target = rport.client_id
    newp.setup(DB)
    newp.provision(hd, log)
    self.addPort(newp)

    hd.createWire(rport.name, newp.name)
    rport.attach()
    newp.attach()


class Datapath(object):
  def __init__ (self, dpname = None):
    if dpname is None:
      self.id = DPNR.getName()
    else:
      self.id = dpname
      DPNR.insertName(self.id)
    self.client_id = None
    self._sliver = None
    self.ports = {}
    self.controllers = []
    self.image = None
    self.sflow_info = None
    self.netflow_info = None
    self.ofver = "1.0"
    self._dpid = None
    self.mirror = None # or (port_client_id, mirror_name)
    self.stpinfo = OVS.STPParams()
    self.autoidx = 0
    self.mactable_size = None
    self.mactable_age = None
    self.public_desc = None

  @property
  def sliver (self):
    return self._sliver

  def __json__ (self):
    d = {"client_id" : self.client_id,
         "image" : self.image,
         "id" : self.id,
         "ofver" : self.ofver,
         "version" : 9,
         "controllers" : self.controllers,
         "ports" : [x.__json__() for x in self.ports.itervalues()],
         "public-desc" : self.public_desc}
    if self.sflow_info:
      d["sflow-info"] = self.sflow_info.__json__()
    if self._dpid:
      d["dpid"] = self._dpid
    d["mirror"] = self.mirror
    d["stpinfo"] = self.stpinfo.__json__()
    d["autoidx"] = self.autoidx
    d["mactable-size"] = self.mactable_size
    d["mactable-age"] = self.mactable_age
    return d

  @staticmethod
  def jconstruct (obj, sliver):
    dp = Datapath(obj["id"])
    dp._sliver = sliver
    dp.client_id = obj["client_id"]
    dp.image = obj["image"]
    for pobj in obj["ports"]:
      port = Ports._PortBase.jconstruct(pobj, dp)
      dp.ports[port.client_id] = port
    if json.getdefault(obj, "version", 1) >= 2:
      dp.controllers = obj["controllers"]
    sfinfo = json.getdefault(obj, "sflow-info", None)
    if sfinfo:
      dp.sflow_info = SFlowInfo.jconstruct(sfinfo)
    if json.getdefault(obj, "version", 1) >= 3:
      dp.ofver = obj["ofver"]
    if json.getdefault(obj, "version", 1) >= 4:
      if obj.has_key("dpid"):
        dp.dpid = obj["dpid"]
        try:
          DPIDR.insertDPID(dp.dpid)
        except DuplicateDPIDError:
          pass  # We only really care when we're making new DPIDs
    if json.getdefault(obj, "version", 1) >= 5:
      dp.mirror = obj["mirror"]
      if dp.mirror:
        DB.MNR.insertName(dp.mirror[1])
    if json.getdefault(obj, "version", 1) >= 6:
      if obj["stpinfo"]["_type"] == "stp":
        dp.stpinfo = OVS.STPParams.jconstruct(obj["stpinfo"])
      elif obj["stpinfo"]["_type"] == "rstp":
        dp.stpinfo = OVS.RSTPParams.jconstruct(obj["stpinfo"])
      elif obj["stpinfo"]["_type"] == "disabled":
        dp.stpinfo = OVS.DisabledSTP.jconstruct(obj["stpinfo"])
    if json.getdefault(obj, "version", 1) >= 7:
      dp.autoidx = obj["autoidx"]
    if json.getdefault(obj, "version", 1) >= 8:
      dp.mactable_size = obj["mactable-size"]
      dp.mactable_age = obj["mactable-age"]
    if json.getdefault(obj, "version", 1) >= 9:
      dp.public_desc = obj["public-desc"]
    return dp

  @property
  def dpid (self):
    if not self._dpid:
      self._dpid = DPIDR.getDPID()
    return self._dpid

  @dpid.setter
  def dpid (self, val):
    if not isinstance(val, DPID):
      val = DPID(val)
    self._dpid = val

  def setOFPortNums (self):
    import foam.vts.drivers.ovsdb as ovs
    driver = ovs.InfoDriver
    driver.log = DB._log
    # TODO: Do this in one transaction
    for port in self.ports.itervalues():
      port._ofportnum = driver.getOFPortNum(port.name)

  def addPort (self, port):
    self.ports[port.client_id] = port

  def getPort (self, client_id):
    return self.ports[client_id]

  def generatePortClientID (self):
    clid = "%s:_x_:%d" % (self.client_id, self.autoidx)
    self.autoidx += 1
    return clid

  # NOT USED WITHIN SLIVERS.  FOR NOW.
  def connectWire (self, rport, log, sname = None):
    hd = HostDriver(log.info)

    newp = Ports.Internal(self)
    newp.client_id = self.generatePortClientID()
    rport.target = newp.client_id

    if rport.vlan:
      newp.vlan = rport.vlan

    newp.target = rport.client_id
    if sname:
      newp.target = "%s.%s" % (sname, rport.client_id)
    newp.setup(DB)
    newp.provision(hd, log)
    self.addPort(newp)

    hd.createWire(rport.name, newp.name)
    rport.attach()
    newp.attach()

  def provision (self, driver, log):
    # TODO: Get this from a registry based on image name
    #driver = HostDriver(log.info)
    driver.createBridge(self)

    # NOTE: This just creates the ports - wiring happens in the main sliver provision call
    for port in self.ports.itervalues():
      port.setup(DB)
      port.provision(driver, log)

    if self.controllers:
      driver.setOFController(self, self.controllers[0])
    if self.sflow_info:
      driver.setSFlow(self.id, self.sflow_info)
    if self.netflow_info:
      driver.setNetFlow(self.id, self.netflow_info)

  def postLocalTopo (self, driver, log):
    if self.mirror:
      port = self.ports[self.mirror[0]]
      driver.attachMirrorPort(self.id, port.name, self.mirror[1])
      port.up = True

  def deprovision (self, driver, db):
    for port in self.ports.itervalues():
      port.deprovision(driver, db)
    db.DPNR.removeName(self.id)


class Container(object):
  def __init__ (self, sliver, cname = None):
    if cname is None:
      self.id = DPNR.getName()
    else:
      self.id = cname
      DPNR.insertName(self.id)

    self._sliver = sliver
    self._ssh_keys = []
    self._cleanSliverKeys()

    self._driver = DockerDriver(self._sliver.log.info)

    self.client_id = None
    self.ports = {}
    self._image = None
    self._image_desc = None
    self.__dead_image_name = None
    self._instance_data = None
    self._shadow_instance_data = None
    self._console_uname = None
    self._instance_handler = None
    self._ram = None
    self._store_root = None
    self._log = None

    # NOTE: I do not like this, but hgvolumes are a core feature (and probably shouldn't be?)
    # I'm not sure how to design a "mix-in" handler for containers, where something
    # like Hg volume support would be bolted onto the side of any container, although they
    # should also be able to have image handlers that don't know anything about special volumes
    self._hgvols = []
    self._dbxvols = []

    self._pending_routes = []

    # Ephemeral jank that Past Nick screwed Future Nick with
    self._vols = []
    self._prebuilt = False

  @property
  def log (self):
    if self._log is None:
      self._log = KeyAdapter(self.client_id, self._sliver.log)
    return self._log

  @property
  def imagename (self):
    if self._image_desc:
      return self._image_desc.name
    else:
      return self.__dead_image_name

  def __json__ (self):
    d = {"client_id" : self.client_id,
         "base-image" : self.imagename,
         "used-image" : self._image,
         "id" : self.id,
         "version" : 8,
         "console-username" : self._console_uname,
         "handler-data" : self._instance_handler,
         "ports" : [x.__json__() for x in self.ports.itervalues()],
         "ssh-keys" : self._ssh_keys,
         "hgvols" : self._hgvols,
         "dbxvols" : self._dbxvols,
         "ram" : self._ram}
    return d

  @staticmethod
  def jconstruct (obj, sliver):
    ct = Container(sliver, obj["id"])
    ct.client_id = obj["client_id"]
    try:
      ct._image_desc = CM.getImageDesc(obj["base-image"])
    except CM.UnknownImagespecError:
      sliver.log.info("Loading container (%s) with dead image (%s)" % (ct.client_id, obj["base-image"]))
      ct._image_desc = None
      ct.__dead_image_name = obj["base-image"]
    for pobj in obj["ports"]:
      port = Ports._PortBase.jconstruct(pobj, ct)
      ct.ports[port.client_id] = port
    if obj["version"] >= 2:
      ct._console_uname = obj["console-username"]
      if ct._console_uname:
        UNR.insertUsername(ct._console_uname)
    if obj["version"] >= 3:
      if obj["handler-data"]:
        if ct._image_desc:
          ct._instance_handler = ct._image_desc._handler.jconstruct(ct._image_desc, obj["handler-data"])
          ct._instance_handler.container = ct
    if obj["version"] >= 4:
      ct._ssh_keys = obj["ssh-keys"]
    if obj["version"] >= 5:
      ct._image = obj["used-image"]
    if obj["version"] >= 6:
      ct._hgvols = [HGVolumeInfo.jconstruct(x) for x in obj["hgvols"]]
    if obj["version"] >= 7:
      ct._dbxvols = [DropboxVolumeInfo.jconstruct(x) for x in obj["dbxvols"]]
    if obj["version"] >= 8:
      ct._ram = obj["ram"]
    return ct

  @property
  def sshport (self):
    try:
      return int(self.instance_data[0]["NetworkSettings"]["Ports"]["22/tcp"][0]["HostPort"])
    except IndexError, e:
      DB._log.debug(str(self.instance_data))
      return -1

  @property
  def ssh_keys (self):
    for x in self._ssh_keys:
      yield x

  @property
  def pid (self):
    return int(self.instance_data[0]["State"]["Pid"])

  @property
  def image (self):
    return self._image

  @image.setter
  def image (self, val):
    kstr = "\\n".join(self._ssh_keys)
    self._image_desc = CM.getImageDesc(val)
    if self._image_desc._handler:
      self._instance_handler = self._image_desc._handler(self._image_desc)
      self._instance_handler.container = self
    self._image = CM.buildImageForSliver(self._sliver, val, {"pkey" : kstr})

  @property
  def ram (self):
    if not self._ram:
      if not self._image_desc:
        return 0  # We will not restart dead images
      return self._image_desc.mem_default
    return self._ram

  @ram.setter
  def ram (self, val):
    if val > self._image_desc.mem_max or val < self._image_desc.mem_min:
      raise CM.MemoryOutOfBoundsError(self.imagename, val, self._image_desc.mem_min, self._image_desc.mem_max)
    self._ram = val

  @property
  def longid (self):
    return self.instance_data[0]["Id"]

  @property
  def instance_data (self):
    if not self._instance_data:
      self._instance_data = self._driver.container_info(self.id)
      if self._image_desc.network_shadow:
        self._shadow_instance_data = self._driver.container_info(self.id, True)
      else:
        self._shadow_instance_data = self._instance_data

    return self._instance_data

  @property
  def hostname (self):
    return self.client_id

  def _cleanSliverKeys (self):
    ukeys = []
    for user in self._sliver.login_users:
      ukeys.extend(user["keys"])
    for key in ukeys:
      nk = key.strip()
      nk = nk.split("\n")[0]
      self._ssh_keys.append(nk)

  def setAttribute (self, name, val):
    if not self._instance_handler:
      return # TODO: Raise something
    DB._log.debug("Setting container attribute %s to %s" % (name, val))
    if not self._image_desc:
      return # TODO: Log/raise something - using dead image
    try:
      adesc = self._image_desc.image_attributes[name]
    except KeyError:
      raise UnknownImageAttributeError(name, self.client_id)
    self._instance_handler.attrs[name] = adesc.builder(val)

  def validateAttributes (self):
    # Make sure required attributes are actually here
    for adesc in self._image_desc.image_attributes.values():
      if not adesc.optional:
        try:
          attr = self._instance_handler.attrs[adesc.name]
        except KeyError:
          raise RequiredAttributeError(adesc.name, self.client_id)

  def addPort (self, port):
    self.ports[port.client_id] = port

  def getPort (self, client_id):
    try:
      return self.ports[client_id]
    except KeyError:
      DB._log.debug("No port [%s].  Known: %s" % (client_id, str(self.ports)))
      raise

  def addSSHKeys (self, key_list):
    if not self._image_desc:
      return # Dead image
    self._ssh_keys.extend(key_list)
    hd = HostDriver(DB._log.info)
    if not self._image_desc.console_shell:
      return # TODO: Maybe be like wtf dude, no keys for this container.  NICE TRY.
    if self._image_desc.forward_ssh_cmd:
      hd.addUserSSHKeys(self._console_uname, key_list,
                       "sudo /usr/bin/docker exec -it %s %s \\\"${SSH_ORIGINAL_COMMAND}\\\"" % (
                                                      self.id, self._image_desc.console_shell))
    else:
      hd.addUserSSHKeys(self._console_uname, key_list,
                       "sudo /usr/bin/docker exec -it %s %s" % (self.id, self._image_desc.console_shell))

  def addIPRoute (self, network, gateway):
    self._driver.addIPRoute(self, network, gateway)
      
  def prebuild (self):
    if self._instance_handler and not self._prebuilt:
      self._instance_handler.prebuild(self)
      self._prebuilt = True

  def provision (self, sliver, log):
    hd = HostDriver(log.info)
    vols = []
    vols.extend([x.hostvol for x in self._hgvols])
    vols.extend([x.hostvol for x in self._dbxvols])

    self._vols = vols

    for vinfo in self._hgvols:
      source = vinfo.getRequestedSource()
      DB.cloneVolume(source)

    self._shadow_instance_data = self._driver.run_container(self, vols)
    self._instance_data = self._shadow_instance_data

    for idx,port in enumerate(self.ports.itervalues()):
      port.setup(DB)
      port.provision(self._driver, log)
      port.alias = "eth%d" % (idx+1)

    if self._image_desc.console_shell:
      self._console_uname = UNR.getUsername()
      if self._image_desc.forward_ssh_cmd:
        hd.addUser(self._console_uname, ["rdocker"], self._ssh_keys,
                   "sudo /usr/bin/docker exec -it %s %s \\\"${SSH_ORIGINAL_COMMAND}\\\"" % (self.id,
                   self._image_desc.console_shell))
      else:
        hd.addUser(self._console_uname, ["rdocker"], self._ssh_keys,
                   "sudo /usr/bin/docker exec -it %s %s" % (self.id, self._image_desc.console_shell))
    for vinfo in self._hgvols:
      DB.checkAndWait(vinfo.getRequestedSource())
      vinfo.clone()

  def deprovision (self, driver, db):
    for port in self.ports.itervalues():
      port.deprovision(driver, db)
    db.DPNR.removeName(self.id)
    if self._console_uname:
      hd = HostDriver(DB._log.info)
      hd.delUser(self._console_uname)
      db.UNR.removeUsername(self._console_uname)
    driver.destroyContainer(self.id, self)

  def postLocalTopo (self):
    if self._image_desc.network_shadow:
      # TODO: We should not know the ID value via the raw docker json
      self._instance_data = self._driver.run_container(self, self._vols,
                                                       netctid = self._shadow_instance_data[0]["Id"])

    if self._instance_handler:
      self._instance_handler.postlocaltopo(self)

  def postNetTopo (self):
    if self._instance_handler:
      self._instance_handler.postnettopo(self)


class VTSSliver(Allocation):
  EXPFORMAT = "%Y%m%d-%H%M"

  def __init__ (self, slice_urn, user_urn):
    super(VTSSliver, self).__init__(self)
    cmtag = ConfigDB.getConfigItemByKey("vts.geni.cm-tag").getValue()
    self.urn = "urn:publicid:IDN+%s+sliver+%s" % (cmtag, self.getUUID())
    self._login_users = []
    self.slice_urn = slice_urn
    self.user_urn = user_urn
    self.create_ts = datetime.datetime.now()
    self.expiration = self.create_ts + datetime.timedelta(days=5)
    self.datapaths = []
    self.functions = []
    self.containers = []
    self.rspec = None
    self._patches = []
    self._ports_by_cid = {}
    self._log = None
    self.gapi_lock = policy.GAPIActionPolicy()
    self._provisioned = False
    self._topo_name = None

  @property
  def shortslice (self):
    return self.slice_urn.split("+")[-1]

  @property
  def log (self):
    if not self._log:
      self._log = KeyAdapter(self.urn, DB._log)
    return self._log

  @property
  def uuid (self):
    return self.urn.rsplit("+", 1)[1]

  def __json__ (self):
    d = {}
    d["version"] = 6
    d["urn"] = self.urn
    d["slice_urn"] = self.slice_urn
    d["user_urn"] = self.user_urn
    d["datapaths"] = [x.__json__() for x in self.datapaths]
    d["functions"] = [x.__json__() for x in self.functions]
    d["containers"] = [x.__json__() for x in self.containers]
    d["patches"] = self._patches
    d["rspec"] = self.rspec
    d["expiration"] = self.expiration.strftime(VTSSliver.EXPFORMAT)
    d["create-time"] = self.create_ts.strftime(VTSSliver.EXPFORMAT)
    d["provisioned"] = self._provisioned
    d["topo-name"] = self._topo_name
    return d

  @staticmethod
  def jconstruct (obj):
    DB._log.log(5, "SLIVER DATA: %s" % (json.dumps(obj)))

    from ..geni.resources import Registry as ResourceRegistry
    s = VTSSliver(obj["slice_urn"], obj["user_urn"])
    s.urn = obj["urn"]
    for dpobj in obj["datapaths"]:
      s.addDatapath(Datapath.jconstruct(dpobj, s))
    s._patches = obj["patches"]
    for (br, port) in s._patches:
      PNR.insertName(port)
    flist = json.getdefault(obj, "functions", [])
    for func in flist:
      s.functions.append(FunctionFactory(func))
    # TODO: Get load orders right - don't do this here, do it in the GENI
    # module load

    # Pre-v5 slivers are always provisioned if we're loading them
    s._provisioned = True

    ResourceRegistry.registerSliver(s.slice_urn, s.urn)
    version = json.getdefault(obj, "version", 1)
    if version == 1:
      try:
        s.rspec = obj["rspec"]
      except KeyError:
        pass
      try:
        s.expiration = datetime.datetime.strptime(obj["expiration"], VTSSliver.EXPFORMAT)
      except KeyError:
        pass
    if version >= 2:
      s.rspec = obj["rspec"]
      for cobj in obj["containers"]:
        s.addContainer(Container.jconstruct(cobj, s))
    if version >= 3:
      s.expiration = datetime.datetime.strptime(obj["expiration"], VTSSliver.EXPFORMAT)
      s.create_ts = datetime.datetime.strptime(obj["create-time"], VTSSliver.EXPFORMAT)
    if version >= 4:
      # REMOVED urn-lock
      pass
    if version >= 5:
      s._provisioned = obj["provisioned"]
    if version >= 6:
      s._topo_name = obj["topo-name"]
      
    return s

  @property
  def expstr (self):
    return self.expiration.strftime(VTSSliver.EXPFORMAT)

  @property
  def login_users (self):
    return self._login_users

  @login_users.setter
  def login_users (self, val):
    self._login_users = val
    for ct in self.containers:
      ct._cleanSliverKeys()

  def _rebuildPortsByCID (self):
    self._ports_by_cid = {}
    for dp in self.datapaths:
      for port in dp.ports.itervalues():
        self._ports_by_cid[port.client_id] = (port, dp)
    for ct in self.containers:
      for port in ct.ports.itervalues():
        self._ports_by_cid[port.client_id] = (port, ct)
      
  def addDatapath (self, dp):
    dp._sliver = self
    self.datapaths.append(dp)
    for port in dp.ports.itervalues():
      self._ports_by_cid[port.client_id] = (port, dp)

  def getDatapath (self, client_id):
    for dp in self.datapaths:
      if dp.client_id == client_id:
        return dp
    # TODO: Raise exception
    return None

  def getFunction (self, client_id):
    for func in self.functions:
      if func.client_id == client_id:
        return func
    return None

  def addContainer (self, ct):
    self.containers.append(ct)
    for port in ct.ports.itervalues():
      self._ports_by_cid[port.client_id] = (port, ct)

  def getContainer (self, client_id):
    for ct in self.containers:
      if ct.client_id == client_id:
        return ct
    # TODO: Raise Exception
    return None
      
  @property
  def client_ids (self):
    cids = set()
    for c in self.containers:
      cids.add(c.client_id)
    for dp in self.datapaths:
      cids.add(dp.client_id)
    for f in self.functions:
      cids.add(f.client_id)
    return cids

  def getPortByCID (self, cid):
    try:
      (port, dp) = self._ports_by_cid[cid]
      return port
    except KeyError:
      raise UnknownPortError(cid)

  def addFunction (self, vf):
    vf.sliver = self
    self.functions.append(vf)

  def removeFunction (self, vf):
    vf.sliver = None
    self.functions.remove(vf)

  def portUp (self, client_id):
    port = self.getPortByCID(client_id)
    port.up = True

  def portDown (self, client_id):
    port = self.getPortByCID(client_id)
    port.up = False

  def setController (self, dp_cid, controller_url, ofver):
    hd = HostDriver(self.log.info)
    for dp in self.datapaths:
      if dp.client_id == dp_cid:
        if ofver:
          dp.ofver = ofver
        dp.controllers = []
        dp.controllers.append(controller_url)
        hd = HostDriver(self.log.info)
        hd.setOFController(dp, controller_url)
        return
    # TODO: Raise some kind of exception that we didn't find your client_id
        

  def addDatapathPort (self, dpname, port):
    # TODO: If this creates a patch we have to fix it up
    hd = HostDriver(self.log.info)
    for dp in self.datapaths:
      if dp.id == dpname:
        dp.ports[port.client_id] = port
        port.datapath = dp
        port.setup(DB)
        port.provision(hd, log)
    DB._write_slivers()

  def removeDatapathPort (self, dpclid, client_id):
    # NOTE: This is somewhat intentionally inefficient - the fact that the current
    # host driver can delete it directly doesn't mean that will always be the case
    log = KeyAdapter(self.urn, DB._log)

    # TODO: Get a real driver - this needs to know which one to use
    hd = HostDriver(log.info)
    dp = self.getDatapath(dpclid)
    if dp:
      port = dp.ports[client_id]
      hd.detachPort(dp.id, port.name)
      port.deprovision(hd, DB)
      del dp.ports[client_id]
    DB._write_slivers()

  def provision (self, bundle = None):
    if self._provisioned:
      return

    log = KeyAdapter(self.urn, DB._log)

    # TODO: This has to move into the datapath
    # It may make patching more complex
    hd = HostDriver(log.info)

    for vf in self.functions:
      vf.provision(DB, log)

    for dp in self.datapaths:
      dp.provision(hd, log)

    for ct in self.containers:
      ct.provision(self, log)

    # Find local patches and make them
    # TODO: There's a local port-by-cid map, maybe we can use that?
    patch_map = {}
    for vf in self.functions:
      patch_map[vf.client_id] = vf

    for dp in self.datapaths:
      for port in dp.ports.itervalues():
        if port.needs_patch:
          if port.target in patch_map:
            rp = patch_map[port.target]
            hd.createWire(port.name, rp.name)
            rp.attach()
            port.attach()
#            hd.patch(port.datapath.id, port.name, rp.datapath.id, rp.name)
            del patch_map[port.target]
          else:
            patch_map[port.client_id] = port
        if port.needs_vf_attach:
          # TODO: We can't attach to the same VF twice on the same dp, right now
          vfp = patch_map[port.target]
          port.name = vfp.port_name
          hd.attachPort(dp.id, port)
          del patch_map[port.target]

    for ct in self.containers:
      for port in ct.ports.itervalues():
        if port.needs_patch:
          if port.target in patch_map:
            rp = patch_map[port.target]
            hd.createWire(port.name, rp.name)
            rp.attach()
            port.attach()
            del patch_map[port.target]
          else:
            patch_map[port.client_id] = port

    # Anything unpatched is hopefully in the bundle
    if bundle and patch_map:
      for port in patch_map.values():
        dp = bundle.getDatapath(port.target)
        if not dp:
          if port.target in DB.target_bridges:
            dp = TargetBridge(DB.target_bridges[port.target])
          elif port.target in DB.public_bridges:
            dp = DB.public_bridges[port.target]
          else:
            # TODO: Complain?
            continue
          dp.connectWire(port, log, self.shortslice)
        else:
          # This DP is in our slice, so don't screw with the client-id
          dp.connectWire(port, log)

        del patch_map[port.client_id]
        
    # TODO: Set policy on what to do for updated dps in other slivers
    for dp in self.datapaths:
      dp.postLocalTopo(hd, log)

    for ct in self.containers:
      ct.postLocalTopo()

    # Handle post-local dynamic configuration
    for ct in self.containers:
      for port in ct.ports.itervalues():
        if port._pending_addresses:
          for item,vals in port._pending_addresses.iteritems():
            if item == "ipv4":
              for addr in vals:
                port.addIPv4Address(addr)
            elif item == "ipv6":
              continue
            elif item == "eth-mac":
              continue
      for network,gateway in ct._pending_routes:
        # TODO: Make route object that has metadata for each container to decide how to set it
        ct.addIPRoute(network, gateway)

    for vf in self.functions:
      vf.start(log)

    for ct in self.containers:
      ct.postNetTopo()

    StatDriver.addSliver(self)
    self._provisioned = True

    # Sliver state may have changed, write it back into the database
    DB._write_slivers()

  def deprovision (self, db):
    StatDriver.delSliver(self)

    HGVolumeManager.removeVolumes(self)
    DropboxVolumeManager.removeVolumes(self)

    hd = HostDriver(db._log.info)
    for dp in self.datapaths:
      dp.deprovision(hd, db)
      hd.destroyBridge(dp.id)

    dd = DockerDriver(db._log.info)
    for ct in self.containers:
      ct.deprovision(dd, db)

    for vf in self.functions:
      vf.deprovision(db, db._log)

    RawHostVolume.destroyForSliver(self)


class _DB(object):
  def __init__ (self):
    # TODO: All the registries should be a member of this object
    self._log = None
    self.enabled = False
    self._pgvlans = {}
    self._pgused = set()
    self._images = []
    self._wan_interfaces = None
    self._pgcmid = None

    # {type : { label : cp }}
    self._circuit_planes = {}

    self.target_bridges = {}
    self.public_bridges = {}

    # These should be weakrefs
    self._sliversByURN = {}
    self._sliversBySlice = {}

    # Primary owner of slivers
    self._slivers = []

    # Old module-level registries
    # TODO: We need to reload these from restart
    self.DPNR = _DatapathNameRegistry()
    self.PNR = _PatchNameRegistry()
    self.SFIR = _SFlowIndexRegistry()
    self.DPIDR = _DPIDRegistry()
    self.UNR = _UsernameRegistry()
    self.MNR = _MirrorNameRegistry()

    self.CVM = CVM()

  def setLogger (self, logger):
    self._log = KeyAdapter("VTS-DB", logger)
    self.CVM.setLog(logger)

  def _write_pgvlans (self):
    self._log.debug("Writing PG VLANs")
    ConfigDB.getConfigItemByKey("vts.pgvlans").write(self._pgvlans)

  def _write_pgused (self):
    self._log.debug("Writing used PG circuits")
    ConfigDB.getConfigItemByKey("vts.pgused").write(self._pgused)

  def _write_images (self):
    self._log.debug("Writing images")
    ConfigDB.getConfigItemByKey("vts.images").write(self._images)

  def _write_target_bridges (self):
    self._log.debug("Writing target bridges")
    ConfigDB.getConfigItemByKey("vts.target-bridges").write(self.target_bridges)

  def _write_slivers (self):
    l = []
    for sliver in self._slivers:
      l.append(json.fdumps(sliver))
    ConfigDB.getConfigItemByKey("vts.slivers").write(l)

  def _write_circuitplanes (self):
    l = []
    for typdict in self._circuit_planes.itervalues():
      for cp in typdict.itervalues():
        l.append(json.fdumps(cp))
    ConfigDB.getConfigItemByKey("vts.circuit-planes").write(l)

  def _load (self):
    self._pgvlans = ConfigDB.getConfigItemByKey("vts.pgvlans").getValue()
    self._pgused = ConfigDB.getConfigItemByKey("vts.pgused").getValue()
    self._images = ConfigDB.getConfigItemByKey("vts.images").getValue()
    self._pgcmid = ConfigDB.getConfigItemByKey("vts.pgcmid").getValue()
    self.enabled = ConfigDB.getConfigItemByKey("vts.status").getValue()
    self.target_bridges = ConfigDB.getConfigItemByKey("vts.target-bridges").getValue()

    self.CVM.loadVolumes() # should be before __loadSlivers() to create vols before slivers check for them

    self.__loadCircuitPlanes()
    self.__loadSlivers()

  def __loadSlivers (self):
    self._log.debug("Loading slivers")
    data = ConfigDB.getConfigItemByKey("vts.slivers").getValue()
    for item in data:
      try:
        sliver = VTSSliver.jconstruct(json.loads(item))
        StatDriver.addSliver(sliver)
        self._slivers.append(sliver)
      except Exception, e:
        self._log.exception("Exception Loading Sliver")
        continue

    for sliver in self._slivers:
      self._sliversByURN[sliver.urn] = sliver
      self._sliversBySlice.setdefault(sliver.slice_urn, []).append(sliver)

    self.refreshPublicBridges()

  def refreshPublicBridges (self):
    self.public_bridges = {}
    for sliver in self._slivers:
      for dp in sliver.datapaths:
        if dp.public_desc:
          self.public_bridges["%s.%s" % (sliver.uuid, dp.client_id)] = dp

  def __loadCircuitPlanes (self):
    self._log.debug("Loading circuit planes")
    data = ConfigDB.getConfigItemByKey("vts.circuit-planes").getValue()
    for istr in data:
      item = json.loads(istr)
      if item["type"] == "ipv4":
        cp = CircuitPlanes.IPv4.jconstruct(item)
        self.addCircuitPlane(cp)

  def getVlanID (self, circuit):
    return self._pgvlans[circuit]

  def getSliversBySlice (self, slice_urn):
    return self._sliversBySlice[slice_urn]

  def getSliver (self, sliver_urn):
    self._log.debug("Getting status for %s" % (sliver_urn))
    try:
      return self._sliversByURN[sliver_urn]
    except KeyError:
      self._log.info("Known sliver URNs: %s" % ", ".join(self._sliversByURN.iterkeys()))
      raise

  def deallocateSliver (self, sliver_urn):
    from ..geni.resources import Registry as ResourceRegistry

    sliver = self.getSliver(sliver_urn)
    if sliver._provisioned:
      return

    slice_slivers = self.getSliversBySlice(sliver.slice_urn)
    ResourceRegistry.unregisterSliver(sliver.slice_urn, sliver_urn)
    del self._sliversByURN[sliver_urn]
    slice_slivers.remove(sliver)
    self._slivers.remove(sliver)
    self._write_slivers()

  def addTargetBridge (self, uiname, brname):
    self.target_bridges[uiname] = brname
    self._write_target_bridges()

  def setPGComponentManager (self, cmid):
    self._pgcmid = cmid
    ConfigDB.getConfigItemByKey("vts.pgcmid").write(self._pgcmid)

  def reserveLocalCircuits (self, count):
    # TODO: Needs to be reservePGCircuits, or change what it does
    # TODO: We need a lock here, probably needs to be a queue
    avail = set(self.vlans)
    avail -= self._pgused
    if len(avail) < count:
      raise InsufficientLocalCircuitsError()
    circuits = [avail.pop() for x in xrange(count)]
    self.allocateLocalCircuit(circuits)
    return circuits

  def freeLocalCircuit (self, val):
    self._pgused.discard(val)
    self._write_pgused()

  def allocateLocalCircuit (self, val):
    if isinstance(val, int):
      s = set([val])
    elif isinstance(val, list):
      s = set(val)
    elif isinstance(val, set):
      s = val
    self._pgused |= s
    self._write_pgused()

  def addPGVLAN (self, name, vid):
    self._pgvlans[name] = vid
    self._write_pgvlans()

  def addImage (self, name):
    self._images.append(name)
    self._write_images()

  def sliceExists (self, slice_urn):
    return self._sliversBySlice.has_key(slice_urn)

  def insertSliver (self, sliver):
    self._log.debug("Inserting sliver %s into DB" % (sliver.urn))
    self._slivers.append(sliver)
    self._sliversByURN[sliver.urn] = sliver
    self._sliversBySlice.setdefault(sliver.slice_urn, []).append(sliver)
    self._write_slivers()

  def removeSliver (self, sliver):
    self._slivers.remove(sliver)
    del self._sliversByURN[sliver.urn]
    self._sliversBySlice[sliver.slice_urn].remove(sliver)
    self._write_slivers()
    self.refreshPublicBridges()

  def deleteSlice (self, slice_urn):
    slivers = self._sliversBySlice[slice_urn]
    for sliver in slivers:
      sliver.deprovision(self)
      self._slivers.remove(sliver)
      del self._sliversByURN[sliver.urn]
    del self._sliversBySlice[slice_urn]
    self._write_slivers()
    self.refreshPublicBridges()

  def renewSliver (self, sliver, exp):
    sliver.expiration = exp
    self._write_slivers()
    
  def addCircuitPlane (self, cp):
    self._circuit_planes.setdefault(cp.TYPE, {})[cp.label] = cp
    self._write_circuitplanes()

  def getCircuitPlane (self, label):
    # TODO: If we end up with lots of planes we might need a label index
    for tdict in self._circuit_planes.itervalues():
      if tdict.has_key(label):
        return tdict[label]

  def removeCircuitPlane (self, label):
    for tdict in self._circuit_planes.itervalues():
      if tdict.has_key(label):
        del tdict[label]
        return

  @property
  def circuit_planes (self):
    return self._circuit_planes

  @property
  def images (self):
    return self._images

  @property
  def vlans (self):
    return self._pgvlans.keys()

  @property
  def usedvlans (self):
    return self._pgused

  @property
  def slivers (self):
    return self._slivers

  @property
  def slices (self):
    return self._sliversBySlice.keys()

  def cloneVolume (self, source):
    self.CVM.cloneVolume(source)

  def getManagedSource (self, source):
    return self.CVM.getManagedSource(source)

  def checkAndWait (self, source):
    self.CVM.checkStatusAndWait(source)

DB = _DB()
DPNR = DB.DPNR
PNR = DB.PNR
SFIR = DB.SFIR
DPIDR = DB.DPIDR
UNR = DB.UNR

def setup (app):
  # Can't load sslvpn slivers unless this is first
  from .functions import sslvpn
  sslvpn.setup(app)

  DB.setLogger(app.logger)
  StatDriver.log = DB._log

  citems = []
  citems.append(ConfigItem().setKey("vts.status").setValue(False)
                .setDesc("VTS Status")
                .setUpdateFuncName("foam.core.configdb.coerceBool"))
  citems.append(ConfigItem().setKey("vts.slivers").setValue([]).setDesc("")
                .setUpdateFuncName("foam.core.configdb.checkList"))
  citems.append(ConfigItem().setKey("vts.pgvlans").setValue({}).setDesc("")
                .setUpdateFuncName("foam.core.configdb.checkDict"))
  citems.append(ConfigItem().setKey("vts.pgused").setValue(set()).setDesc("")
                .setUpdateFuncName("foam.core.configdb.checkSet"))
  citems.append(ConfigItem().setKey("vts.images").setValue([])
                .setDesc("Storage for supported VTS image descriptors")
                .setUpdateFuncName("foam.core.configdb.checkList"))
  citems.append(ConfigItem().setKey("vts.initial-lease").setValue(datetime.timedelta(hours=12))
                .setDesc("Default initial lease time for VTS slivers")
                .setUpdateFuncName("foam.core.configdb.checkDatetimeHours"))
  citems.append(ConfigItem().setKey("vts.target-bridges").setValue({})
                .setDesc("Storage for site-local VTS targets")
                .setUpdateFuncName("foam.core.configdb.checkDict"))
  citems.append(ConfigItem().setKey("vts.circuit-planes").setValue({})
                .setDesc("Storage for site-local circuit-plane metadata")
                .setUpdateFuncName("foam.core.configdb.checkList"))
  citems.append(ConfigItem().setKey("vts.max-initial-lease").setValue(datetime.timedelta(days=2))
                .setDesc("Max lease time")
                .setUpdateFuncName("foam.core.configdb.checkDatetimeHours"))
  citems.append(ConfigItem().setKey("vts.max-lease").setValue(datetime.timedelta(days=28))
                .setDesc("Max total lease time")
                .setUpdateFuncName("foam.core.configdb.checkDatetimeHours"))
  citems.append(ConfigItem().setKey("vts.pgcmid").setValue(None)
                .setDesc("Local component manager ID for PG/IG Aggregate"))
  [ConfigDB.installConfigItem(x) for x in citems]

  vinfo = {}
#  vinfo["vts_enable"] = ConfigDB.getConfigItemByKey("vts.status").getValue()
  vinfo["vts.enable"] = True
  vinfo["vts.version"] = "0.7"
  APIRegistry.addStaticInfo(vinfo)

  DB._load()

