# Copyright (c) 2015-2021  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import errno
import importlib
import logging
import logging.handlers
import os
import os.path
import sys

import IPy

from .exceptions import OutOfMountSpaceError
from .. import tasks
from ..hostdriver import call, HostDriver
from ...core.configdb import ConfigDB, ConfigItem
from ...core.log import KeyAdapter
from ...core.exception import CoreException
from ...core import json
from ...config import LOGDIR, LOGFORMAT

try:
  from ...config import DOCKERLOGLEVEL
except ImportError:
  DOCKERLOGLEVEL = logging.INFO

DKRLOG = logging.getLogger("driver.docker")
lh = logging.handlers.RotatingFileHandler(os.path.normpath("%s/docker.log" % (LOGDIR)),
                                           maxBytes = 1000000, backupCount = 10)
lh.setLevel(DOCKERLOGLEVEL)
lh.setFormatter(logging.Formatter(LOGFORMAT))
DKRLOG.addHandler(lh)
DKRLOG.setLevel(DOCKERLOGLEVEL)

class HostPathVolume(object):
  def __init__ (self, hpath, cpath):
    self.hpath = hpath
    self.cpath = cpath

  def __json__ (self):
    return {"hpath" : self.hpath, "cpath" : self.cpath}

  @property
  def args(self):
    return "-v %s:%s" % (self.hpath, self.cpath)

class SeededDataVolume(object):
  def __init__ (self, source_repo, container_path):
    self.cpath = container_path
    self.srcpath = "%s/%s" % (SEEDED_ROOT, source_repo)

class RawHostVolume(object):
  def __init__ (self, cpath):
    self.cpath = cpath

  def convertToHostVolume (self, cobj):
    from ..hostdriver import HostDriver
    hd = HostDriver(cobj.log.info)
    hpath = "%s/_raw-host/%s/%s%s" % (hd.volroot, cobj._sliver.uuid, cobj.client_id, self.cpath)
    try:
      os.makedirs(hpath)
      os.chmod(hpath, 0o777)
    except OSError as e:
      if e.errno == errno.ENOSPC:
        raise OutOfMountSpaceError("host-vol")
      raise

    return HostPathVolume(hpath, self.cpath)

  @staticmethod
  def destroyForSliver (sliver):
    from ..hostdriver import HostDriver
    hd = HostDriver(sliver.log.info)
    basepath = "%s/_raw-host/%s" % (hd.volroot, sliver.uuid)
    tasks.destroyTree(basepath)


class ValueOutOfRangeError(CoreException):
  def __init__ (self, val, mn, mx):
    super(ValueOutOfRangeError, self).__init__()
    self.val = val
    self.min = mn
    self.max = mx
  def __str__ (self):
    return "Supplied value (%s) out of range [%d, %d]" % (self.val, self.min, self.max)

def uint16Builder (val):
  val = int(val)
  if val < 0 or val > 65535:
    raise ValueOutOfRangeError(val, 0, 65535)
  return val

def ipv4Builder (val):
  if val:
    return IPy.IP(val)
  return None
 
def ipv4ListBuilder (val):
  parts = val.split(",")
  out = []
  for part in parts:
    out.append(ipv4Builder(part.strip()))
  return out

def boolBuilder (val):
  if isinstance(val, bool):
    return val
  elif isinstance(val, (str, unicode)):
    val = val.lower()
    if val == "false":
      return False
    elif val == "true":
      return True
    else:
      # Well at least we can debug it later...
      return None

class ImageDesc(object):
  class Attribute(object):
    TYPES = {"client-id" : str,
             "ip-subnet" : ipv4Builder,
             "ip-subnet-list" : ipv4ListBuilder,
             "ip-address" : ipv4Builder,
             "ip-address-list" : ipv4ListBuilder,
             "integer" : int,
             "port-client-id" : str,
             "string" : str,
             "bool" : boolBuilder,
             "float" : float,
              "uint16" : uint16Builder}

    def __init__ (self, name, type, default = None, optional = False):
      self.name = name
      self.type = type
      self.builder = ImageDesc.Attribute.TYPES[type]
      self.default = None
      self.optional = optional
      if default:
        self.default = self.builder(default)
        self.optional = True

  def __init__ (self):
    self.name = None
    self.image_name = None
    self.max_slots = 0
    self.mem_default = 0
    self.mem_min = None
    self.mem_max = None
    self.hidden = False
    self.max_ports = None
    self.capabilities = ["NET_ADMIN", "NET_RAW"]
    self.console_shell = "/bin/bash"
    self.network_shadow = False
    self.volumes = []
    self.make_user_image = True
    self.image_attributes = {} 
    self.port_attributes = {}
    self.port_attrs_optional = True
    self.port_attrs_exclusive = False
    self.data_dir = None
    self._handler = None
    self.forward_ssh_cmd = False

  def loadHandler (self, info):
    mod = importlib.import_module(info["module"])
    self._handler = getattr(mod, info["class"])

  def convertVolumes (self, cobj):
    hvl = []
    for vol in self.volumes:
# do this with the user volume manager
      hvl.append(vol.convertToHostVolume(cobj))
    return hvl


class _ContainerManager(object):
  class UnknownImagespecError(CoreException):
    def __init__ (self, image_name):
      super(_ContainerManager.UnknownImagespecError, self).__init__()
      self.image_name = image_name
    def __str__ (self):
      return "Requested base imagespec '%s' not found." % (self.image_name)

  class MemoryOutOfBoundsError(CoreException):
    def __init__ (self, image, mem, mmin, mmax):
      super(_ContainerManager.MemoryOutOfBoundsError, self).__init__()
      self.imagename = image
      self.request = mem
      self.mmin = mmin
      self.mmax = mmax
    def __str__ (self):
      return "Requested memory (%s MB) for image (%s) is outside of allowable bounds (%s - %s)" % (
        self.request, self.imagename, self.mmin, self.mmax)


  def __init__ (self):
    super(_ContainerManager, self).__init__()
    self._specroot = None
    self.log = None
    self._baseimgs = {}
    self._sliverimgs = set()
    self.dlog = KeyAdapter("ContainerManager", DKRLOG)

  def call (self, cmd, log_cmd = None, log_output = None):
    if not log_cmd:
      log_cmd = self.lcmd
    (o,e,r) = call(cmd, log_cmd, log_output)
    if o and e:
      self.dlog.warning(cmd)
      self.dlog.warning("[stdout] %s" % (o.strip()))
      self.dlog.warning("[stderr] %s" % (e.strip()))
    elif o:
      self.dlog.debug(cmd)
      self.dlog.debug("[stdout] %s" % (o.strip()))
    elif e:
      self.dlog.warning(cmd)
      self.dlog.warning("[stderr] %s" % (e.strip()))
    return (o,e,r)

  def _load (self):
    self._specroot = ConfigDB.getConfigItemByKey("docker.imagespec-root").getValue()
    baseroot = "%s/_basespecs" % (self._specroot)
    for dname in os.listdir(baseroot):
      if not os.path.isdir("%s/%s" % (baseroot, dname)):
        continue
      spath = "%s/%s/spec.json" % (baseroot, dname)
      if not os.path.exists(spath):
        ### TODO: Log this
        continue
      self._loadbase(spath)
    self.log.info("Loaded")

  def _loadbase (self, path):
    self.log.debug("Loading %s" % (path))
    f = open(path, "rb")
    obj = json.loads(f.read())
    imgd = None
    if obj["version"] >= 1:
      imgd = ImageDesc()
      imgd.name = obj["name"]
      imgd.max_slots = obj["max-slots"]
      imgd.mem_default = obj["memory"]
      imgd.mem_min = imgd.mem_default
      imgd.mem_max = imgd.mem_default
      imgd.image_name = json.getdefault(obj, "image-name", imgd.name)
    if obj["version"] >= 2:
      imgd.console_shell = obj["console-shell"]
    if obj["version"] >= 3:
      imgd.capabilities = obj["capabilities"]
      for volspec in obj["volumes"]:
        if volspec["type"] == "seeded":
          imgd.volumes.append(SeededDataVolume(volspec["src-repo"], volspec["container-path"]))
        elif volspec["type"] == "raw":
          imgd.volumes.append(RawHostVolume(volspec["mount-path"]))
        else:
          self.log.warning("Unknown volume type: %s" % (volspec["type"]))
      imgd.network_shadow = obj["network-shadow"]
    if obj["version"] >= 4:
      imgd.make_user_image = obj["make-user-image"]
    if obj["version"] >= 5:
      for ainfo in json.getdefault(obj, "attributes", []):
        attr = ImageDesc.Attribute(ainfo["name"], ainfo["type"], json.getdefault(ainfo, "default", None),
                                   json.getdefault(ainfo, "optional", False))
        imgd.image_attributes[attr.name] = attr
      if json.getdefault(obj, "handler", None):
        imgd.loadHandler(obj["handler"])
      imgd.data_dir = obj["data-dir"]
    if obj["version"] >= 6:
      if json.getdefault(obj, "port-attributes", None):
        paspec = obj["port-attributes"]
        imgd.port_attrs_optional = paspec["optional"]
        imgd.port_attrs_exclusive = paspec["exclusive"]
    if obj["version"] >= 7:
      imgd.mem_min = json.getdefault(obj, "memory-min", imgd.mem_default)
      imgd.mem_max = json.getdefault(obj, "memory-max", imgd.mem_default)
    if obj["version"] >= 8:
      imgd.hidden = json.getdefault(obj, "hidden", False)
    if obj["version"] >= 9:
      imgd.forward_ssh_cmd = json.getdefault(obj, "forward-ssh-command", False)

    if imgd:
      self._baseimgs[imgd.name] = imgd
      self.log.info("Added image: %s" % (imgd.name))

  def setLogger (self, logger):
    self.log = KeyAdapter("ContainerManager", logger)

  def getImageSpec (self, name):
    path = "%s/_basespecs/%s/Dockerfile" % (self._specroot, name)
    if not os.path.exists(path):
      raise _ContainerManager.UnknownImagespecError(name)

    data = open(path, "rb").read()
    return data

  def validImage (self, name):
    return self._baseimgs.has_key(name)

  def getImageDesc (self, name):
    try:
      return self._baseimgs[name]
    except KeyError:
      raise _ContainerManager.UnknownImagespecError(name)

  def buildImageForSliver (self, sliver, imagename, opts):
    desc = self.getImageDesc(imagename)
    if not desc.make_user_image:
      return desc.image_name

    sliver_image = "%s-%s" % (sliver.uuid, imagename)
    if sliver_image in self._sliverimgs:
      return sliver_image

    data = self.getImageSpec(imagename)
    odata = data % (opts)

    path = "%s/%s/%s" % (self._specroot, sliver.uuid, imagename)
    if not os.path.exists(path):
      os.makedirs(path)
    nf = open("%s/Dockerfile" % (path), "w")
    nf.write(odata)
    nf.close()

    (o, e, ret) = self.call("sudo docker build -t %s %s/" % (sliver_image, path), log_cmd=self.log.info)
    self._sliverimgs.add(sliver_image)
    return sliver_image

  @property
  def imagedescs (self):
    return self._baseimgs.itervalues()


CM = _ContainerManager()

class DockerDriver(object):
  def __init__ (self, log_cmd):
    super(DockerDriver, self).__init__()
    self.lcmd = log_cmd
    self.dlog = KeyAdapter("Driver", DKRLOG)

  def call (self, cmd, log_cmd = None, log_output = None):
    if not log_cmd:
      log_cmd = self.lcmd
    (o,e,r) = call(cmd, log_cmd, log_output)
    if o and e:
      self.dlog.warning(cmd)
      self.dlog.warning("[stdout] %s" % (o.strip()))
      self.dlog.warning("[stderr] %s" % (e.strip()))
    elif o:
      self.dlog.debug(cmd)
      self.dlog.debug("[stdout] %s" % (o.strip()))
    elif e:
      self.dlog.warning(cmd)
      self.dlog.warning("[stderr] %s" % (e.strip()))
    return (o,e,r)

  def ccall (self, cobj, cmd, log_cmd = None, log_output = None):
    if not log_cmd:
      log_cmd = self.lcmd
    newcmd = "sudo docker exec %s %s" % (cobj.id, cmd)
    (o,e,r) = self.call(newcmd, log_cmd, log_output)
    return (o,e,r)

  def build_image (self, urn, basename):
    uname = urn.split("+")[-1]
    (o, e, ret) = self.call("sudo docker build -t %s-%s %s/" % (basename, uname, urn))
    return (o, e)

  def hup (self, cobj, pid):
    (o,e,r) = self.call("sudo ip netns exec %d kill -HUP %d" % (cobj.pid, pid))

  def supervisor (self, cobj, action, name = None):
    if name:
      (o,e,r) = self.call("sudo docker exec %s supervisorctl %s %s" % (cobj.id, action, name))
    else:
      (o,e,r) = self.call("sudo docker exec %s supervisorctl %s" % (cobj.id, action))
    return (o,e,r)

  def setInterfaceIP (self, cobj, intf_name, ipstr):
    (o,e,r) = self.call("sudo ip netns exec %d ip addr flush dev %s" % (cobj.pid, intf_name))
    (o,e,r) = self.call("sudo ip netns exec %d ip addr add %s dev %s" % (cobj.pid, ipstr, intf_name))

  def setupNAT (self, pnr, cname, ip_addr, gateway):
    # TODO: Clean these names up
    p1 = pnr.getName()
    p2 = pnr.getName()

    hd = HostDriver(self.lcmd)
    hd.createWire(p1, p2)

    (o,e,r) = self.call("sudo brctl addif docker0 %s" % (p1))
    (o,e,r) = self.call("sudo ip link set %s up" % (p1))

    obj = self.container_info(cname)
    pid = int(obj[0]["State"]["Pid"])

    (o,e,r) = self.call("sudo ip link set %s netns %d" % (p2, pid))
    (o,e,r) = self.call("sudo ip netns exec %d ip link set dev %s name eth0" % (pid, p2))
    (o,e,r) = self.call("sudo ip netns exec %d ip link set dev eth0 up" % (pid))
    (o,e,r) = self.call("sudo ip netns exec %d ip addr add %s dev eth0" % (pid, ip_addr))
    (o,e,r) = self.call("sudo ip netns exec %d ip route add default via %s" % (pid, gateway))

  def run_container (self, cobj, vols = None, netctid = None):
    caps = []
    for cap in cobj._image_desc.capabilities:
      caps.append("--cap-add=%s"% (cap))

    cobj.prebuild()

    if vols is None:
      vols = []


    # We only pass volumes to the stacked container, BOTH need the same caps
    if cobj._image_desc.network_shadow and not netctid:
      (o, e, r) = self.call("sudo docker run -d -h %s --name=sleep-%s --memory=6m --memory-swap=-1 --net=\"none\" %s tianon/sleeping-beauty" % (
                            cobj.client_id, cobj.id, " ".join(caps)))
      obj = self.container_info(cobj.id, True)
      pid = int(obj[0]["State"]["Pid"])

    elif cobj._image_desc.network_shadow and netctid:
      vols.extend(cobj._image_desc.convertVolumes(cobj))
      if cobj._instance_handler:
        vols.extend(cobj._instance_handler.volumes)

      (o, e, r) = self.call("sudo docker run -d --name=%s --memory=%dm --memory-swap=-1 --net=container:%s %s %s %s" % (
                            cobj.id, cobj.ram, netctid, " ".join(caps), " ".join([x.args for x in vols]),
                            cobj._image_desc.image_name))

      obj = self.container_info(cobj.id)
      pid = int(obj[0]["State"]["Pid"])

    else: # No shadow, we do everything here
      vols.extend(cobj._image_desc.convertVolumes(cobj))
      if cobj._instance_handler:
        vols.extend(cobj._instance_handler.volumes)

      (o, e, r) = self.call("sudo docker run -d -h %s --name=%s --memory=%dm --memory-swap=-1 --net=\"none\" %s %s %s" % (
                            cobj.client_id, cobj.id, cobj.ram, " ".join(caps),
                            " ".join([x.args for x in vols]), cobj._image_desc.image_name))

      obj = self.container_info(cobj.id)
      pid = int(obj[0]["State"]["Pid"])

    # Put volumes back for lookup in handlers
    cobj._vols = vols

    # TODO: If there's an IndexError here, or obj is basically junk, we should go back to o/e
    # from the run call to find out what happened
    (o,e,r) = call("sudo ln -s /proc/%d/ns/net /var/run/netns/%d" % (pid, pid))
    return obj

  def container_info (self, name, shadow = False):
    if shadow:
      name = "sleep-%s" % (name)

    (o, e, r) = self.call("sudo docker inspect %s" % (name))

    obj = json.loads(o)
    return obj

  def setPortImpairment (self, cobj, pname, dinfo, linfo, rinfo = None):
    argl = ["sudo ip netns exec %d tc qdisc add dev %s root netem" % (cobj.pid, pname)]

    if dinfo:
      argl.append("delay %dms" % (dinfo.time))
      if dinfo.jitter:
        argl.append("%dms" % (dinfo.jitter))
        if dinfo.correlation:
          argl.append("%d%%" % (dinfo.correlation))
      if dinfo.distribution:
        argl.append("distribution %s" % (dinfo.distribution))

    if linfo:
      argl.append("loss random %s%%" % (linfo.percent))

    if rinfo:
      if rinfo.gap:
        argl.append("reorder %s%% %s%% gap %s" % (str(rinfo.percent), str(rinfo.correlation), str(rinfo.gap)))
      else:
        argl.append("reorder %s%% %s%%" % (str(rinfo.percent), str(rinfo.correlation)))

    cmd = " ".join(argl)
    (o,e,r) = self.call(cmd)

#  def setDelay (self, cobj, pname, dinfo, change = False):
#    if dinfo.time == 0:
#      return
#    argl = ["sudo ip netns exec %d tc qdisc" % (cobj.pid)]
#    if change:
#      argl.append("change")
#    else:
#      argl.append("add")
#    argl.append("dev %s root netem delay %dms" % (pname, dinfo.time))
#    if dinfo.jitter:
#      argl.append("%dms" % (dinfo.jitter))
#      if dinfo.correlation:
#        argl.append("%d%%" % (dinfo.correlation))
#    if dinfo.distribution:
#      argl.append("distribution %s" % (dinfo.distribution))
#    cmd = " ".join(argl)
#
#    call(cmd, log_cmd = self.lcmd)
#
#  def setLoss (self, cobj, pname, linfo, change = False):
#    if linfo.percent == 0:
#      return
#    argl = ["sudo ip netns exec %d tc qdisc" % (cobj.pid)]
#    if change:
#      argl.append("change")
#    else:
#      argl.append("add")
#    argl.append("dev %s root netem loss random %s%%" % (pname, linfo.percent))
#    cmd = " ".join(argl)
#
#    call(cmd, log_cmd = self.lcmd)

  def clearImpairment (self, cobj, pname):
    (o,e,r) = self.call("sudo ip netns %d tc qdisc delete dev %s root" % (cobj.pid, pname))

  def attachPort (self, cobj, port):
    pid = cobj.pid
    (o,e,r) = self.call("sudo ip link set %s netns %d" % (port.name, pid))
    (o,e,r) = self.call("sudo ip netns exec %d ip link set dev %s name %s" % (pid, port.name, port.alias))
    (o,e,r) = self.call("sudo ip netns exec %d ip link set dev %s up" % (pid, port.alias))

  def portDown (self, port):
    (o,e,r) = self.call("sudo ip netns exec %d ip link set dev %s down" % (port.datapath.pid, port.alias))

  def portUp (self, port):
    (o,e,r) = self.call("sudo ip netns exec %d ip link set dev %s up" % (port.datapath.pid, port.alias))

  def getMACAddress (self, cobj, pname):
    (o,e,r) = self.call("sudo ip netns exec %d cat /sys/class/net/%s/address" % (cobj.pid, pname))
    mac = o.strip()
    return mac

  def addPortIPv4Address (self, cobj, pname, address):
    pid = cobj.pid
    (o,e,r) = self.call("sudo ip netns exec %d ip address add %s dev %s" % (pid, address, pname))

  def getPortIPv4Addresses (self, cobj, pname):
    addrs = []
    (o,e,r) = self.call("sudo ip netns exec %d ip address show %s | grep 'inet '" % (cobj.pid, pname))
    if o.strip():
      for line in o.split("\n"):
        if line.strip():
          addrs.append(line.strip().split()[1])
    return addrs

  def getProcData (self, cobj, path):
    (o,e,r) = self.call("sudo ip netns exec %d cat /proc/%s" % (cobj.pid, path))
    return o.strip().splitlines()

  def addIPRoute (self, cobj, network, gateway):
    (o,e,r) = self.call("sudo ip netns exec %d ip route add %s via %s" % (cobj.pid, network, gateway))

  def destroyContainer (self, name, cobj):
    (o,e,r) = self.call("sudo docker kill %s" % (name), log_cmd = self.lcmd)
    (o,e,r) = self.call("sudo docker rm -v %s" % (name), log_cmd = self.lcmd)

    if cobj._image_desc.network_shadow:
      (o,e,r) = self.call("sudo docker kill sleep-%s" % (name), log_cmd = self.lcmd)
      (o,e,r) = self.call("sudo docker rm sleep-%s" % (name), log_cmd = self.lcmd)

    if cobj._instance_handler:
      cobj._instance_handler.postrm(cobj)

  def changeMemoryLimits (self, cobj, mem_max):
    # mem_max in mb
    memfile = "/sys/fs/cgroup/memory/docker/%s/memory.limit_in_bytes" % (cobj.longid)
    (o,e,r) = self.call("sudo sh -c 'echo %dm > %s'" % (mem_max, memfile))


#  def attach_port (self, cobj, dpname, p_a, p_b, internal_name = None):
#  #  pid = int(cobj[0]["NetworkSettings"]["Ports"]["22/tcp"][0]["HostPort"])
#    pid = int(cobj[0]["State"]["Pid"])
#    (o,e,r) = call("sudo ip link add %s type veth peer name %s" % (p_a, p_b))
#    (o,e,r) = call("sudo ip link set %s netns %d" % (p_a, pid))
#    if internal_name:
#      (o,e,r) = call("sudo ip netns exec %d ip link set dev %s name %s" % (pid, p_a, internal_name))
#      p_a = internal_name
#    (o,e,r) = call("sudo ip netns exec %d ip link set dev %s up" % (pid, p_a))
#
#    (o,e,r) = call("sudo ovs-vsctl add-port %s %s" % (dpname, p_b))
#    (o,e,r) = call("sudo ip link set %s up" % (p_b))

def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("docker.imagespec-root").setValue("/opt/foam/docker-imagespecs")
                .setDesc("Storage root for per-user container metadata"))
  [ConfigDB.installConfigItem(x) for x in citems]

  if not os.path.exists("/var/run/netns"):
    call("sudo mkdir /var/run/netns")

  CM.setLogger(app.logger)
  CM._load()
