# Copyright (c) 2016-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime
import errno
import os
import os.path
import shutil
import subprocess
import tempfile

from .exceptions import OutOfMountSpaceError
from .. import tasks
from ...core import json
from ...core.configdb import ConfigDB, ConfigItem
from ...core.exception import CoreException
from ...core.log import KeyAdapter
from ...geni.resources import Registry as ResourceRegistry
from ...geni.version import Registry as APIRegistry
from ...types import UUID

SUDO = "/usr/bin/sudo"
UPLOADBIN = "/opt/foam/bin/dropbox-upload"

class VolumeInfo(object):
  class BuildError(object):
    def __init__ (self, obj):
      self._obj = obj
    def __str__ (self):
      return "Could not reconstruct volume from %s" % (str(obj))

  def __init__ (self, name, local_dir, cmount):
    self.volid = UUID()
    self.name = name
    self.local_dir = local_dir
    self.container_mount = cmount
    self.dropbox_path = None

  def __json__ (self):
    d = {"version" : 1, 
         "volid" : self.volid,
         "name" : self.name,
         "dropbox-path" : self.dropbox_path,
         "local-dir" : self.local_dir,
         "container-mount" : self.container_mount}
    return d

  @staticmethod
  def jconstruct (obj):
    if obj["version"] == 1:
      vd = VolumeInfo(obj["name"], obj["local-dir"], obj["container-mount"])
      vd.volid = UUID.construct(obj["volid"])
      vd.dropbox_path = obj["dropbox-path"]
      return vd
    raise VolumeInfo.BuildError(obj)

  @property
  def hostvol (self):
    # Translate relevant info into a HostVolume for mounting
    from .docker import HostPathVolume
    return HostPathVolume(self.local_dir, self.container_mount)

  def upload (self, userurn):
    from .dropbox import Driver as DropboxDriver
    try:
      token = DropboxDriver.getUserToken(userurn)
    except KeyError:
      # Raise error
      return

    p = subprocess.Popen([str(x) for x in [SUDO, UPLOADBIN, token, self.local_dir, self.dropbox_path]],
                         stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    return p.pid



class _DropboxVolumeManager(object):
  class NotEnabledError(CoreException):
    def __str__ (self):
      return "The Dropbox Volume Manager has not been enabled."

  def __init__ (self):
    self._log = None
    self._root = None
    self._reservations = {} # { <sliver-uuid> : { <container> : [volume_info, ...], ... }, ... }

  @property
  def log (self):
    return self._log

  @log.setter
  def log (self, logger):
    self._log = KeyAdapter("DbxVolMgr", logger)

  def _load (self):
    from .dropbox import Driver as DropboxDriver
    if not DropboxDriver.enabled:
      return

    vinfo = {"vts.dropbox.enabled" : True}
    APIRegistry.addStaticInfo(vinfo)

    self._root = ConfigDB.getConfigItemByKey("dropbox.user-vols.root").getValue()
    rsv = json.loads(ConfigDB.getConfigItemByKey("dropbox.user-vols.reservations").getValue())
    for slv_uuid,info in rsv.iteritems():
      for ctname,vol_list in info.iteritems():
        # It's about time for the config db to manage this for us...
        self._reservations.setdefault(slv_uuid, {})[ctname] = [VolumeInfo.jconstruct(x) for x in vol_list]
          
    if self._root:
      ResourceRegistry.addPOACallback("vts:dropbox:upload", poaDropboxUpload)

  def createVolume (self, sliver, container, name, cmount):
    if not self._root:
      raise _DropboxVolumeManager.NotEnabledError()

    vpath = "%s/%s/%s/%s" % (self._root, sliver.uuid, container.client_id, name)
    vinfo = VolumeInfo(name, vpath, cmount)
    vinfo.dropbox_path = "/%s/%s/%s" % (sliver.uuid, container.client_id, name)
    try:
      os.makedirs(vpath)
    except OSError as e:
      if e.errno == errno.ENOSPC:
        raise OutOfMountSpaceError("dropbox")
      raise

    self._reservations.setdefault(sliver.uuid, {}).setdefault(container.client_id, []).append(vinfo)
    ConfigDB.getConfigItemByKey("dropbox.user-vols.reservations").write(self._reservations)

    return vinfo

  def removeVolume (self, sliver, container, volid):
    try:
      vlist = self._reservations[sliver.uuid][container.client_id]
      for vol in vlist:
        if vol.volid == volid:
          shutil.rmtree("%s/%s/%s/%s" % (self._root, sliver.uuid, container.client_id, vol.name), True)
          ConfigDB.getConfigItemByKey("dropbox.user-vols.reservations").write(self._reservations)
          return
    except KeyError:
      # We can't find the thing you're trying to remove
      return
  
  def removeVolumes (self, sliver, container = None):
    # This clears all volumes for a sliver or container in one shot
    # TODO: These paths should be constructed from a single place
    try:
      if container:
        tasks.destroyTree("%s/%s/%s" % (self._root, sliver.uuid, container.client_id))
        del self._reservations[sliver.uuid][container.client_id]
      else:
        tasks.destroyTree("%s/%s" % (self._root, sliver.uuid))
        del self._reservations[sliver.uuid]
      ConfigDB.getConfigItemByKey("dropbox.user-vols.reservations").write(self._reservations)
    except KeyError:
      # Nothing to see here, please move along
      return


def poaDropboxUpload (action, urns, options, user_urn):
  # options:
  # { "vols" :
  #  [
  #   { container-name : [volid, ...] },
  #  ...]
  # }

  from .dropbox import Driver as DropboxDriver
  from ..data import DB
  from ..geni import SliverBundleProxy
  from ...geni.resources import UnknownSliceError, UnsupportedResourceActionError, Registry

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  # Sliver Data
  skip = False
  slice_urn = urns[0]
  datadir = "%s-%s" % (slice_urn, datetime.datetime.now().strftime("%Y%m%d%H%M"))
  dropbox_path = "/%s/%s" % (sliver.uuid, datadir)
  path = os.path.normpath("%s/%s" % (tempfile.gettempdir(), datadir))
  try:
    os.makedirs(path)
  except OSError:
    # TODO: This might be in progress, lets just skip the whole mess since we did it in the last minute.
    skip = True
  if not skip:
    token = DropboxDriver.getUserToken(user_urn)
    mfest = Registry.generateManifest(slice_urn)
    slivers = Registry.getSlivers(slice_urn)
    sstatus = Registry.sliverStatus(slivers, {})
    with open("%s/manifest.xml" % (path), "w+") as mf:
      mf.write(mfest)
    with open("%s/sliverstatus.xml" % (path), "w+") as ssf:
      ssf.write(str(sstatus))
    p = subprocess.Popen([str(x) for x in [SUDO, UPLOADBIN, token, path, dropbox_path, "delete"]],
                         stdout = subprocess.PIPE, stderr = subprocess.PIPE)

# Upload it, tell the uploader to delete it

  # User Data
  for data in options["vols"]:
    for cname,vlist in data.iteritems():
      container = sliver.getContainer(cname)
      if not container:
        sliver.log.warning("Did not find container [%s]" % (cname))
        continue
      sliver.log.debug("Volume ids: %s" % (",".join([str(x.volid) for x in container._dbxvols])))
      for vid in vlist:
        for volinfo in container._dbxvols:
          if str(volinfo.volid) != str(vid):
            continue
          pid = volinfo.upload(user_urn)
          if pid:
            sliver.log.info("Uploading %s for %s" % (vid, user_urn))
          else:
            sliver.log.warning("User-urn %s tried upload without dropbox token" % (user_urn))
          break

DropboxVolumeManager = _DropboxVolumeManager()

def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("dropbox.user-vols.root").setValue(None)
                .setDesc("Storage root for dropbox user volumes"))
  citems.append(ConfigItem().setKey("dropbox.user-vols.reservations").setValue("{}")
                .setDesc("Storage of current reservations")
                .setUpdateFuncName("foam.core.configdb.checkDictJson"))
  [ConfigDB.installConfigItem(x) for x in citems]

  DropboxVolumeManager.log = app.logger
  DropboxVolumeManager._load()
