# Copyright (c) 2016-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import shutil
import subprocess
import urlparse

from ...core import json
from ...core.configdb import ConfigDB, ConfigItem
from ...core.exception import CoreException, UserException
from ...core.log import KeyAdapter
from ...geni.resources import Registry as ResourceRegistry
from ...geni.version import Registry as APIRegistry
from ...types import UUID

WORKERBIN = "/opt/foam/bin/hg-worker"

class NoPullWithPasswordError(UserException):
  def __str__ (self):
    return "You cannot pull repositories that require a password."

class VolumeInfo(object):
  class BuildError(object):
    def __init__ (self, obj):
      self._obj = obj
    def __str__ (self):
      return "Could not reconstruct volume from %s" % (str(obj))

  class CommandError(object):
    def __init__ (self, error):
      self.error = error
    def __str__ (self):
      return "Command error: %s" % (self.error)

  def __init__ (self, source, local_dir, cmount, sensitive = False, requested_source = None):
    self.volid = UUID()
    self.source = source
    self.requested_source = requested_source
    self.local_dir = local_dir
    self.branch = "default"
    self.container_mount = cmount
    self.sensitive = sensitive
    self.name = None

  def __json__ (self):
    d = {"version" : 4, 
         "volid" : self.volid,
         "branch" : self.branch,
         "source" : self.source,
         "requested-source" : self.requested_source,
         "container-mount" : self.container_mount,
         "local-dir" : self.local_dir,
         "sensitive" : self.sensitive,
         "name" : self.name}
    return d

  def __json_ephemeral__ (self): 
    d = {"version" : 4,
         "volid" : self.volid,
         "branch" : self.branch,
         "container-mount" : self.container_mount,
         "local-dir" : self.local_dir,
         "name" : self.name,
         "source" : self.source}
    if self.sensitive:
      d["requested-source"] = "<elided>"
    else:
      d["requested-source"] = self.requested_source
    return d

  @staticmethod
  def jconstruct (obj):
    from ..data import DB
    if obj["version"] >= 1:
      vd = VolumeInfo(obj["source"], obj["local-dir"], obj["container-mount"])
      vd.volid = UUID.construct(obj["volid"])
      vd.branch = obj["branch"]
    if obj["version"] >= 2:
      vd.sensitive = obj["sensitive"]
    if obj["version"] >= 3:
      vd.name = obj["name"]
    if obj["version"] >= 4:
      vd.source = DB.getManagedSource(obj["requested-source"])
      vd.requested_source = obj["requested-source"]
    return vd
      
  @property
  def hostvol (self):
    # Translate relevant info into a HostVolume for mounting
    from .docker import HostPathVolume
    return HostPathVolume(self.local_dir, self.container_mount)

  def clone (self):
    arglist = [str(x) for x in [WORKERBIN, "clone", self.source, self.local_dir, self.branch]]
    HGVolumeManager.log.debug("Cloning: %s" % (" ".join(arglist)))
    p = subprocess.Popen(arglist, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    return p.pid

  def getRequestedSource (self):
    return self.requested_source

  def pull (self, update = True):
    udata = urlparse.urlparse(self.source)
    if udata.password:
      raise NoPullWithPasswordError()

    p = subprocess.Popen([str(x) for x in [WORKERBIN, "pull", self.local_dir]],
                         stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    return p.pid


class _HGVolumeManager(object):
  class NotEnabledError(CoreException):
    def __str__ (self):
      return "The HG Volume Manager has not been enabled."

  def __init__ (self):
    self._log = None
    self._root = None
    self._reservations = {} # { <sliver-uuid> : { <container> : [volume_info, ...], ... }, ... }

  @property
  def log (self):
    return self._log

  @log.setter
  def log (self, logger):
    self._log = KeyAdapter("HGVolMgr", logger)

  def _load (self):
    self._root = ConfigDB.getConfigItemByKey("hg.user-vols.root").getValue()
    rsv = json.loads(ConfigDB.getConfigItemByKey("hg.user-vols.reservations").getValue())
    for slv_uuid,info in rsv.iteritems():
      for ctname,vol_list in info.iteritems():
        # It's about time for the config db to manage this for us...
        self._reservations.setdefault(slv_uuid, {})[ctname] = [VolumeInfo.jconstruct(x) for x in vol_list]
          
    if self._root:
      vinfo = {"vts.hg.enabled" : True}
      APIRegistry.addStaticInfo(vinfo)
      ResourceRegistry.addPOACallback("vts:hg:pull", poaHgPull)

  def createVolume (self, sliver, container, name, source, cmount, branch = None, req_source = None):
    if not self._root:
      raise _HGVolumeManager.NotEnabledError()

    vpath = "%s/%s/%s/%s" % (self._root, sliver.uuid, container.client_id, name)
    if req_source:
      vinfo = VolumeInfo(source, vpath, cmount, requested_source = req_source)
    else:
      vinfo = VolumeInfo(source, vpath, cmount)
    vinfo.name = name
    if branch:
      vinfo.branch = branch
    os.makedirs(vpath)

    self._reservations.setdefault(sliver.uuid, {}).setdefault(container.client_id, []).append(vinfo)
    ConfigDB.getConfigItemByKey("hg.user-vols.reservations").write(self._reservations)

    return vinfo

  def removeVolume (self, sliver, container, volid):
    try:
      vlist = self._reservations[sliver.uuid][container.client_id]
      for vol in vlist:
        if vol.volid == volid:
          shutil.rmtree("%s/%s/%s/%s" % (self._root, sliver.uuid, container.client_id, vol.name), True)
          ConfigDB.getConfigItemByKey("hg.user-vols.reservations").write(self._reservations)
          return
    except KeyError:
      # We can't find the thing you're trying to remove
      return
  
  def removeVolumes (self, sliver, container = None):
    # This clears all volumes for a sliver or container in one shot
    # TODO: These paths should be constructed from a single place
    try:
      if container:
        shutil.rmtree("%s/%s/%s" % (self._root, sliver.uuid, container.client_id), True)
        del self._reservations[sliver.uuid][container.client_id]
      else:
        shutil.rmtree("%s/%s" % (self._root, sliver.uuid), True)
        del self._reservations[sliver.uuid]
      ConfigDB.getConfigItemByKey("hg.user-vols.reservations").write(self._reservations)
    except KeyError:
      # Nothing to see here, please move along
      return
    

def poaHgPull (action, urns, options, user_urn):
  # options:
  # { "vols" :
  #  [
  #   { container-name : [(volid, update), ...] },
  #  ...]
  # }
  from ..data import DB
  from ..geni import SliverBundleProxy
  from ...geni.resources import UnknownSliceError, UnsupportedResourceActionError

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  vols_to_pull = []

  for data in options["vols"]:
    for cname,vlist in data.iteritems():
      container = sliver.getContainer(cname)
      if not container:
        sliver.log.info("[vts:hg:pull] Can't find container '%s'" % (cname))
        continue # Accumulate error?  Punt?
      sliver.log.debug("[vts:hg:pull] Searching vols [%s]" % (", ".join([str(x.volid) for x in container._hgvols])))
      for (vid, update) in vlist:
        for volinfo in container._hgvols:
          if str(volinfo.volid) != str(vid):
            continue
          sliver.log.info("[vts:hg:pull] Pulling for volume '%s'" % (str(volinfo.volid)))
          DB.CVM.pull(volinfo.requested_source)
          vols_to_pull.append((volinfo, update))
          break
  
  for volinfo, update in vols_to_pull:
    DB.checkAndWait(volinfo.requested_source)
    volinfo.pull(update)

  return True

HGVolumeManager = _HGVolumeManager()

def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("hg.user-vols.root").setValue(None)
                .setDesc("Storage root for hg user volumes"))
  citems.append(ConfigItem().setKey("hg.user-vols.reservations").setValue("{}")
                .setDesc("Storage of current reservations")
                .setUpdateFuncName("foam.core.configdb.checkDictJson"))
  [ConfigDB.installConfigItem(x) for x in citems]
  
  HGVolumeManager.log = app.logger
  HGVolumeManager._load()
