# Copyright (c) 2016-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import os.path

from .. import tasks
from ...core import json
from ...core.configdb import ConfigDB, ConfigItem
from ...core.exception import CoreException
from ...core.log import KeyAdapter
from ...geni.resources import Registry as ResourceRegistry
from ...geni.version import Registry as APIRegistry
from ...types import UUID

class VolumeInfo(object):
  class BuildError(object):
    def __init__ (self, obj):
      self._obj = obj
    def __str__ (self):
      return "Could not reconstruct volume from %s" % (str(obj))

  def __init__ (self, name, local_dir, cmount):
    self.volid = UUID()
    self.name = name
    self.local_dir = local_dir
    self.container_mount = cmount

  def __json__ (self):
    d = {"version" : 1, 
         "volid" : self.volid,
         "name" : self.name,
         "local-dir" : self.local_dir,
         "container-mount" : self.container_mount}
    return d

  @staticmethod
  def jconstruct (obj):
    if obj["version"] == 1:
      vd = VolumeInfo(obj["name"], obj["local-dir"], obj["container-mount"])
      vd.volid = UUID.construct(obj["volid"])
      return vd
    raise VolumeInfo.BuildError(obj)

  @property
  def hostvol (self):
    # Translate relevant info into a HostVolume for mounting
    from .docker import HostPathVolume
    return HostPathVolume(self.local_dir, self.container_mount)


class _UserVolumeManager(object):
  class OutOfMountSpaceError(CoreException):
    def __init__ (self, type):
      super(OutOfMountSpaceError, self).__init__()
    def __str__ (self):
      return "There is no more space available to create user volumes"

  class NotEnabledError(CoreException):
    def __str__ (self):
      return "The User Volume Manager has not been enabled."

  def __init__ (self):
    self._log = None
    self._root = None
    self._reservations = {} # { <sliver-uuid> : { <container> : [volume_info, ...], ... }, ... }

  @property
  def log (self):
    return self._log

  @log.setter
  def log (self, logger):
    self._log = KeyAdapter("UserVolMgr", logger)

  def _load (self):
    self._root = ConfigDB.getConfigItemByKey("docker.user-vols.root").getValue()
    if not self._root:
      return

    vinfo = {"vts.docker-vols.enabled" : True}
    APIRegistry.addStaticInfo(vinfo)

    rsv = json.loads(ConfigDB.getConfigItemByKey("docker.user-vols.reservations").getValue()
    for slv_uuid,info in rsv.iteritems():
      for ctname,vol_list in info.iteritems():
        # It's about time for the config db to manage this for us...
        self._reservations.setdefault(slv_uuid, {})[ctname] = [VolumeInfo.jconstruct(x) for x in vol_list]

  def _write (self):
    ConfigDB.getConfigItemByKey("docker.user-vols.reservations").write(self._reservations)

  def createVolume (self, sliver, container, name, cmount):
    if not self._root:
      raise _UserVolumeManager.NotEnabledError()

    vpath = "%s/%s/%s/%s" % (self._root, sliver.uuid, container.client_id, name)
    vinfo = VolumeInfo(name, vpath, cmount)
    try:
      os.makedirs(vpath)
    except OSError as e:
      if e.errno == errno.ENOSPC:
        raise OutOfMountSpaceError()
      raise

    self._reservations.setdefault(sliver.uuid, {}).setdefault(container.client_id, []).append(vinfo)
    self._write()

    return vinfo

  def removeVolumes (self, sliver, container = None):
    # This clears all volumes for a sliver or container in one shot
    # TODO: These paths should be constructed from a single place
    try:
      if container:
        tasks.destroyTree("%s/%s/%s" % (self._root, sliver.uuid, container.client_id))
        del self._reservations[sliver.uuid][container.client_id]
      else:
        tasks.destroyTree("%s/%s" % (self._root, sliver.uuid))
        del self._reservations[sliver.uuid]
      self._write()
    except KeyError:
      # Nothing to see here, please move along
      return


UserVolumeManager = _UserVolumeManager()


def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("docker.user-vols.root").setValue(None)
                .setDesc("Storage root for docker user volumes"))
  citems.append(ConfigItem().setKey("docker.user-vols.reservations").setValue({})
                .setDesc("Storage of current reservations")
                .setUpdateFuncName("foam.core.configdb.checkDict"))
  [ConfigDB.installConfigItem(x) for x in citems]

  VolumeManager.setLogger(app.logger)
  Volumemanager._load()
