# Copyright (c) 2016-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import dropbox

from ...core.configdb import ConfigDB, ConfigItem
from ...core.exception import CoreException
from ...core.log import KeyAdapter
from ...geni.resources import Registry

class UnknownAuthFlowError(CoreException):
  def __init__ (self, urn):
    super(UnknownAuthFlowError, self).__init__()
    self.userurn = urn
  def __str__ (self):
    return "No dropbox authorization pending for user %s" % (self.userurn)

class _Driver(object):
  def __init__ (self):
    self.tokens = {}
    self._inprogress_auth = {}
    self._APPKEY = None
    self._APPSECRET = None
    self._log = None
    self.enabled = False

  @property
  def log (self):
    return self._log

  @log.setter
  def log (self, val):
    self._log = KeyAdapter("Dropbox", val)

  def _load (self):
    self._APPKEY = ConfigDB.getConfigItemByKey("dropbox.app-key").getValue()
    self._APPSECRET = ConfigDB.getConfigItemByKey("dropbox.app-secret").getValue()
    self.tokens = ConfigDB.getConfigItemByKey("dropbox.users").getValue()

    if self._APPKEY is None or self._APPSECRET is None:
      return

    self.enabled = True

    Registry.addPAACallback("vts:dropbox:link-account", self.paaLinkAccount)
    Registry.addPAACallback("vts:dropbox:complete-link", self.paaLinkComplete)

  def _write (self):
    ConfigDB.getConfigItemByKey("dropbox.users").write(self.tokens)

  def startAuth (self, user_urn):
    auth_flow = dropbox.DropboxOAuth2FlowNoRedirect(self._APPKEY, self._APPSECRET)
    self._inprogress_auth[user_urn] = auth_flow
    return auth_flow.start()

  def finalizeAuth (self, user_urn, auth_code):
    try:
      auth_flow = self._inprogress_auth[user_urn]
    except KeyError:
      err = UnknownAuthFlowError(user_urn)
      err._unexpected = False
      raise err

    try:
      res = auth_flow.finish(auth_code)
      self.tokens[user_urn] = (res.access_token, res.account_id, res.user_id)
      del self._inprogress_auth[user_urn]
      self._write()
    except Exception, e:
      self.log.exception("Finalize Exception")

  def getUserToken (self, userurn):
    return self.tokens[userurn][0]

  def paaLinkAccount (self, action, options, user_urn):
    url = self.startAuth(user_urn)
    return {"authorize_url" : url}

  def paaLinkComplete (self, action, options, user_urn):
    self.finalizeAuth(user_urn, options["auth-code"])
    return {}


Driver = _Driver()

def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("dropbox.users").setValue({})
                .setDesc("Dropbox User Token Mapping")
                .setUpdateFuncName("foam.core.configdb.checkDict"))
  citems.append(ConfigItem().setKey("dropbox.app-key").setValue(None)
                .setDesc("Dropbox App Key"))
  citems.append(ConfigItem().setKey("dropbox.app-secret").setValue(None)
                .setDesc("Dropbox App Secret"))
  [ConfigDB.installConfigItem(x) for x in citems]

  Driver.log = app.logger
  Driver._load()
