# Copyright (c) 2017  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Start of an OVS handler outside of the core of VTS

from ...core.exception import CoreException

class IntegerParameterOutOfRangeError(CoreException):
  def __init__ (self, name, val, vmin, vmax):
    super(IntegerParameterOutOfRangeError, self).__init__()
    self.name = name
    self.val = val
    self.vmin = vmin
    self.vmax = vmax

  def __str__ (self):
    return "Value (%d) for parameter (%s) out of range (%d, %d)" % (self.val, self.name, self.vmin, self.vmax)


def _intrange (name, val, vmin, vmax):
  if val >= vmin and val <= vmax:
    return val
  raise IntegerParameterOutOfRangeError(name, val, vmin, vmax)


class DisabledSTP(object):
  def __json__ (self):
    return {"_type" : "disabled",
            "_version" : 1}

  @classmethod
  def jconstruct (cls, obj):
    return cls()

  @property
  def cmdlist (self):
    return []


class STPParams(object):
  def __init__ (self):
    self.system_id = None
    self._priority = None
    self._hello_time = None
    self._max_age = None
    self._forward_delay = None

  def __json__ (self):
    return {"_type" : "stp",
            "system-id" : self.system_id,
            "priority" : self.priority,
            "hello-time" : self.hello_time,
            "max-age" : self.max_age,
            "forward-delay" : self.forward_delay,
            "_version" : 1}

  @classmethod
  def jconstruct (cls, obj):
    if obj["_version"] == 1:
      o = cls()
      o.system_id = obj["system-id"]
      o._priority = obj["priority"]
      o._hello_time = obj["hello-time"]
      o._max_age = obj["max-age"]
      o._forward_delay = obj["forward-delay"]
      return o

  @property
  def priority (self):
    return self._priority

  @priority.setter
  def priority (self, val):
    self._priority = _intrange("stp-priority", val, 0, 65535)

  @property
  def hello_time (self):
    return self._hello_time

  @hello_time.setter
  def hello_time (self, val):
    self._hello_time = _intrange("stp-hello-time", val, 1, 10)

  @property
  def max_age (self):
    return self._max_age

  @max_age.setter
  def max_age (self, val):
    self._max_age = _intrange("stp-max-age", val, 6, 40)

  @property
  def forward_delay (self):
    return self._forward_delay

  @forward_delay.setter
  def forward_delay (self, val):
    self._forward_delay = _intrange("stp-forward-delay", val, 4, 30)

  @property
  def cmdlist (self):
    ocl = []
    if self.system_id: ocl.append("other_config:stp-system-id=%s" % (self.system_id))
    if self.priority: ocl.append("other_config:stp-priority=%d" % (self.priority))
    if self.hello_time: ocl.append("other_config:stp-hello-time=%d" % (self.hello_time))
    if self.max_age: ocl.append("other_config:stp-max-age=%d" % (self.max_age))
    if self.forward_delay: ocl.append("other_config:stp-forward-delay=%d" % (self.forward_delay))
    return ocl


class RSTPParams(object):
  def __init__ (self):
    self.address = None
    self._priority = None
    self._ageing_time = None
    self.force_version = None  # No idea what valid values are here
    self._max_age = None
    self._forward_delay = None
    self._hold_count = None

  def __json__ (self):
    return {"_type" : "rstp",
            "address" : self.address,
            "priority" : self.priority,
            "ageing-time" : self.ageing_time,
            "max-age" : self.max_age,
            "forward-delay" : self.forward_delay,
            "hold-count" : self.hold_count,
            "_version" : 1}

  @classmethod
  def jconstruct (cls, obj):
    if obj["_version"] == 1:
      o = cls()
      o.address = obj["address"]
      o._priority = obj["priority"]
      o._ageing_time = obj["ageing-time"]
      o._max_age = obj["max-age"]
      o._forward_delay = obj["forward-delay"]
      o._hold_count = obj["hold-count"]
      return o

  @property
  def priority (self):
    return self._priority

  @priority.setter
  def priority (self, val):
    self._priority = _intrange("rstp-priority", val, 0, 61440)

  @property
  def ageing_time (self):
    return self._ageing_time

  @ageing_time.setter
  def ageing_time (self, val):
    self._ageing_time = _intrange("rstp-ageing-time", val, 10, 1000000)

  @property
  def max_age (self):
    return self._max_age

  @max_age.setter
  def max_age (self, val):
    self._max_age = _intrange("rstp-max-age", val, 6, 40)

  @property
  def forward_delay (self):
    return self._forward_delay

  @forward_delay.setter
  def forward_delay (self, val):
    self._forward_delay = _intrange("rstp-forward-delay", val, 4, 30)

  @property
  def hold_count (self):
    return self._hold_count

  @hold_count.setter
  def hold_count (self, val):
    self._hold_count = _intrange("rstp-xmit-hold-count", val, 1, 10)

  @property
  def cmdlist (self):
    ocl = []
    if self.address: ocl.append("other_config:rstp-address=%s" % (self.address))
    if self.priority: ocl.append("other_config:rstp-priority=%d" % (self.priority))
    if self.ageing_time: ocl.append("other_config:rstp-ageing-time=%d" % (self.ageing_time))
    if self.max_age: ocl.append("other_config:rstp-max-age=%d" % (self.max_age))
    if self.forward_delay: ocl.append("other_config:rstp-forward-delay=%d" % (self.forward_delay))
    if self.hold_count: ocl.append("other_config:rstp-transmit-hold-count=%d" % (self.hold_count))
    return ocl

