# Copyright (c) 2015-2017  Barnstormer Softworks, Ltd.

import itertools

from ...core.log import KeyAdapter
from ..hostdriver import call
from ...core import json
from ...openflow.types import DPID

ALLSELECT = """'["Open_vSwitch",{"op":"select","table":"%(table)s","where":[],"columns":%(cols)s}]'"""
SELECT = """'["Open_vSwitch",{"op":"select","table":"%(table)s","where":[["%(wname)s","==","%(wval)s"]],"columns":%(cols)s}]'"""
SELECTUUID = """'["Open_vSwitch",{"op":"select","table":"%(table)s","where":[["_uuid","==",["uuid","%(uuid)s"]]],"columns":%(cols)s}]'"""

OVS_BASE = """'["Open_vSwitch",%s]'"""
SCLAUSE = """{"op":"select","table":"%(table)s","where":[["%(wname)s","==","%(wval)s"]],"columns":%(cols)s}"""
SCLAUSEUUID = """{"op":"select","table":"%(table)s","where":[["_uuid","==",["uuid","%(uuid)s"]]],"columns":%(cols)s}"""

def getField (data, fname):
  j = json.loads(data)
  return j[0]["rows"][0][fname]

def getUUIDs (data, fname):
  j = json.loads(data)
  slist = j[0]["rows"][0][fname][1]
  if j[0]["rows"][0][fname][0] == "uuid":
    return [slist] # If there's just one UUID, it's the raw value instead of a set
  return [x[1] for x in slist]

def getMultiUUIDs (objl, fname):
  retval = []
  for row in objl:
    rval = row["rows"][0][fname]
    if rval[0] == "uuid":
      retval.append(rval[1])
    elif rval[0] == "set":
      for val in rval[1]:
        retval.append(val[1])
  return retval

def convertMapDeep (data, fname):
  j = json.loads(data)
  obj = j[0]["rows"][0]
  return convertMap(obj, fname)

def convertMap (obj, fname, str_convert = False):
  slist = obj[fname][1]
  d = {}
  for x in slist:
    if str_convert:
      d[x[0]] = str(x[1])
    else:
      d[x[0]] = x[1]
  return d


class OVSDBInfoDriver(object):
  def __init__ (self, log, url = "unix:/var/run/openvswitch/db.sock"):
    self._url = url
    self._log = KeyAdapter("OVS-Info", log)
    self._cmdpath = "/usr/bin/ovsdb-client"

  @property
  def log (self):
    return self._log

  @log.setter
  def log (self, val):
    self._log = KeyAdapter("OVS-Info", val)

  @property
  def cmd (self):
    return "sudo %s transact %s" % (self._cmdpath, self._url)

  def getDPID (self, brname):
    expr = SELECT % {"wname" : "name", "wval" : brname, "table" : "Bridge", "cols" : json.dumps(["datapath_id"])}
    (sout, serr, retcode) = call("%s %s" % (self.cmd, expr), self.log.info)

    dpid = DPID(getField(sout, "datapath_id"))
    return dpid

  def getConnectionStatus (self, brname):
    expr = SELECT % {"wname" : "name", "wval" : brname, "table" : "Bridge", "cols" : json.dumps(["controller"])}
    (sout, serr, retcode) = call("%s %s" % (self.cmd, expr), self.log.info)
    cuuid = getField(sout, "controller")[1]

    expr = SELECTUUID % {"table" : "Controller", "uuid" : cuuid, "cols" : json.dumps(["is_connected"])}
    (sout, serr, retcode) = call("%s %s" % (self.cmd, expr), self.log.info)
    return getField(sout, "is_connected")
    
  def getOFPortNum (self, intf_name):
    expr = SELECT % {"wname" : "name", "wval" : intf_name, "table" : "Interface", "cols" : json.dumps(["ofport"])}
    (sout, serr, retcode) = call("%s %s" % (self.cmd, expr), self.log.info)
    return getField(sout, "ofport")

  def getBridges (self):
    expr = ALLSELECT % {"wname" : "name", "table" : "Bridge", "cols" : json.dumps(["name"])}
    (sout, serr, retcode) = call("%s %s" % (self.cmd, expr), self.log.info)
    j = json.loads(sout)

    names = []
    for obj in j[0]["rows"]:
      names.append(obj["name"])
    return names

  def getInterfaces (self, port_uuids):
    sdata = {"table":"Port", "cols" : json.dumps(["interfaces"])}
    clist = []
    for puuid in port_uuids:
      sdata["uuid"] = puuid
      clist.append(SCLAUSEUUID % (sdata))
    expr = OVS_BASE % (",".join(clist))
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)
    objlist = json.loads(o)
    return getMultiUUIDs(objlist, "interfaces")

  def getPorts (self, brname):
    expr = SELECT % {"wname" : "name", "wval" : brname, "table" : "Bridge", "cols" : json.dumps(["ports"])}
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)
    return getUUIDs(o, "ports")

  def getTableUUIDFields (self, table, uuid_list, fields):
    sdata = {"table" : table, "cols" : json.dumps(fields)}
    clist = []
    for uuid in uuid_list:
      sdata["uuid"] = uuid
      clist.append(SCLAUSEUUID % (sdata))
    expr = OVS_BASE % (",".join(clist))
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)

    objlist = json.loads(o)
    retl = []
    for obj in objlist:
      retl.append(obj["rows"][0])
    return retl

  def getPortInfo (self, brname):
    port_uuids = self.getPorts(brname)
    intf_uuids = self.getInterfaces(port_uuids)

    port_data = self.getTableUUIDFields("Port", port_uuids, ["name", "tag"])
    intf_data = self.getTableUUIDFields("Interface", intf_uuids,
                                        ["ifindex", "mtu", "link_state", "admin_state", "statistics"])

    if len(port_data) != len(intf_data):
      self.log.error("Bad shit happened")

    retlist = []
    for p,i in itertools.izip(port_data, intf_data):
      retobj = p
      retobj.update(i)
      retobj["statistics"] = convertMap(i, "statistics", True)
      retlist.append(retobj)
    return retlist
    
  def getRSTPInfo (self, brname):
    expr = SELECT % {"wname" : "name", "wval" : brname, "table" : "Bridge", "cols" : json.dumps(["ports", "rstp_status"])}
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)

    brstp_info = convertMapDeep(o, "rstp_status")
    port_uuids = getUUIDs(o, "ports")

    sdata = {"table":"Port", "cols" : json.dumps(["rstp_statistics","rstp_status","name"])}
    clist = []
    for puuid in port_uuids:
      sdata["uuid"] = puuid
      clist.append(SCLAUSEUUID % (sdata))
    expr = OVS_BASE % (",".join(clist))
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)

    objlist = json.loads(o)
    pinfo = []
    for obj in objlist:
      iobj = obj["rows"][0]
      info = {}
      info.update(convertMap(iobj, "rstp_statistics"))
      info.update(convertMap(iobj, "rstp_status"))
      pinfo.append({"name" : iobj["name"], "info" : info})

    ret = brstp_info
    ret["ports"] = pinfo
    return ret

  def getSTPInfo (self, brname):
    expr = SELECT % {"wname" : "name", "wval" : brname, "table" : "Bridge", "cols" : json.dumps(["ports", "status"])}
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)

    brstp_info = convertMapDeep(o, "status")
    port_uuids = getUUIDs(o, "ports")

    sdata = {"table":"Port", "cols" : json.dumps(["statistics","status","name"])}
    clist = []
    for puuid in port_uuids:
      sdata["uuid"] = puuid
      clist.append(SCLAUSEUUID % (sdata))
    expr = OVS_BASE % (",".join(clist))
    (o,e,r) = call("%s %s" % (self.cmd, expr), self.log.info)

    objlist = json.loads(o)
    pinfo = []
    for obj in objlist:
      iobj = obj["rows"][0]
      info = {}
      info.update(convertMap(iobj, "statistics"))
      info.update(convertMap(iobj, "status"))
      pinfo.append({"name" : iobj["name"], "info" : info})

    ret = brstp_info
    ret["ports"] = pinfo
    return ret

  def deleteBridge (self, name):
    cmd = "sudo ovs-vsctl del-br %s" % (name)
    (o,e,r) = call(cmd, self.log.info)

InfoDriver = OVSDBInfoDriver(None)
