# Copyright (c) 2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Driver for manipulating OVS topologies within their own namespaces

import sys
import os
import os.path

from ..hostdriver import call
from ..interfaces import VTSDriver
from ...core.log import KeyAdapter

NSDIR = "/var/run/netns"

TOOLBIN = "/usr/bin/ovsdb-tool"
SERVERBIN = "/usr/sbin/ovsdb-server"
SWITCHBIN = "/usr/sbin/ovs-vswitchd"
SCHEMAPATH = "/usr/share/openvswitch/vswitch.ovsschema"
VSCTLBIN = "/usr/bin/ovs-vsctl"

SERVERCMD = "%s %s -vconsole:emer -vsyslog:err -vfile:info --remote=punix:%s --private-key=db:Open_vSwitch,SSL,private_key --certificate=db:Open_vSwitch,SSL,certificate --bootstrap-ca-cert=db:Open_vSwitch,SSL,ca_cert --no-chdir --log-file=%s --pidfile=%s --detach --monitor"
SERVERCONF = "/etc/openvswitch/{ns-}conf.db"
SERVERSOCK = "/var/run/openvswitch/{ns-}db.sock"
SERVERLOG = "/var/log/openvswitch/{ns-}ovsdb-server.log"
SERVERPID = "/var/run/openvswitch/{ns-}ovsdb-server.pid"

SWITCHCMD = "%s unix:%s -vconsole:emer -vsyslog:err -vfile:info --mlockall --no-chdir --log-file=%s --pidfile=%s --detach --monitor"
SWITCHLOG = "/var/log/openvswitch/{ns-}ovs-vswitchd.log"
SWITCHPID = "/var/log/openvswitch/{ns-}ovs-vswitchd.pid"

class OVSNSDriver(VTSDriver):
  TESTRUN = False

  def __init__ (self, namespace, log):
    self.ns = str(namespace)
    self.sudo = "sudo"

    self._log = KeyAdapter("OVSNS %s" % (self.ns), log)

    if namespace == 0:
      self.nssudo = "sudo"
      self.vars = {"ns" : "0", "ns-" : ""}
      self.vsctl = VSCTLBIN
    else:
      self.nssudo = "sudo ip netns exec %s" % (self.ns)
      self.vars = {"ns" : self.ns, "ns-" : "%s-" % (self.ns)}
      self.vsctl = "%s --db=%s" % (VSCTLBIN, SERVERSOCK % (self.vars))
      if not os.path.exists("%s/%s" % (NSDIR, self.ns)):
        self._makeNamespace()

    if not os.path.exists(SERVERCONF.format(**self.vars)):
      self._makeDB()
  
  def destroy (self):
    serverpid = open(SERVERPID.format(**self.vars)).read().strip()
    self.call("%s kill -9 %s" % (self.sudo, serverpid))

    switchpid = open(SWITCHPID.format(**self.vars)).read().strip()
    self.call("%s kill -9 %s" % (self.sudo, switchpid))

  def call (self, cmd):
    if OVSNSDriver.TESTRUN:
      self._log.info(cmd)
    else:
      call(cmd, log_cmd=self._log.debug)

  def _makeNamespace (self):
    self.call("sudo ip netns add %s" % (self.ns))

  def _makeDB (self):
    self.call("%s %s create %s %s" % (self.sudo, TOOLBIN, SERVERCONF.format(**self.vars), SCHEMAPATH))

  def startDB (self):
    srvcmd = SERVERCMD % (SERVERBIN, SERVERCONF, SERVERSOCK, SERVERLOG, SERVERPID)
    self.call("%s %s" % (self.nssudo, srvcmd.format(**self.vars)))

  def startSwitches (self):
    switchcmd = SWITCHCMD % (SWITCHBIN, SERVERSOCK, SWITCHLOG, SWITCHPID)
    self.call("%s %s" % (self.nssudo, switchcmd.format(**self.vars)))
    
  def createBridge (self, dp):
    self.call("%s %s add-br %s" % (self.sudo, self.vsctl, dp.id))
    self.call("%s %s set bridge %s other-config:datapath-id=%s" % (self.sudo, self.vsctl, dp.id, dp.dpid.hexstr()))
    if dp.mactable_size:
      self.call("%s %s set bridge %s other-config:mac-table-size=%d" % (self.sudo, self.vsctl, dp.mactable_size))
    if dp.mactable_age:
      call("%s %s set bridge %s other-config:mac-aging-time=%d" % (self.sudo, self.vsctl, dp.mactable_age))
    self.setBridgeSTPMode(dp)

  def destroyBridge (self, name):
    pass

  def attachPort (self, bridge, port):
    pass

  def detachPort (self, bridge, port):
    pass

