# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import random
import time

from lxml import etree as ET
import psutil

from . import NetworkFunction, FuncReg 
from ..hostdriver import call
from .. import tasks
from ...geni.resources import Namespace
from ...core.configdb import ConfigDB, ConfigItem
from ...core.exception import CoreException
from ...core.log import KeyAdapter
from ...core import json

# TODO: Need to move GENI leaves to other files
VTSMNS = Namespace("vts", "http://geni.bssoftworks.com/rspec/ext/vts/manifest/1")

# TODO: Should be a config item
DATA_ROOT = "/opt/vts/functions/sslvpn"
OPENVPN = ""

SERVER_CFG = """dev %(dev)s
secret %(path)s/secret.key
port %(port)d
proto %(protocol)s
status-version 3
status %(path)s/status.log
log %(path)s/vpn.log
cipher AES-256-CBC
verb 4
keepalive 10 120
ping-timer-rem
"""

CLIENT_CFG = """
dev-type tap
dev %(pname)s
secret secret.key
remote %(rip)s %(rport)d
port %(localport)d
cipher AES-256-CBC
"""

L3SERVER_CFG = """
dev-type tun
dev %(pname)s
port %(port)d
proto udp
ca %(path)s/ca.crt
cert %(path)s/server.crt
key %(path)s/server.key
dh %(path)s/dh2048.pem
topology subnet
server %(network)s %(mask)s
ifconfig-pool-persist %(path)s/ipp.txt
push "route %(rnet)s %(rmask)s"
client-to-client
keepalive 10 120
cipher AES-256-CBC
comp-lzo
max-clients 50
status-version 3
status %(path)s/status.log
verb 4
log %(path)s/vpn.log
"""

class PortAlreadyUsedError(CoreException):
  def __init__ (self, port):
    super(PortAlreadyUsedError, self).__init__()
    self.port = port
  def __str__ (self):
    return "The requested port (%d) is already in use." % (self.port)


class _PortRegistry(object):
  def __init__ (self, start, end):
    self.low = start
    self.high = end
    self.used = set([])

  def isPortUsed (self, port):
    return port in self.used

  def reservePort (self, port):
    self.used.add(port)

  def allocatePort (self):
    p = random.randint(self.low, self.high)
    while p in self.used:
      p = random.randint(self.low, self.high)
    self.used.add(p)
    return p

  def releasePort (self, port):
    self.used.remove(port)


TPReg = _PortRegistry(20000, 21000)
CPReg = _PortRegistry(2000, 2500)

class L2SSLVPNServer(NetworkFunction):
  def __init__ (self):
    super(L2SSLVPNServer, self).__init__("sslvpn")
    self.port_name = None
    self.protocol = "udp"
    self.transport_port = None
    self.data_dir = None
    self.user_urn = None
    self.note = None
    self.sliver = None

  def __json__ (self):
    pd = super(L2SSLVPNServer, self).__json__()
    d = {"port-name" : self.port_name, "protocol" : self.protocol, "tp-port" : self.transport_port,
         "data-dir" : self.data_dir, "version" : 3, "user-urn" : self.user_urn, "note" : self.note}
    pd.update(d)
    return pd

  @classmethod
  def jconstruct (cls, obj):
    ver = json.getdefault(obj, "version", 1)

    o = cls()
    o.client_id = obj["client-id"]
    o.port_name = obj["port-name"]
    o.protocol = obj["protocol"]
    o.transport_port = obj["tp-port"]
    TPReg.reservePort(o.transport_port)
    o.data_dir = obj["data-dir"]

    if ver >= 2:
      o.user_urn = obj["user-urn"]
    if ver >= 3:
      o.note = obj["note"]
    return o

  @classmethod
  def fromdom (cls, elem):
    o = cls()
    o.protocol = elem.get("protocol", "udp")
    o.client_id = elem.get("client_id")
    return o

  @classmethod
  def fromdict (cls, obj):
    o = cls()
    o.protocol = obj["protocol"]
    o.note = json.getdefault(obj, "note", None)
    return o

  @property
  def key_data (self):
    f = open("%s/secret.key" % (self.data_dir))
    data = f.read()
    f.close()
    return data

  def userInfo (self):
    obj = {}
    obj["tp-port"] = self.transport_port
    obj["local-ip"] = ConfigDB.getConfigItemByKey("vts.sslvpn.local-ip").getValue()
    obj["protocol"] = self.protocol
    obj["key"] = self.key_data
    return obj

  def addManifest (self, root):
    fe = ET.SubElement(root, "{%s}function" % (VTSMNS.name))
    fe.attrib["client_id"] = self.client_id
    fe.attrib["type"] = "sslvpn"
    fe.attrib["tp-port"] = str(self.transport_port)
    fe.attrib["local-ip"] = ConfigDB.getConfigItemByKey("vts.sslvpn.local-ip").getValue()
    fe.attrib["protocol"] = self.protocol
    if self.user_urn:
      fe.attrib["user-urn"] = str(self.user_urn)
    if self.note:
      fe.attrib["note"] = unicode(self.note)
#    fe.text = ET.CDATA(self.key_data)

  def checkAndSetPort (self, port):
    if TPReg.isPortUsed(port):
      raise PortAlreadyUsedError(port)
    TPReg.reservePort(port)

  def provision (self, db, plog):
    log = KeyAdapter("sslvpn", plog)
    if not self.transport_port:
      self.transport_port = TPReg.allocatePort()
    if not self.client_id:
      self.client_id = "_vpn_%d" % (self.transport_port)
    self.port_name = "tap%d" % (self.transport_port)
    self.data_dir = "%s/%s" % (DATA_ROOT, self.port_name)
    kpath = "%s/secret.key" % (self.data_dir)

    os.makedirs(self.data_dir)
    call("%s --genkey --secret %s" % (OPENVPN, kpath), log_cmd = log.info)
    call("sudo ip tuntap add dev %s mode tap" % (self.port_name), log_cmd = log.info, log_output = log.info)

    cfg_data = SERVER_CFG % ({"dev" : self.port_name, "path" : self.data_dir, "port" : self.transport_port,
                              "protocol" : self.protocol})
    cfgf = open("%s/server.cfg" % (self.data_dir), "w+")
    cfgf.write(cfg_data)
    cfgf.close()

  def start (self, plog):
    log = KeyAdapter("sslvpn", plog)
    call("sudo %s --config %s/server.cfg --daemon" % (OPENVPN, self.data_dir), log_cmd = log.info, log_output=log.info)
    call("sudo ip link set %s up" % (self.port_name), log_cmd = log.info, log_output=log.info)
    
  def deprovision (self, db, plog):
    try:
      for p in psutil.process_iter():
        if p.name() == "openvpn":
          if p.cmdline()[2].count(self.port_name):
            call("sudo kill -9 %d" % (p.pid), log_cmd = plog.info, log_output=plog.info)

      call("sudo ip link delete %s" % (self.port_name), log_cmd = plog.info, log_output=plog.info)
    except TypeError:
      pass # If you really screw things up there's no port_name

    try:
      TPReg.releasePort(self.transport_port)
    except KeyError:
      pass  # Wasn't really here
    tasks.destroyTree(self.data_dir)


class L3SSLVPNServer(NetworkFunction):
  def __init__ (self):
    super(L3SSLVPNServer, self).__init__("l3vpn-server")
    self.port_name = None
    self.protocol = "udp"
    self.transport_port = None
    self.data_dir = None
    self.sliver = None
    self.client_net = None # (network, mask)
    self.routes = [] # [(network, mask), ...]

  def __json__ (self):
    pd = super(L3SSLVPNServer, self).__json__()
    d = {"port-name" : self.port_name, "protocol" : self.protocol, "tp-port" : self.transport_port,
         "data-dir" : self.data_dir, "version" : 1, "client-net" : self.client_net, "routes" : self.routes}
    pd.update(d)
    return pd

  @classmethod
  def fromdom (cls, elem):
    o = cls()
    return o

  @classmethod
  def jconstruct (cls, obj):
    ver = obj["version"]

    o = cls()
    return o


class L2SSLVPNClient(NetworkFunction):
  def __init__ (self):
    super(L2SSLVPNClient, self).__init__("sslvpn-client")
    self.port_name = None
    self.protocol = "udp"
    self.local_port = None
    self.remote_ip = None
    self.remote_port = None
    self.data_dir = None
    self.note = None

    # Temporary storage pre-provision
    self.__pkey = None

  def __json__ (self):
    pd = super(L2SSLVPNClient, self).__json__()
    d = {"port-name" : self.port_name, "protocol" : self.protocol, "local-port" : self.local_port,
         "data-dir" : self.data_dir, "version" : 1, "note" : self.note, "remote-ip" : self.remote_ip,
         "remote-port" : self.remote_port}
    pd.update(d)
    return pd

  @classmethod
  def jconstruct (cls, obj):
    ver = obj["version"]

    o = cls()
    o.client_id = obj["client-id"]
    o.port_name = obj["port-name"]
    o.protocol = obj["protocol"]
    o.local_port = obj["local-port"]
    o.remote_ip = obj["remote-ip"]
    o.remote_port = obj["remote-port"]
    CPReg.reservePort(o.local_port)
    o.data_dir = obj["data-dir"]
    o.note = obj["note"]

    return o

  @classmethod
  def fromdom (cls, elem):
    o = cls()
    o.client_id = elem.get("client_id")
    o.remote_ip = elem.get("remote-ip")
    o.remote_port = int(elem.get("remote-port"))
    o.note = elem.get("note")
    o.__pkey = elem.text.strip()
    return o

  def addManifest (self, root):
    fe = ET.SubElement(root, "{%s}function" % (VTSMNS))
    fe.attrib["client_id"] = self.client_id
    fe.attrib["type"] = "sslvpn-client"
    fe.attrib["remote-ip"] = str(self.remote_ip)
    fe.attrib["remote-port"] = str(self.remote_port)
    fe.attrib["tp-port"] = str(self.local_port)
    fe.attrib["protocol"] = self.protocol
    fe.attrib["note"] = str(self.note)

  def checkAndSetPort (self, port):
    if CPReg.isPortUsed(port):
      raise PortAlreadyUsedError(port)
    CPReg.reservePort(port)

  def provision (self, db, plog):
    log = KeyAdapter("sslvpn-client", plog)
    if not self.local_port:
      self.local_port = CPReg.allocatePort()
    if not self.client_id:
      self.client_id = "_vpnc_%d" % (self.local_port)

    self.port_name = "ctap%d" % (self.local_port)
    self.data_dir = "%s/%s" % (DATA_ROOT, self.port_name)
    kpath = "%s/secret.key" % (self.data_dir)

    os.makedirs(self.data_dir)

    keyf = open(kpath, "w+")
    keyf.write(self.__pkey)
    keyf.close()
    self.__pkey = None

    call("sudo ip tuntap add dev %s mode tap" % (self.port_name), log_cmd = log.info, log_output = log.info)
    cfg_data = CLIENT_CFG % ({"pname" : self.port_name, "rip" : self.remote_ip, "rport" : self.remote_port,
                              "localport" : self.local_port})

    cfgf = open("%s/client.cfg" % (self.data_dir), "w+")
    cfgf.write(cfg_data)
    cfgf.close()

  def start (self, plog):
    log = KeyAdapter("sslvpn-client", plog)
    call("sudo %(openvpn)s --daemon --status %(data_dir)s/slice.status 30 --cd %(data_dir)s --config %(data_dir)s/client.cfg --script-security 2" % (
          {"data_dir" : self.data_dir, "openvpn" : OPENVPN}), log_cmd = log.info, log_output = log.info)
    call("sudo ip link set %s up" % (self.port_name), log_cmd = log.info, log_output = log.info)

  def deprovision (self, db, plog):
    for p in psutil.process_iter():
      if p.name() == "openvpn":
        if p.cmdline()[3].count(self.port_name):
          call("sudo kill -9 %d" % (p.pid), log_cmd = plog.info, log_output=plog.info)

    call("sudo ip link delete %s" % (self.port_name), log_cmd = plog.info, log_output=plog.info)
    try:
      CPReg.releasePort(self.local_port)
    except KeyError:
      pass  # Wasn't really here
    tasks.destroyTree(self.data_dir)


def setup (app):
  global OPENVPN

  citems = []
  citems.append(ConfigItem().setKey("vts.sslvpn.local-ip").setValue(None)
                .setDesc("Local IP endpoint for SSL VPN functions"))
  citems.append(ConfigItem().setKey("vts.sslvpn.bin").setValue("/usr/sbin/openvpn")
                .setDesc("Path to OpenVPN binary"))
  [ConfigDB.installConfigItem(x) for x in citems]

  OPENVPN = ConfigDB.getConfigItemByKey("vts.sslvpn.bin").getValue()

  FuncReg.registerType("sslvpn", L2SSLVPNServer.jconstruct, L2SSLVPNServer.fromdom)
  FuncReg.registerType("sslvpn-client", L2SSLVPNClient.jconstruct, L2SSLVPNClient.fromdom)
  FuncReg.registerType("l3vpn-server", L3SSLVPNServer.jconstruct, L3SSLVPNServer.fromdom)
