# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import base64
import json

from lxml import etree as ET
import IPy

from ..geni.resources import Registry as ResourceRegistry
from ..geni.version import Registry as APIRegistry
from ..geni.resources import Namespace, UnknownSliverError, UnknownSliceError, SkipHandler
from ..geni.lib import getMaxExpiration, DuplicateSliver
from ..core.configdb import ConfigDB, ConfigItem
from ..core.log import KeyAdapter
from ..core.exception import CoreException
from .data import (DB, Datapath, VTSSliver, SFlowInfo, Container, NetFlowInfo, UnknownImageAttributeError,
                   RequiredAttributeError, UnknownDatapathError, UnknownPortError)
from . import ports as Ports
from .functions import FuncReg as FR
from .hostdriver import HostDriver
from .drivers.docker import CM
from .drivers.dockerhgvols import HGVolumeManager
from .drivers.dockerdropboxvols import DropboxVolumeManager
from .drivers import ovs as OVS


GENINS = Namespace("geni", "http://www.geni.net/resources/rspec/3", "http://www.geni.net/resources/rspec/3/request.xsd")
GENIUSERNS = Namespace("user", "http://www.geni.net/resources/rspec/ext/user/1")
VTSNS = Namespace("vts", "http://geni.bssoftworks.com/rspec/ext/vts/ad/1")
VTSRNS = Namespace("vts", "http://geni.bssoftworks.com/rspec/ext/vts/request/1")
VTSMNS = Namespace("vts", "http://geni.bssoftworks.com/rspec/ext/vts/manifest/1")
SDNRNS = Namespace("sdn", "http://geni.bssoftworks.com/rspec/ext/sdn/request/1")
SDNMNS = Namespace("sdn", "http://geni.bssoftworks.com/rspec/ext/sdn/manifest/1")

XPNS = {'g' : GENINS.name, 'v' : VTSRNS.name, 's' : SDNRNS.name}

class AmbiguousUUIDError(CoreException):
  def __str__ (self):
    return "Sliver UUID ambiguous for multi-sliver slice"

class POAValidationError(CoreException):
  def __init__ (self, poa, errors):
    super(POAValidationError, self).__init__()
    self.poa = poa
    self.errors = errors
  def __str__ (self):
    return "%s input validation failed:\n%s" % (self.poa, "\n".join(self.errors))

class BadPortVLANFormatError(CoreException):
  def __init__ (self, val):
    super(BadPortVLANFormatError, self).__init__()
    self._val = val
  def __str__ (self):
    return "(%s) is not a proper (port,vlan-id) pair" % (self._val)

class IllegalVlanIDError(CoreException):
  def __init__ (self, vid):
    super(IllegalVlanIDError, self).__init__()
    self.vid = vid
  def __str__ (self):
    return "The requested VLAN (%d) is not a legal value" % (self.vid)

class DuplicateSliceClientIDError(CoreException):
  def __init__ (self, cid_set):
    super(DuplicateSliceClientIDError, self).__init__()
    self.cid_set = cid_set
  def __str__ (self):
    return "The requested client IDs (%s) are already used in this slice at this aggregate" % (", ".join([x for x in self.cid_set]))

def coerceNoneStr (val, noneval="Unknown"):
  if val is None:
    return noneval
  else:
    return str(val)

def addConstraint (parent, key, value):
  cse = ET.SubElement(parent, "{%s}constraint" % (VTSNS.name))
  cse.attrib["key"] = key
  cse.attrib["value"] = value

def generateGENIAd (root):
#  vlans = ET.SubElement(root, "{%s}local-circuits" % (VTSNS.name))
#  vlans.attrib["total"] = str(len(DB.vlans))
#  vlans.attrib["avail"] = str(len(DB.vlans) - len(DB.usedvlans))

  planes = ET.SubElement(root, "{%s}circuit-planes" % (VTSNS.name))
  for typ,cpdict in DB.circuit_planes.iteritems():
    for cp in cpdict.itervalues():
      cpe = ET.SubElement(planes, "{%s}circuit-plane" % (VTSNS.name))
      cpe.attrib["type"] = typ
      if typ == "ipv4":
        cpe.attrib["label"] = cp.label
        cste = ET.SubElement(cpe, "{%s}constraints" % (VTSNS.name))
        addConstraint(cste, "mtu", str(cp.mtu))
        addConstraint(cste, "multipoint", "false")
        addConstraint(cste, "transparent", "true")
        addConstraint(cste, "mac-learning", "false")

        epe = ET.SubElement(cpe, "{%s}endpoint" % (VTSNS.name))
        if cp.encoded:
          d = {"ip" : str(cp.endpoint), "key" : None}
          epe.attrib["value"] = base64.b32encode(json.dumps(d))
        else:
          epe.attrib["value"] = str(cp.endpoint)
        etle = ET.SubElement(cpe, "{%s}supported-tunnels" % (VTSNS.name))
        for tt in cp.supported_types:
          tte = ET.SubElement(etle, "{%s}tunnel-type" % (VTSNS.name))
          tte.attrib["name"] = tt

  # Stub in local if we have one
  if DB._pgcmid:
    cpe = ET.SubElement(planes, "{%s}circuit-plane" % (VTSNS.name))
    cpe.attrib["type"] = "l2-eth"
    cpe.attrib["label"] = "pg-local"
    cste = ET.SubElement(cpe, "{%s}constraints" % (VTSNS.name))
    addConstraint(cste, "mtu", "7000")
    addConstraint(cste, "multipoint", "true")
    addConstraint(cste, "transparent", "true")
    addConstraint(cste, "mac-learning", "true")
    addConstraint(cste, "local-cmid", DB._pgcmid)
    addConstraint(cste, "circuits.max", str(len(DB.vlans)))
    addConstraint(cste, "circuits.available", str(len(DB.vlans) - len(DB.usedvlans)))

    etle = ET.SubElement(cpe, "{%s}supported-tunnels" % (VTSNS.name))
    tte = ET.SubElement(etle, "{%s}tunnel-type" % (VTSNS.name))
    tte.attrib["name"] = "vlan"
  

  images = ET.SubElement(root, "{%s}images" % (VTSNS.name))
  for image in DB.images:
    i = ET.SubElement(images, "{%s}image" % (VTSNS.name))
    i.attrib["name"] = image
    i.attrib["type"] = "raw"

  for imgdesc in CM.imagedescs:
    if imgdesc.hidden:
      continue
    i = ET.SubElement(images, "{%s}image" % (VTSNS.name))
    i.attrib["name"] = imgdesc.name
    i.attrib["type"] = "container"
    i.attrib["mem"] = str(imgdesc.mem_default)
    i.attrib["mem-min"] = str(imgdesc.mem_min)
    i.attrib["mem-max"] = str(imgdesc.mem_max)
    if imgdesc.image_attributes:
      attrse = ET.SubElement(i, "{%s}image-attributes" % (VTSNS))
      for attr in imgdesc.image_attributes.values():
        a = ET.SubElement(attrse, "{%s}image-attribute" % (VTSNS))
        a.attrib["name"] = attr.name
        a.attrib["type"] = attr.type
        if attr.default:
          a.attrib["default"] = str(attr.default)
        a.attrib["optional"] = str(attr.optional)

  dps = ET.SubElement(root, "{%s}datapaths" % (VTSNS.name))
  for uiname in DB.target_bridges.keys():
    t = ET.SubElement(dps, "{%s}datapath" % (VTSNS.name))
    t.attrib["name"] = uiname
  for name,ptb in DB.public_bridges.items():
    t = ET.SubElement(dps, "{%s}datapath" % (VTSNS))
    t.attrib["name"] = name
    t.attrib["desc"] = ptb.public_desc


def generateManifest (sliver, root):
  log = KeyAdapter(sliver.uuid, DB._log)
  log.info("Generating manifest for sliver %s" % (sliver.urn))

  infoe = ET.SubElement(root, "{%s}info" % (VTSMNS))
  infoe.attrib["host"] = ConfigDB.getConfigItemByKey("vts.geni.cm-tag").getValue()
  infoe.attrib["slice"] = sliver.slice_urn
  if sliver._topo_name:
    infoe.attrib["client-topo-name"] = sliver._topo_name

  for dp in sliver.datapaths:
    dpe = ET.SubElement(root, "{%s}datapath" % (VTSMNS.name))
    dpe.attrib["client_id"] = dp.client_id
    dpe.attrib["image"] = dp.image
    dpe.attrib["sliver_id"] = sliver.urn
    if dp.controllers:
      dpe.attrib["ofver"] = dp.ofver

    if dp.mactable_size:
      dpe.attrib["mac-table-size"] = str(dp.mactable_size)
    if dp.mactable_age:
      dpe.attrib["mac-aging-time"] = str(dp.mactable_age)

    # TODO: Add all the controllers back
    if dp.controllers:
      ce = ET.SubElement(dpe, "{%s}controller" % (SDNMNS.name))
      ce.attrib["url"] = dp.controllers[0]
      dpide = ET.SubElement(dpe, "{%s}openflow-dpid" % (VTSMNS.name))
      dpide.attrib["value"] = str(dp.dpid)

    stpe = ET.SubElement(dpe, "{%s}stp" % (VTSMNS))
    if isinstance(dp.stpinfo, OVS.STPParams):
      stpe.attrib["type"] = "stp"
    elif isinstance(dp.stpinfo, OVS.RSTPParams):
      stpe.attrib["type"] = "rstp"
    elif isinstance(dp.stpinfo, OVS.DisabledSTP):
      stpe.attrib["type"] = "disabled"

    # TODO: SFlow info
    # TODO: NetFlow info

    if dp.mirror:
      ce = ET.SubElement(dpe, "{%s}mirror" % (VTSMNS))
      ce.attrib["target"] = dp.mirror[0]

    for port in dp.ports.itervalues():
      pe = ET.SubElement(dpe, "{%s}port" % (VTSMNS.name))
      pe.attrib["client_id"] = port.client_id
      pe.attrib["type"] = port.type
      if port.type == "internal":
        pe.attrib["remote-clientid"] = port.target
        if port.vlan:
          pe.attrib["vlan-id"] = str(port.vlan)
      elif port.type == "vf-port":
        pe.attrib["remote-clientid"] = port.target
      elif port.type == "pglocal":
        pe.attrib["shared-lan"] = port.circuit
        pe.attrib["type"] = "pg-local"  # Inconsistent mapping between internal structure and API
      elif port.type == "gre":
        epe = ET.SubElement(pe, "{%s}endpoint" % (VTSMNS.name))
        epe.attrib["circuit-plane"] = port.circuit_plane.label

        led = {"ip" : str(port.circuit_plane.endpoint), "key" : port.key}
        red = {"ip" : str(port.endpoint_addr), "key" : port.key}
        if port.circuit_plane.encoded:
          epe.attrib["local"] = base64.b32encode(json.dumps(led))
          epe.attrib["remote"] = base64.b32encode(json.dumps(red))
        else:
          epe.attrib["local-ip"] = str(port.circuit_plane.endpoint)
          epe.attrib["key"] = port.key
          epe.attrib["remote-ip"] = str(port.endpoint_addr)
      try:
        if port.delay_info:
          de = ET.SubElement(pe, "{%s}egress-delay" % (VTSMNS.name))
          if port.delay_info.time: de.attrib["time"] = str(port.delay_info.time)
          if port.delay_info.jitter: de.attrib["jitter"] = str(port.delay_info.jitter)
          if port.delay_info.correlation: de.attrib["correlation"] = str(port.delay_info.correlation)
          if port.delay_info.distribution: de.attrib["distribution"] = port.delay_info.distribution
      except AttributeError:
        pass
      try:
        if port.loss_info:
          le = ET.SubElement(pe, "{%s}egress-loss" % (VTSMNS.name))
          le.attrib["percent"] = str(port.loss_info.percent)
      except AttributeError:
        pass
      try:
        if port.reorder_info:
          le = ET.SubElement(pe, "{%s}egress-reorder" % (VTSMNS))
          le.attrib["percent"] = str(port.reorder_info.percent)
          le.attrib["correlation"] = str(port.reorder_info.correlation)
          le.attrib["gap"] = str(port.reorder_info.gap)
      except AttributeError:
        pass
   
  froot = ET.SubElement(root, "{%s}functions" % (VTSMNS.name))
  for func in sliver.functions:
    func.addManifest(froot)

  for ct in sliver.containers:
    cte = ET.SubElement(root, "{%s}container" % (VTSMNS.name))
    cte.attrib["client_id"] = ct.client_id
    cte.attrib["image"] = ct._image_desc.name
    cte.attrib["sliver_id"] = sliver.urn
    cte.attrib["ram"] = str(ct.ram)

    if ct._console_uname:
      se = ET.SubElement(cte, "{%s}services" % (GENINS.name))
      ue = ET.SubElement(se, "{%s}login" % (GENINS.name))
      ue.attrib["authentication"] = "ssh-keys"
      ue.attrib["username"] = ct._console_uname
      ue.attrib["port"] = "22"
      ue.attrib["hostname"] = "%s" % (ConfigDB.getConfigItemByKey("vts.geni.cm-tag").getValue())
      pkl = ET.SubElement(se, "{%s}services_user" % (GENIUSERNS))
      pkl.attrib["login"] = ct._console_uname
      for key in ct.ssh_keys:
        pke = ET.SubElement(pkl, "{%s}public_key" % (GENIUSERNS))
        pke.text = key

    for volinfo in ct._hgvols:
      me = ET.SubElement(cte, "{%s}mount" % (VTSMNS.name))
      me.attrib["type"] = "hg"
      me.attrib["vol-id"] = str(volinfo.volid)
      me.attrib["path"] = volinfo.container_mount
      me.attrib["name"] = volinfo.name

    for volinfo in ct._dbxvols:
      me = ET.SubElement(cte, "{%s}mount" % (VTSMNS.name))
      me.attrib["type"] = "dropbox"
      me.attrib["vol-id"] = str(volinfo.volid)
      me.attrib["name"] = volinfo.name
      me.attrib["path"] = volinfo.container_mount

    for port in ct.ports.itervalues():
      pe = ET.SubElement(cte, "{%s}port" % (VTSMNS.name))
      pe.attrib["client_id"] = port.client_id
      if port.type == "internal-container":
        pe.attrib["type"] = "internal"
        pe.attrib["remote-clientid"] = port.target
        pe.attrib["mac-address"] = coerceNoneStr(port.macaddr)
        pe.attrib["name"] = coerceNoneStr(port.alias)
        if port.delay_info:
          de = ET.SubElement(pe, "{%s}egress-delay" % (VTSMNS.name))
          if port.delay_info.time: de.attrib["time"] = str(port.delay_info.time)
          if port.delay_info.jitter: de.attrib["jitter"] = str(port.delay_info.jitter)
          if port.delay_info.correlation: de.attrib["correlation"] = str(port.delay_info.correlation)
          if port.delay_info.distribution: de.attrib["distribution"] = port.delay_info.distribution
      

def genSliceManifest (slice_urn, root):
  try:
    slivers = DB.getSliversBySlice(slice_urn)
    for sliver in slivers:
      generateManifest(sliver, root)
    return True 
  except KeyError:
    # TODO: Separate no slice from other errors
    return None 


def deleteSlice (slice_urn, user_info):
  try:
    slivers = DB.getSliversBySlice(slice_urn)
    for sliver in slivers:
      ResourceRegistry.unregisterSliver(slice_urn, sliver.urn)
    DB.deleteSlice(slice_urn)
    return True
  except KeyError:
    # TODO: Separate no slice from other errors
    return None

def deleteSliver (slice_urn, sliver_urns, user_info):
  outlist = []
  for urn in sliver_urns:
    try:
      sliver = DB.getSliver(urn)
      outlist.append((urn, sliver.expiration, None))
      ResourceRegistry.unregisterSliver(slice_urn, sliver.urn)
      sliver.deprovision(DB)
      DB.removeSliver(sliver)
    except KeyError:
      outlist.append((urn, None, "Unknown Sliver"))

  proxy = SliverBundleProxy(DB.getSliversBySlice(slice_urn))
  proxy.cleanInterconnects()

  return outlist


def renewSliver (slice_urn, reqexp):
  try:
    slivers = DB.getSliversBySlice(slice_urn)
    for sliver in slivers:
      DB.renewSliver(sliver, reqexp)
    return True
  except KeyError:
    return None

def sliverStatus (res, urns, options):
  import foam.vts.drivers.ovsdb as ovs
  db = ovs.InfoDriver
  db.log = DB._log

  res["vts_datapaths"] = []
  res["vts_containers"] = []

  for urn in urns:
    sliver = DB.getSliver(urn)

    for dp in sliver.datapaths:
      d = {"client-id" : dp.client_id,
           "dpid" : str(dp.dpid),
           "ports" : []}
      for port in dp.ports.itervalues():
        d["ports"].append({"client-id" : port.client_id, "name" : port.name,
                           "number" : port.ofportnum, "up" : port.up})

      try:
       if dp.controllers:
          d["connected"] = db.getConnectionStatus(dp.id)
      except:
        pass
      res["vts_datapaths"].append(d)

    for ct in sliver.containers:
      d = {"client-id" : ct.client_id,
           "ports" : []}
      for port in ct.ports.itervalues():
        d["ports"].append({"client-id" : port.client_id, "name" : port.alias,
                           "mac-address" : port.macaddr, "ipv4-addresses" : port.ipv4addrs})
      res["vts_containers"].append(d)


def parseRequest (slice_urn, credentials, rdom, user_info, mdom):
  bundle = None

  if DB.sliceExists(slice_urn):
    if user_info['_api'] != 3:
      raise DuplicateSliver(slice_urn)
    else:
      bundle = SliverBundleProxy(DB.getSliversBySlice(slice_urn))

  exp = getMaxExpiration(ConfigDB.getConfigItemByKey("vts.initial-lease").getValue(), credentials)

  sliver = VTSSliver(slice_urn, user_info["urn"])
  sliver.rspec = ET.tostring(rdom)
  sliver.login_users = user_info["login_users"]

  # TODO: We need to preflight target ports or something here, to avoid ending up in a situation
  # where we can't provision them in the next step

  # Grab local circuits we're going to need
  # TODO: We need to return these if something goes awry later
  # Or have some kind of periodic circuit garbage collection
  pgports = rdom.xpath("/g:rspec/v:datapath/v:port[@type='pg-local']", namespaces = XPNS)
  circuitpool = DB.reserveLocalCircuits(len(pgports))

  cinfo = rdom.findall("{%s}client-info" % (VTSRNS))
  if cinfo:
    sliver._topo_name = cinfo[0].get("topo-name")

  datapaths = rdom.findall("{%s}datapath" % (VTSRNS.name))
  for dpelem in datapaths:
    dp = Datapath()
    dp.client_id = dpelem.get("client_id")
    image_handler = rspecImageHandlerFactory(dpelem.xpath('v:image', namespaces = XPNS)[0], dp)
    image_handler.log = sliver.log
    dp.image = image_handler.image
    # TODO: We need to support some arbitrary number of local circuit pools
    image_handler.circuitpool = circuitpool
    ports = dpelem.xpath('v:port', namespaces = XPNS)
    for pelem in ports:
      dp.addPort(image_handler.buildPort(pelem))
    sliver.addDatapath(dp)

  containers = rdom.findall("{%s}container" % (VTSRNS.name))
  try:
    for celem in containers:
      ct = DockerXmlHandler(sliver, celem)
      if ct:
        sliver.addContainer(ct)
  except (UnknownImageAttributeError, RequiredAttributeError,
          CM.UnknownImagespecError) as e:
    e._unexpected = False
    raise

  functions = rdom.findall("{%s}function" % (VTSRNS.name))
  for funcelem in functions:
    rbuilder = FR.getRSpecBuilder(funcelem.get("type"))
    func = rbuilder(funcelem)
    sliver.addFunction(func)

  if bundle:
    bad_cids = sliver.client_ids & bundle.client_ids
    if bad_cids:
      sliver.log.debug("Overlapping client IDs: %s" % (str(bad_cids)))
      raise DuplicateSliceClientIDError(bad_cids)

  DB.insertSliver(sliver)
  ResourceRegistry.registerSliver(slice_urn, sliver.urn)


def provisionSliver (slice_urn, credentials, options):
  # TODO: We need to not block on this - complex topologies take a while to build
  slivers = DB.getSliversBySlice(slice_urn)
  bundle = SliverBundleProxy(slivers)
  for sliver in slivers:
    if not sliver.login_users:
      sliver.login_users = options["user-info"]["login_users"]
      sliver.log.info("Adding login users")
    sliver.provision(bundle)


# TODO: This should create an OVSImage / OVSInstance object from the XML description
# Right now we just make the One True Datapath object (which is OVS)
class OVSXmlHandler(object):
  def __init__ (self, elem, dp):
    self._elem = elem
    self._dp = dp
    self._ver = "1.0"
    self.circuitpool = []
    self.log = None
    if self._elem is not None:
      self._build()

  def _build (self):
    # TODO: If you supply certain elements too many times, we just ignore it
    # rather than telling you we're not using the extra ones
    self.image = self._elem.get("name")

    sfelems = self._elem.xpath('v:sflow', namespaces = XPNS)
    if sfelems:
      self._parse_sflow(sfelems[0])

    nfelems = self._elem.xpath('v:netflow', namespaces = XPNS)
    if nfelems:
      self._parse_netflow(nfelems[0])

    mrelems = self._elem.xpath('v:mirror', namespaces = XPNS)
    if mrelems:
      self._parse_mirror(mrelems[0])

    velems = self._elem.xpath('v:openflow-version', namespaces = XPNS)
    if velems:
      self._parse_version(velems[0])
    self._dp.ofver = self._ver

    idelems = self._elem.xpath('v:openflow-dpid', namespaces = XPNS)
    if idelems:
      self._dp.dpid = idelems[0].get("value")

    stpelems = self._elem.xpath('v:stp', namespaces = XPNS)
    if stpelems:
      self._parse_stp(stpelems[0])

    macelems = self._elem.xpath('v:mac-table-params', namespaces = XPNS)
    if macelems:
      self._parse_mac_params(macelems[0])

    # This resets STP, so it has to come after
    celems = self._elem.xpath('s:controller', namespaces = XPNS)
    if celems:
      self._parse_controllers(celems)

  def _parse_mac_params (self, elem):
    for child in elem:
      if child.tag == "{%s}max-size" % (VTSRNS):
        self._dp.mactable_size = int(child.get("value"))
      elif child.tag == "{%s}max-age" % (VTSRNS):
        self._dp.mactable_age = int(child.get("value"))

  def _parse_stp (self, elem):
    if elem.get("type") == "stp":
      params = OVS.STPParams()
      self._dp.stpinfo = params
      for child in elem:
        if child.tag == "{%s}system-id" % (VTSRNS):
          params.system_id = child.get("value")
        elif child.tag == "{%s}priority" % (VTSRNS):
          params.priority = int(child.get("value"))
        elif child.tag == "{%s}hello-time" % (VTSRNS):
          params.hello_time = int(child.get("value"))
        elif child.tag == "{%s}max-age" % (VTSRNS):
          params.max_age = int(child.get("value"))
        elif child.tag == "{%s}forward-delay" % (VTSRNS):
          params.forward_delay = int(child.get("value"))
    elif elem.get("type") == "rstp":
      params = OVS.RSTPParams()
      self._dp.stpinfo = params
      for child in elem:
        if child.tag == "{%s}address" % (VTSRNS):
          params.address = child.get("value")
        elif child.tag == "{%s}priority" % (VTSRNS):
          params.priority = int(child.get("value"))
        elif child.tag == "{%s}ageing-time" % (VTSRNS):
          params.ageing_time = int(child.get("value"))
        elif child.tag == "{%s}max-age" % (VTSRNS):
          params.max_age = int(child.get("value"))
        elif child.tag == "{%s}forward-delay" % (VTSRNS):
          params.forward_delay = int(child.get("value"))
        elif child.tag == "{%s}xmit-hold-count" % (VTSRNS):
          params.hold_count = int(child.get("value"))
    elif elem.get("type") == "disabled":
      self._dp.stpinfo = OVS.DisabledSTP()

  def _parse_sflow (self, sfe):
    c,p = sfe.get("collector").split(":")

    info = SFlowInfo()
    info.collector_ip = c
    info.collector_port = int(p)
    info.header_bytes = int(sfe.get("header-bytes"))
    info.sampling_n = int(sfe.get("sampling-n"))
    info.polling_secs = int(sfe.get("polling-secs"))
    self._dp.sflow_info = info

  def _parse_netflow (self, nfe):
    c,p = nfe.get("collector").split(":")

    info = NetFlowInfo()
    info.collector_ip = c
    info.collector_port = int(p)
    info.timeout = int(nfe.get("timeout"))
    self._dp.netflow_info = info

  def _parse_mirror (self, mre):
    self._dp.mirror = (mre.get("target"), DB.MNR.getName())

  def _parse_controllers (self, celems):
    self._dp.stpinfo = OVS.DisabledSTP()
    for celem in celems:
      self._dp.controllers.append(celem.get("url"))

  def _parse_version (self, velem):
    v = velem.get("value")
    if v in set(["1.3", "1.2", "1.0"]):
      self._ver = v

  def buildPort (self, pelem):
    typ = pelem.get("type")
    if typ == "internal":
      p = Ports.Internal(self._dp) 
      p.client_id = pelem.get("client_id")
      p.vlan = pelem.get("vlan-id")
      telem = pelem.xpath('v:target', namespaces = XPNS)[0]
      p.target = telem.get("remote-clientid")
      sup_vlans = telem.get("tagged-vlans")
      if sup_vlans:
        spec_list = sup_vlans.strip().split(",")
        for spec in spec_list:
          p._supported_vlans = Ports.VLANRange(spec)

    elif typ == "pg-local":
      p = Ports.PGLocal(self._dp)
      p.client_id = pelem.get("client_id")
      p.circuit = self.circuitpool.pop()
    elif typ == "gre":
      p = Ports.GRE(self._dp)
      p.client_id = pelem.get("client_id")
      p.circuit_plane = DB.getCircuitPlane(pelem.get("circuit-plane"))
      endpoint_info = pelem.get("endpoint")
      if p.circuit_plane.encoded:
        # Encoded info is {ip : xx, key : xx}
        epobj = json.loads(base64.b32decode(endpoint_info))
        p.endpoint_addr = IPy.IP(epobj["ip"])
        if epobj["key"]:
          p.key = epobj["key"]
          p.circuit_plane.reserveGREKey(p.endpoint_addr, p.key)
      else:
        # Unencoded just has IP and we allocate key here
        p.endpoint_addr = IPy.IP(endpoint_info)
    elif typ == "vf-port":
      p = Ports.VFPort(self._dp)
      p.client_id = pelem.get("client_id")
      telem = pelem.xpath('v:target', namespaces = XPNS)[0]
      p.target = telem.get("remote-clientid")

    # One place for loss and delay handling
    if typ in set(["internal", "pg-local"]):
      try:
        edelem = pelem.xpath('v:egress-delay', namespaces = XPNS)[0]
        p.setDelayInfo(edelem.get("time"), edelem.get("jitter"), edelem.get("correlation"), edelem.get("distribution"))
      except (TypeError, IndexError) as e:
        pass

      try:
        lelem = pelem.xpath('v:egress-loss', namespaces = XPNS)[0]
        linfo = Ports.LossInfo()
        linfo.percent = lelem.get("percent")
        p.loss_info = linfo
      except (TypeError, IndexError) as e:
        pass

      try:
        relem = pelem.xpath('v:egress-reorder', namespaces = XPNS)[0]
        rinfo = Ports.ReorderInfo()
        rinfo.percent = relem.get("percent")
        rinfo.correlation = relem.get("correlation")
        rinfo.gap = relem.get("gap", None)
        p.reorder_info = rinfo
      except (TypeError, IndexError):
        pass

    return p


def DockerXmlHandler(sliver, elem):
  log = KeyAdapter(sliver.uuid, DB._log)

  c = Container(sliver)
  c.client_id = elem.get("client_id")

#  log.debug("[%s] Container parsing" % (c.client_id))

  imgelem = elem.xpath('v:image', namespaces=XPNS)[0]
  c.image = imgelem.get("name")
  c.ram = int(elem.get("ram", c._image_desc.mem_default))
  attrs = imgelem.xpath('v:image-attribute', namespaces = XPNS)
  for aelem in attrs:
    c.setAttribute(aelem.get("name"), aelem.get("value"))
  c.validateAttributes()

#  if not CM.validImage(c.image):
    # TODO: Complain
#    return None

  # TODO: Start checking image desc attributes for whether these features
  # are actually supported by this container type
  mounts = elem.xpath('v:mount', namespaces=XPNS)
  for melem in mounts:
    if melem.get("type") == "hg":
      name = melem.get("name")
      cpath = melem.get("path")
      source = melem.get("source")
      branch = melem.get("branch")
      mgd_source = DB.getManagedSource(source)
      c._hgvols.append(HGVolumeManager.createVolume(sliver, c, name, mgd_source, cpath, branch, req_source = source))
    elif melem.get("type") == "hg-secure":
      name = melem.get("name")
      cpath = melem.get("path")
      source = ResourceRegistry.decrypt(melem.get("source"))
      branch = melem.get("branch")
      vinfo = HGVolumeManager.createVolume(sliver, c, name, source, cpath, branch)
      vinfo.sensitive = True
      c._hgvols.append(vinfo)
    elif melem.get("type") == "dropbox":
      name = melem.get("name")
      cpath = melem.get("path")
      c._dbxvols.append(DropboxVolumeManager.createVolume(sliver, c, name, cpath))
    else:
      sliver.log.warning("Unknown mount type (%s) requested" % (melem.get("type")))
      continue

  for relem in elem.xpath('v:route', namespaces = XPNS):
    c._pending_routes.append((relem.get("network"), relem.get("gateway")))

  ports = elem.xpath('v:port', namespaces = XPNS)
  for pelem in ports:
    if pelem.get("type") != "internal":
      # TODO: Error
      continue
    p = Ports.InternalContainer(c)
    p.client_id = str(pelem.get("client_id"))
#    log.debug("[%s] Port parsing" % (p.client_id))
    telem = pelem.xpath('v:target', namespaces = XPNS)[0]
    p.target = telem.get("remote-clientid")
    try:
      edelem = pelem.xpath('v:egress-delay', namespaces = XPNS)[0]
      p.setDelayInfo(edelem.get("time"), edelem.get("jitter"), edelem.get("correlation"), edelem.get("distribution"))
    except (TypeError, IndexError):
      pass

    try:
      lelem = pelem.xpath('v:egress-loss', namespaces = XPNS)[0]
      linfo = Ports.LossInfo()
      linfo.percent = lelem.get("percent")
      p.loss_info = linfo
    except (TypeError, IndexError):
      pass

    try:
      relem = pelem.xpath('v:egress-reorder', namespaces = XPNS)[0]
      rinfo = Ports.ReorderInfo()
      rinfo.percent = relem.get("percent")
      rinfo.correlation = relem.get("correlation")
      rinfo.gap = relem.get("gap", None)
      p.reorder_info = rinfo
    except (TypeError, IndexError):
      pass

    for v4elem in pelem.xpath('v:ipv4-address', namespaces = XPNS):
#      log.debug("[%s] IP address: %s" % (p.client_id, v4elem.get("value")))
      p._pending_addresses.setdefault("ipv4", []).append(v4elem.get("value"))

    c.addPort(p)
  return c

def rspecImageHandlerFactory (elem, dp):
  # TODO: Switch handler class based on image name
  return OVSXmlHandler(elem, dp)


class SliverBundleProxy(object):
  # Wraps a list of slivers to provide the same interface as a single one
  # The wrapper is done by hand instead of dealing with the side effects of trying to be magic

  # All client_ids in the bundle must be unique

  def __init__ (self, slivers):
    self._slivers = slivers
    self.log = KeyAdapter("%s Bundle" % (slivers[0].slice_urn), DB._log)

  def addSliver (self, sliver):
    self._slivers.append(sliver)

  @property
  def client_ids (self):
    cids = set()
    for sliver in self._slivers:
      cids |= sliver.client_ids
    return cids

  @property
  def uuid (self):
    # If you only have one sliver we'll let you call this, otherwise we'll bust you for Problems.
    if len(self._slivers) == 1:
      return self._slivers[0].uuid
    raise AmbiguousUUIDError()

  @property
  def datapaths (self):
    for sliver in self._slivers:
      for datapath in sliver.datapaths:
        yield datapath

  @property
  def containers (self):
    for sliver in self._slivers:
      for container in sliver.containers:
        yield container

  @property
  def functions (self):
    for sliver in self._slivers:
      for function in sliver.functions:
        yield function

  def getDatapath (self, client_id):
    for sliver in self._slivers:
      dp = sliver.getDatapath(client_id)
      if dp: return dp

  def getFunction (self, client_id):
    for sliver in self._slivers:
      func = sliver.getFunction(client_id)
      if func: return func

  def removeFunction (self, vf):
    vf.sliver.removeFunction(vf)

  def setController (self, client_id, controller_url, ofver):
    for sliver in self._slivers:
      if sliver.getDatapath(client_id):
        sliver.setController(client_id, controller_url, ofver)
        return
    return

  def getContainer (self, client_id):
    for sliver in self._slivers:
      ctr = sliver.getContainer(client_id)
      if ctr: return ctr

  def portUp (self, client_id):
    port = self.getPortByCID(client_id)
    port.up = True

  def portDown (self, client_id):
    port = self.getPortByCID(client_id)
    port.up = False

  def getPortByCID (self, client_id):
    for sliver in self._slivers:
      try:
        port = sliver.getPortByCID(client_id)
        return port
      except UnknownPortError:
        continue
    raise UnknownPortError(client_id)

  def removeDatapathPort (self, dpclid, portclid):
    for sliver in self._slivers:
      if sliver.getDatapath(dpclid):
        sliver.removeDatapathPort(dpclid, portclid)
        break 

  def cleanInterconnects (self):
    """Used to scrub interconnects after sliver deletions.  We can't trivially identify cross-slice ports
       that don't exist, so we try to duck the issue here."""
    ports = set()
    targets = {}

    for dp in self.datapaths:
      for port in dp.ports.itervalues():
        ports.add(port.client_id)
        targets[port.target] = (dp, port)

    for ct in self.containers:
      for port in ct.ports.itervalues():
        ports.add(port.client_id)
        targets[port.target] = (ct, port)

    for func in self.functions:
      ports.add(func.client_id)

    # If we have a target that doesn't exist in ports, we remove the source port
    missing = set(targets.keys()) - ports
    self.log.log(5, "Found ports: %s" % (", ".join(list(ports))))
    self.log.log(5, "Needed targets: %s" % (", ".join(targets.keys())))
    self.log.debug("Missing targets: %s" % (", ".join(list(missing))))
    for name in missing:
      if name.split(":")[0] in DB.public_bridges:
        # TODO: We need to make cross-slice connections an attribute instead of hinky text
        continue # In another slice
      (vertex, port) = targets[name]
      self.log.log(5, "Missing target port: %s for source port %s" % (name, port.client_id))
      del vertex.ports[port.client_id]

    for sliver in self._slivers:
      sliver._rebuildPortsByCID()


def poaChangeController (action, urns, options):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  curl = options["controller-url"]
  ofver = options.setdefault("openflow-version", None)
  for cid in options["datapaths"]:
    sliver.setController(cid, curl, ofver)
  DB._write_slivers()
  return {}
    
def poaAddContainerKeys (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  for clid in options["client-ids"]:
    ct = sliver.getContainer(clid)
    ct.addSSHKeys(options["ssh-keys"])
  return {}

def poaPortStatusChange (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  if action == "vts:port-down":
    sliver.portDown(options["port-client-id"])
  elif action == "vts:port-up":
    sliver.portUp(options["port-client-id"])
  return {}

def poaSetPortBehaviour (action, urns, options, user_urn):
  # { "ports" : [ <client_id> : <delay/loss obj>, ...] }
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  for cid, desc in options["ports"]:
    port = sliver.getPortByCID(cid)

    if desc["type"] == "egress-delay":
      dobj = Ports.DelayInfo.jconstruct(desc)
      port.delay_info = dobj
    elif desc["type"] == "egress-loss":
      dobj = Ports.LossInfo.jconstruct(desc)
      port.loss_info = dobj
    port.resetBehaviour()

  DB._write_slivers()


def poaOFDumpFlows (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  ret = {}
  hd = HostDriver(DB._log.info)
  for clid in options["datapaths"]:
    dp = sliver.getDatapath(clid)
    if not dp:
      continue
    flowstr = hd.dumpFlows(dp.id)
    flows = flowstr.split("\n")
    # Suppress a final empty line
    if not flows[-1]:
      flows = flows[:-1]
    if options.setdefault("show-hidden", False):
      ret[dp.client_id] = flows
    else:
      ret[dp.client_id] = []
      for flow in flows:
        if flow.count("NORMAL") or flow.count("table_id=254"):
          continue
        ret[dp.client_id].append(flow)

  return ret

def poaShareDatapath (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  dp = sliver.getDatapath(options["client-id"])
  dp.public_desc = options["description"]
  DB._write_slivers()
  DB.refreshPublicBridges()

  return True


# The "api:l2-switch" version
def poaOVSDumpFDBAPI (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  ret = {}
  hd = HostDriver(DB._log.info)
  for clid in options["client-ids"]:
    dp = sliver.getDatapath(clid)
    if not dp:
      continue
    dbstr = hd.dumpFDB(dp.id)
    macs = dbstr.split("\n")
    # Suppress a final empty line
    if not macs[-1]:
      macs = macs[:-1]
    ret[dp.client_id] = [x.split() for x in macs]
  if not ret:
    raise SkipHandler()
  return ret

# Legacy POA
def poaOVSDumpFDB (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  ret = {}
  hd = HostDriver(DB._log.info)
  for clid in options["datapaths"]:
    dp = sliver.getDatapath(clid)
    if not dp:
      continue
    dbstr = hd.dumpFDB(dp.id)
    macs = dbstr.split("\n")
    # Suppress a final empty line
    if not macs[-1]:
      macs = macs[:-1]
    ret[dp.client_id] = [x.split() for x in macs]
  return ret
  

def poaOVSSTPInfo (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  errors = []

  import foam.vts.drivers.ovsdb as ovs
  ovs.InfoDriver.log = DB._log
  retl = []
  for clid in options["datapaths"]:
    dp = sliver.getDatapath(clid)
    if not dp:
      errors.append("Datapath [%s] not found in this sliver" % (clid))
      continue

    pmap = {}
    for port in dp.ports.itervalues():
      pmap[port.name] = port

    brdata = ovs.InfoDriver.getSTPInfo(dp.id)
    brdata["client-id"] = clid
    true_plist = []
    for port in brdata["ports"]:
      if port["name"] == dp.id:
        continue
      port["num"] = pmap[port["name"]].ofportnum
      port["up"] = pmap[port["name"]].up
      port["client-id"] = pmap[port["name"]].client_id
      true_plist.append(port)
    brdata["ports"] = true_plist
    retl.append(brdata)
  return retl

def poaOVSRSTPInfo (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  errors = []

  import foam.vts.drivers.ovsdb as ovs
  ovs.InfoDriver.log = DB._log
  retl = []
  for clid in options["datapaths"]:
    dp = sliver.getDatapath(clid)
    if not dp:
      errors.append("Datapath [%s] not found in this sliver" % (clid))
      continue

    pmap = {}
    for port in dp.ports.itervalues():
      pmap[port.name] = port

    brdata = ovs.InfoDriver.getRSTPInfo(dp.id)
    brdata["client-id"] = clid
    true_plist = []
    for port in brdata["ports"]:
      if port["name"] == dp.id:
        continue
      port["num"] = pmap[port["name"]].ofportnum
      port["up"] = pmap[port["name"]].up
      port["client-id"] = pmap[port["name"]].client_id
      true_plist.append(port)
    brdata["ports"] = true_plist
    retl.append(brdata)
  return retl

def poaOVSSetSTPParams (action, urns, options, user_urn):
  from .drivers import ovs as OVS

  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  errors = []

  for clid in options["client-ids"]:
    dp = sliver.getDatapath(clid)
    if not dp:
      errors.append("Datapath [%s] not found in this sliver" % (clid))
      continue
    if dp.controllers:
      errors.append("OpenFlow-enabled Datapath [%s] cannot use STP" % (clid))
      continue
  if errors:
    raise POAValidationError("vts:l2:set-stp-params", errors)

  pinfo = options["params"]
  if pinfo["type"] == "stp":
    stpinfo = OVS.STPParams()
    if pinfo.has_key("system-id"): stpinfo.system_id = pinfo["system-id"]
    if pinfo.has_key("priority"): stpinfo.priority = pinfo["priority"]
    if pinfo.has_key("hello-time"): stpinfo.hello_time = pinfo["hello-time"]
    if pinfo.has_key("max-age"): stpinfo.max_age = pinfo["max-age"]
    if pinfo.has_key("forward-delay"): stpinfo.forward_delay = pinfo["forward-delay"]
  elif pinfo["type"] == "rstp":
    stpinfo = OVS.RSTPParams()
    if pinfo.has_key("address"): stpinfo.system_id = pinfo["address"]
    if pinfo.has_key("priority"): stpinfo.priority = pinfo["priority"]
    if pinfo.has_key("ageing-time"): stpinfo.ageing_time = pinfo["ageing-time"]
    if pinfo.has_key("max-age"): stpinfo.max_age = pinfo["max-age"]
    if pinfo.has_key("xmit-hold-count"): stpinfo.hold_count = pinfo["xmit-hold-count"]
    if pinfo.has_key("forward-delay"): stpinfo.forward_delay = pinfo["forward-delay"]
  else:
    raise POAValidationError("vts:l2:set-stp-params", ["Unknown STP mode [%s]" % (pinfo["type"])])

  hd = HostDriver(DB._log.info)
  for clid in options["client-ids"]:
    dp = sliver.getDatapath(clid)
    dp.stpinfo = stpinfo
    hd.setBridgeSTPMode(dp)
  DB._write_slivers()
  return {}


def poaGetPortInfo (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  import foam.vts.drivers.ovsdb as ovs
  ovs.InfoDriver.log = DB._log
  retd = {}
  for clid in options["datapaths"]:
    dp = sliver.getDatapath(clid)

    if not dp:
      raise UnknownDatapathError(clid)

    pmap = {}
    for port in dp.ports.itervalues():
      pmap[port.name] = port

    port_data = ovs.InfoDriver.getPortInfo(dp.id)
    true_plist = []
    for pdata in port_data:
      if pdata["name"] == dp.id:
        continue
      if pdata["name"] in pmap:
        pdata["num"] = pmap[pdata["name"]].ofportnum
        pdata["client-id"] = pmap[pdata["name"]].client_id
      else:
        pdata["num"] = ovs.InfoDriver.getOFPortNum(pdata["name"])
        pdata["client-id"] = ""

      if not isinstance(pdata["tag"], int):
        pdata["tag"] = None
      true_plist.append(pdata)
    retd[clid] = true_plist
  return retd
  

def poaOFClearFlows (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  hd = HostDriver(DB._log.info)

  errors = []
  if isinstance(options["datapaths"], list):
    for clid in options["datapaths"]:
      if not sliver.getDatapath(clid):
        errors.append("Datapath [%s] not found in this sliver" % (clid))
  else:
    errors.append("Datapaths must be a list")
  if errors:
    raise POAValidationError("vts:of:clear-flows", errors)

  for clid in options["datapaths"]:
    dp = sliver.getDatapath(clid)
    hd.clearFlows(dp.id)
    if not dp.controllers:
      hd.setNormalFlow(dp.id)
  return {}
  

def poaOFAddFlows (action, urns, options, user_urn):
  import foam.vts.drivers.ovsdb as ovs

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  errors = []
  hd = HostDriver(DB._log.info)
  db = ovs.OVSDBInfoDriver(DB._log)

  for fspec in options["rules"]:
    dp = sliver.getDatapath(fspec["datapath"])
    if not dp:
      errors.append("Datapath [%s] not found in this sliver" % (fspec["datapath"]))
      continue

    if fspec["output"]:
      pdata = fspec["output"]
      if isinstance(pdata, (list, tuple)):
        pnum = []
        for pname in pdata:
          port = dp.getPort(pname)
          if not port:
            errors.append("Output port [%s] not found in this sliver" % (fspec["output"]))
            continue
          pnum.append(db.getOFPortNum(port.name))
      else:
        if pdata == "___normal___":
          pnum = -1
        else:
          port = dp.getPort(pdata)
          if not port:
            errors.append("Output port [%s] not found in this sliver" % (fspec["output"]))
            continue
          pnum = db.getOFPortNum(port.name)

      hd.addFlow(dp, fspec["match"], pnum)
    else:
      hd.addFlow(dp, fspec["match"], None)

  return {"errors" : errors}


def poaOFAddFlowsAdvanced (action, urns, options, user_urn):
  import foam.vts.drivers.ovsdb as ovs

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  errors = []
  hd = HostDriver(DB._log.info)
  db = ovs.OVSDBInfoDriver(DB._log)

  for fspec in options["rules"]:
    dp = sliver.getDatapath(fspec["datapath"])
    if not dp:
      errors.append("Datapath [%s] not found in this sliver" % (fspec["datapath"]))
      continue

    pnum = None
    if fspec["output"]:
      pdata = fspec["output"]
      if isinstance(pdata, (list, tuple)):
        pnum = []
        for pname in pdata:
          port = dp.getPort(pname)
          if not port:
            errors.append("Output port [%s] not found in this sliver" % (fspec["output"]))
            continue
          pnum.append(db.getOFPortNum(port.name))
      else:
        if pdata == "___normal___":
          pnum = -1
        else:
          port = dp.getPort(pdata)
          if not port:
            errors.append("Output port [%s] not found in this sliver" % (fspec["output"]))
            continue
          pnum = db.getOFPortNum(port.name)

    hd.addFlowAdvanced(dp, fspec["match"], fspec["actions"], pnum)

  return {"errors" : errors}


def poaSetVLAN (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  hd = HostDriver(DB._log.info)
  # TODO: Need to validate that the passed in options are actually
  # in the format we require
  for val in options["ports"]:
    if not isinstance(val, (tuple, list)) or len(val) != 2:
      raise BadPortVLANFormatError(val)

  errors = []
  for client_id,vlan in options["ports"]:
    if vlan < 1 or vlan > 4095:
      raise IllegalVlanIDError(vlan)
    try:
      port = sliver.getPortByCID(client_id)
      if not isinstance(port, Ports.Internal):
        errors.append("VLAN cannot be set on port [%s]" % (client_id))
        continue
      port.vlan = vlan
      hd.setPortVLAN(port.name, vlan)
    except UnknownPortError:
      errors.append("Port [%s] not found in this sliver" % (client_id))
  DB._write_slivers()
  return {"errors" : errors}

def poaSetTrunk (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  hd = HostDriver(DB._log.info)
  for client_id in options["ports"]:
    port = sliver.getPortByCID(client_id)
    if not isinstance(port, Ports.Internal):
      continue # TODO: Accumulate some kind of error message
    if not port.vlan:
      continue # Already no VLAN here
    hd.removePortVLAN(port.name, port.vlan)
  return {}

def poaAddVPNPort (action, urns, options, user_urn):
  from .functions.sslvpn import L2SSLVPNServer

  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  dp = sliver.getDatapath(options["datapath"])
  if not dp:
    raise UnknownDatapathError(options["datapath"])

  vpnobj = L2SSLVPNServer.fromdict(options)
  vpnobj.user_urn = user_urn
  vpnobj.provision(DB, sliver.log)

  hd = HostDriver(sliver.log.info)

  vfport = Ports.VFPort(dp)
  vfport.client_id = "%s:_%d" % (dp.client_id, vpnobj.transport_port)
  vfport.name = vpnobj.port_name
  vfport.target = vpnobj.client_id
  vfport.provision(hd, sliver.log)
  dp.addPort(vfport)
  dp.sliver.addFunction(vpnobj)

  hd.attachPort(dp.id, vfport)

  vpnobj.start(sliver.log)

  DB._write_slivers()

  return vpnobj.userInfo()

def poaDelVPNPort (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  hd = HostDriver(sliver.log.info)
  func = sliver.getFunction(options["client-id"])

  if not func:
    return

  vdp = None
  vport = None
  for dpath in sliver.datapaths:
    if vdp:
      break
    for cid,port in dpath.ports.items():
      if port.type == "vf-port":
        if port.target == func.client_id:
          vdp = dpath.client_id
          vport = port.client_id
          break

  if vdp:
    sliver.removeDatapathPort(vdp, vport)

  func.deprovision(DB, sliver.log)
  sliver.removeFunction(func)

  DB._write_slivers()

def poaGetVPNInfo (action, urns, options, user_urn):
  if len(urns) > 1:
    return {} # Return some kind of error
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])

  func = sliver.getFunction(options["client-id"])
  if func:
    return func.userInfo()


def setup (app):
  ResourceRegistry.addAdNamespace(GENINS)
  ResourceRegistry.addAdNamespace(VTSNS)
  ResourceRegistry.addAdvertisementCallback(generateGENIAd)

  ResourceRegistry.addManifestNamespace(GENINS)
  ResourceRegistry.addManifestNamespace(GENIUSERNS)
  ResourceRegistry.addManifestNamespace(VTSMNS)
  ResourceRegistry.addManifestNamespace(SDNMNS)
  ResourceRegistry.addManifestCallback(genSliceManifest)

  ResourceRegistry.addAllocateCallback(parseRequest)
  ResourceRegistry.addProvisionCallback(provisionSliver)
  ResourceRegistry.addDeleteCallback(deleteSlice)
  ResourceRegistry.addDeleteSliverCallback(deleteSliver)
  ResourceRegistry.addStatusCallback(sliverStatus)
  ResourceRegistry.addRenewCallback(renewSliver)

  ResourceRegistry.addPOACallback("vts:port-down", poaPortStatusChange)
  ResourceRegistry.addPOACallback("vts:port-up", poaPortStatusChange)
  ResourceRegistry.addPOACallback("vts:share-datapath", poaShareDatapath)
  ResourceRegistry.addPOACallback("vts:raw:set-port-behaviour", poaSetPortBehaviour)
  ResourceRegistry.addPOACallback("vts:of:change-controller", poaChangeController)
  ResourceRegistry.addPOACallback("vts:of:dump-flows", poaOFDumpFlows)
  ResourceRegistry.addPOACallback("vts:of:clear-flows", poaOFClearFlows)
  ResourceRegistry.addPOACallback("vts:of:add-flows", poaOFAddFlows)
  ResourceRegistry.addPOACallback("vts:of:add-flows-advanced", poaOFAddFlowsAdvanced)
  ResourceRegistry.addPOACallback("vts:raw:set-vlan", poaSetVLAN)
  ResourceRegistry.addPOACallback("vts:raw:set-trunk", poaSetTrunk)
  ResourceRegistry.addPOACallback("api:l2-switch:get-l2-table", poaOVSDumpFDBAPI)
  ResourceRegistry.addPOACallback("api:l2-switch:stp-info", poaOVSSTPInfo)
  ResourceRegistry.addPOACallback("api:l2-switch:rstp-info", poaOVSRSTPInfo)
  ResourceRegistry.addPOACallback("api:l2-switch:get-port-info", poaGetPortInfo)
  ResourceRegistry.addPOACallback("api:l2-switch:set-stp-params", poaOVSSetSTPParams)
  ResourceRegistry.addPOACallback("vts:container:add-keys", poaAddContainerKeys)
  ResourceRegistry.addPOACallback("vts:vpn:add-port", poaAddVPNPort)
  ResourceRegistry.addPOACallback("vts:vpn:del-port", poaDelVPNPort)
  ResourceRegistry.addPOACallback("vts:vpn:get-info", poaGetVPNInfo)
#  ResourceRegistry.addPOACallback("vts:pg:add-circuit", poaAddPGCircuit)
#  ResourceRegistry.addPOACallback("vts:pg:del-circuit", poaDelPGCircuit)

  # DEPRECATED STUFF:
  ResourceRegistry.addPOACallback("vts:l2:dump-macs", poaOVSDumpFDB)
  ResourceRegistry.addPOACallback("vts:l2:stp-info", poaOVSSTPInfo)
  ResourceRegistry.addPOACallback("vts:l2:rstp-info", poaOVSRSTPInfo)
  ResourceRegistry.addPOACallback("vts:raw:get-port-info", poaGetPortInfo)
  ResourceRegistry.addPOACallback("vts:l2:set-stp-params", poaOVSSetSTPParams)

  citems = []
  # TODO: This belongs in a global "geni" bit
  citems.append(ConfigItem().setKey("vts.geni.cm-tag").setValue("example.com")
                .setDesc("Unique Component Manager Tag for this instance"))
  citems.append(ConfigItem().setKey("vts.geni.container-policy").setValue({})
                .setDesc("Container policy")
                .setUpdateFuncName("foam.core.configdb.checkDict"))
  citems.append(ConfigItem().setKey("vts.geni.capability-policy").setValue({})
                .setDesc("Capability policy")
                .setUpdateFuncName("foam.core.configdb.checkDict"))
  [ConfigDB.installConfigItem(x) for x in citems]

  ginfo = {}
  ginfo["urn"] = "urn:publicid:IDN+%s+authority+cm" % (ConfigDB.getConfigItemByKey("vts.geni.cm-tag").getValue())
  APIRegistry.addStaticInfo(ginfo)

  for desc in CM.imagedescs:
    if desc._handler:
      desc._handler.geniSetup(ResourceRegistry)


