# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Driver for manipulating VTS bridges and ports on the local host

import sys
import subprocess
import os
import os.path
import pwd

from .drivers import ovs as OVS
from .interfaces import VTSDriver
from ..core.configdb import ConfigDB, ConfigItem
from ..core.exception import CoreException

class NoVolumeStorageError(CoreException):
  def __str__ (self):
    return "Volume storage has not been configured for this host."

def call (cmd, log_cmd = None, log_output = None):
  if log_cmd:
    log_cmd(cmd)
  if log_output:
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
    for line in iter(p.stdout.readline, ""):
      log_output(line.rstrip())
    sout = ""
    serr = ""
  else:
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (sout, serr) = p.communicate()
  return (sout, serr, p.returncode)

class HostDriver(VTSDriver):
  def __init__ (self, log_cmd):
    super(HostDriver, self).__init__()
    self.log_cmd = log_cmd
    self._patches = {}
    self.sudo = "sudo"
    self.nssudo = "sudo"
    self._ns = None
    self._volroot = ConfigDB.getConfigItemByKey("vts.hostvols.root").getValue()

  @property
  def ns (self):
    return self._ns

  @ns.setter
  def ns (self, val):
    self._ns = val
    self.nssudo = "sudo ip netns %s exec" % (val)

  @property
  def volroot (self):
    if not self._volroot:
      raise NoVolumeStorageError()
    return self._volroot

  def call (self, cmd, log_output = None):
    call(cmd, log_cmd = self.log_cmd, log_output = log_output)

  def createBridge (self, dp):
    call("sudo ovs-vsctl add-br %s" % (dp.id), log_cmd=self.log_cmd)
    call("sudo ovs-vsctl set bridge %s other-config:datapath-id=%s" % (dp.id, dp.dpid.hexstr()), log_cmd=self.log_cmd)
    if dp.mactable_size:
      call("sudo ovs-vsctl set bridge %s other-config:mac-table-size=%d" % (dp.id, dp.mactable_size), log_cmd=self.log_cmd)
    if dp.mactable_age:
      call("sudo ovs-vsctl set bridge %s other-config:mac-aging-time=%d" % (dp.id, dp.mactable_age), log_cmd=self.log_cmd)
    self.setBridgeSTPMode(dp)

  def setBridgeSTPMode (self, dp):
    name = dp.id
    if isinstance(dp.stpinfo, OVS.STPParams):
      call("sudo ovs-vsctl set bridge %s stp_enable=true" % (name), log_cmd=self.log_cmd)
      call("sudo ovs-vsctl set bridge %s rstp_enable=false" % (name), log_cmd=self.log_cmd)
    elif isinstance(dp.stpinfo, OVS.RSTPParams):
      call("sudo ovs-vsctl set bridge %s rstp_enable=true" % (name), log_cmd=self.log_cmd)
      call("sudo ovs-vsctl set bridge %s stp_enable=false" % (name), log_cmd=self.log_cmd)
    elif isinstance(dp.stpinfo, OVS.DisabledSTP):
      call("sudo ovs-vsctl set bridge %s rstp_enable=false" % (name), log_cmd=self.log_cmd)
      call("sudo ovs-vsctl set bridge %s stp_enable=false" % (name), log_cmd=self.log_cmd)

    if dp.stpinfo.cmdlist:
      call("sudo ovs-vsctl set bridge %s %s" % (name, " ".join(dp.stpinfo.cmdlist)), log_cmd=self.log_cmd)

  def destroyBridge (self, name):
    call("sudo ovs-vsctl del-br %s" % (name), log_cmd = self.log_cmd)

  def touch(self, fname, times=None):
    with open(fname, 'a'):
      os.utime(fname, times)

# Python 3.3
#  def touch(self, fname, mode=0o666, dir_fd=None, **kwargs):
#    flags = os.O_CREAT | os.O_APPEND
#    with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
#      os.utime(f.fileno() if os.utime in os.supports_fd else fname,
#               dir_fd=None if os.supports_fd else dir_fd, **kwargs)

  def attachPort (self, bridge, port, vid = None):
    if not vid:
      call("sudo ovs-vsctl add-port %s %s" % (bridge, port.alias), log_cmd = self.log_cmd)
    else:
      call("sudo ovs-vsctl add-port %s %s tag=%d" % (bridge, port.alias, vid), log_cmd = self.log_cmd)
    self.portUp(port)

  def attachMirrorPort (self, bridge, port, mirrorname):
#     ovs-vsctl add-port br0 tap1 \
#       -- --id=@p get port tap1 \
#       -- --id=@m create mirror name=m0 select-all=true output-port=@p \
#       -- set bridge br0 mirrors=@m
    cmdl = ["sudo ovs-vsctl add-port %s %s" % (bridge, port)]
    cmdl.append("-- --id=@p get port %s" % (port))
    cmdl.append("-- --id=@m create mirror name=%s select-all=true output-port=@p" % (mirrorname))
    cmdl.append("-- set bridge %s mirrors=@m" % (bridge))
    cmd = " ".join(cmdl)
    call(cmd, log_cmd = self.log_cmd, log_output = self.log_cmd)

  def setPortVLAN (self, port, vid):
    call("sudo ovs-vsctl set Port %s tag=%d" % (port, vid), log_cmd = self.log_cmd)

  def removePortVLAN (self, port, vid):
    call("sudo ovs-vsctl remove Port %s tag %d" % (port, vid), log_cmd = self.log_cmd)

  def attachPatchPort (self, bridge, port):
    # call("sudo ovs-vsctl add-port %s %s" % (bridge, port), log_cmd = self.log_cmd)
    # call("sudo ovs-vsctl set Interface %s type=patch" % (port), log_cmd = self.log_cmd)
    # Do nothing right now, we're using veth pairs in the patch code
    return 

  def setPortImpairment (self, pname, dinfo, linfo, rinfo = None, rate = None):
    # TODO: Hosts that use linux qdiscs probably need a more complex driver now
    parent = None
    handle = None
    if rate:
      self.call("sudo tc qdisc add dev %s root handle 1:0 htb default 50" % (pname))
      self.call("sudo tc class add dev %s parent 1:0 classid 1:50 htb rate %d%s" % (pname,
                rate.speed, rate.tcunits))
      parent = "1:50"
      handle = "50"

    # If we rate limit, the root is the HTB limiter, otherwise it's the netem
    if parent:
      argl = ["sudo tc qdisc add dev %s parent %s handle %s netem" % (pname, parent, handle)]
    else:
      argl = ["sudo tc qdisc add dev %s root netem" % (pname)]

    if dinfo:
      argl.append("delay %dms" % (dinfo.time))
      if dinfo.jitter:
        argl.append("%dms" % (dinfo.jitter))
        if dinfo.correlation:
          argl.append("%d%%" % (dinfo.correlation))
      if dinfo.distribution:
        argl.append("distribution %s" % (dinfo.distribution))

    if linfo:
      argl.append("loss random %s%%" % (linfo.percent))

    if rinfo:
      if rinfo.gap:
        argl.append("reorder %s%% %s%% gap %s" % (rinfo.percent, rinfo.correlation, rinfo.gap))
      else:
        argl.append("reorder %s%% %s%%" % (rinfo.percent, rinfo.correlation))

    cmd = " ".join(argl)
    call(cmd, log_cmd = self.log_cmd)

  def clearImpairment (self, pname):
    call("sudo tc qdisc delete dev %s root" % (pname), log_cmd = self.log_cmd)

  def detachPort (self, bridge, port):
    call("sudo ovs-vsctl del-port %s %s" % (bridge, port), log_cmd = self.log_cmd)

  def deleteInterface (self, port):
    call("sudo ip link delete %s type veth" % (port), log_cmd = self.log_cmd)

  def attachGRE (self, bridge, port, remote, key):
    self.attachPort(bridge, port)
    call("sudo ovs-vsctl set Interface %s type=gre options:remote_ip=%s options:key=%d" % (port, str(remote), key), log_cmd = self.log_cmd)

  def setOFController (self, dp, controller):
    call("sudo ovs-vsctl set-fail-mode %s secure" % (dp.id), log_cmd = self.log_cmd)
    self.setBridgeSTPMode(dp)
    proto = "OpenFlow10"
    if dp.ofver == "1.3":
      proto = "OpenFlow13"
    elif dp.ofver == "1.2":
      proto = "OpenFlow12"
    call("sudo ovs-vsctl set bridge %s protocols=%s" % (dp.id, proto), log_cmd = self.log_cmd)
    call("sudo ovs-vsctl set-controller %s %s" % (dp.id, controller), log_cmd = self.log_cmd)

  def setSFlow (self, bridge, sflow_info):
    sflow_info.bridge_id = bridge
    call('sudo ovs-vsctl --id=@sflow%(idx)s create sflow agent=%(agent_intf)s target="%(collector_ip)s\:%(collector_port)d"'
         ' header=%(header_bytes)d sampling=%(sampling_n)d polling=%(polling_secs)d -- set bridge %(bridge_id)s'
         ' sflow=@sflow%(idx)s' % (sflow_info), log_cmd = self.log_cmd)

  def setNetFlow (self, bridge, nf_info):
    nf_info.bridge_id = bridge
    call('sudo ovs-vsctl -- set Bridge %(bridge_id)s netflow=@nf%(idx)s -- --id=@nf%(idx)s create netflow '
         'target="%(collector_ip)s\:%(collector_port)d" active-timeout=%(timeout)d')


#  def patch (self, br1, p1, br2, p2):
#    call("sudo ovs-vsctl set Interface %s options:peer=%s" % (p1, p2), log_cmd = self.log_cmd)
#    call("sudo ovs-vsctl set Interface %s options:peer=%s" % (p2, p1), log_cmd = self.log_cmd)

  def patch (self, br1, p1, br2, p2):
    self.createWire(p1, p2)
    self.attachPort(br1, p1)
    self.attachPort(br2, p2)
    self.portUp(p1)
    self.portUp(p2)

  def createWire (self, p1, p2):
    call("sudo ip link add %s type veth peer name %s" % (p1, p2), log_cmd = self.log_cmd)

  def portUp (self, p):
    call("sudo ip link set dev %s up" % (p.alias), log_cmd = self.log_cmd)

  def portDown (self, p):
    call("sudo ip link set dev %s down" % (p.alias), log_cmd = self.log_cmd)

  def dumpFlows (self, brname):
    (out, err, ret) = call("sudo ovs-appctl bridge/dump-flows %s" % (brname), log_cmd = self.log_cmd)
    return out

  def dumpFDB (self, brname):
    (out, err, ret) = call("sudo ovs-appctl fdb/show %s" % (brname), log_cmd = self.log_cmd)
    return out

  def clearFlows (self, brname):
    (out, err, ret) = call("sudo ovs-ofctl del-flows %s" % (brname), log_cmd = self.log_cmd)

  def setNormalFlow (self, brname):
    (o,e,r) = call("sudo ovs-ofctl add-flow %s priority=0,actions=output:NORMAL" % (brname),
                   log_cmd = self.log_cmd)

  def addFlow (self, dp, match, pnum):
    mlist = []
    for k,v in match.iteritems():
      mlist.append("%s=%s" % (k, str(v)))

    if pnum:
      if isinstance(pnum, (list, tuple)):
        action_list = ["output:%s" % (x) for x in pnum]
        (out, err, ret) = call("sudo ovs-ofctl add-flow %s %s,actions=%s" % (dp.id,
                               ",".join(mlist), ",".join(action_list)), log_cmd=self.log_cmd)
      else:
        if pnum == -1:
          (out, err, ret) = call("sudo ovs-ofctl add-flow %s %s,actions=normal" % (dp.id,
                                 ",".join(mlist)), log_cmd=self.log_cmd)
        else:
          (out, err, ret) = call("sudo ovs-ofctl add-flow %s %s,actions=output:%s" % (dp.id,
                                 ",".join(mlist), pnum), log_cmd=self.log_cmd)
    else:
      (out, err, ret) = call("sudo ovs-ofctl add-flow %s %s,actions=" % (dp.id,
                             ",".join(mlist)), log_cmd=self.log_cmd)

  def addFlowAdvanced (self, dp, match, actions, pnum):
    mlist = []
    for k,v in match.iteritems():
      mlist.append("%s=%s" % (k, str(v)))

    if pnum:
      actions.append("output:%d" % (pnum))

    action_str = ",".join(actions)
    match_str = ",".join(mlist)

    (out, err, ret) = call("sudo ovs-ofctl add-flow %s %s,actions=%s" % (dp.id, match_str, action_str),
                           log_cmd=self.log_cmd)

  def getMACAddress (self, pname):
    f = open("/sys/class/net/%s/address" % (pname))
    mac = f.read().strip()
    return mac

  def addUser (self, uname, group_list, key_list, command = None):
    # Make the user
    cmd = "sudo /usr/sbin/useradd -G %s -m %s" % (",".join(group_list), uname)
    (out, err, ret) = call(cmd, log_cmd = self.log_cmd)
    
    # Make the .ssh directory and add the keys, with the optionally provided command
    # If a command is provided we add some security filters to keep you in that jail
    userhome = os.path.expanduser("~%s" % (uname))
    (o,e,r) = call("sudo mkdir -p %s/.ssh/" % (userhome))
    (o,e,r) = call("sudo chown www-data:www-data %s/.ssh/" % (userhome))
    f = open("%s/.ssh/authorized_keys" % (userhome), "w+")

    aklines = []
    if command:
      for key in key_list:
        aklines.append("command=\"%s\",no-port-forwarding,no-X11-forwarding,no-agent-forwarding %s" % (command, key))
    else:
      aklines = key_list

    f.write("\n".join(aklines))
    f.close()

    (o,e,r) = call("sudo chmod 0644 %s" % (f.name), log_cmd = self.log_cmd)
    (o,e,r) = call("sudo chown -R %s:%s %s" % (uname, uname, userhome), log_cmd = self.log_cmd)

  def delUser (self, uname):
    (o, e, r) = call("sudo /usr/sbin/userdel -r %s" % (uname), log_cmd = self.log_cmd)

  def addUserSSHKeys (self, uname, keys, command = None):
    if not isinstance(keys, list):
      keys = [keys]
    userhome = os.path.expanduser("~%s" % (uname))

    (o,e,r) = call("sudo chown -R www-data:www-data %s/.ssh/" % (userhome))

    nklines = []
    if command:
      for key in keys:
        nklines.append("command=\"%s\",no-port-forwarding,no-X11-forwarding,no-agent-forwarding %s" % (command, key))
    else:
      nklines = keys

    f = open("%s/.ssh/authorized_keys" % (userhome), "a+")
    f.write("\n")
    f.write("\n".join(nklines))
    f.close()

    (o,e,r) = call("sudo chmod 0644 %s" % (f.name), log_cmd = self.log_cmd)
    (o,e,r) = call("sudo chown -R %s:%s %s" % (uname, uname, userhome), log_cmd = self.log_cmd)


def setup (app):
  citems = []
  citems.append(ConfigItem().setKey("vts.hostvols.root").setValue(None)
                .setDesc("Storage root for HostVolumes created by image handlers"))
  [ConfigDB.installConfigItem(x) for x in citems]


