# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import decimal

import IPy

from ..core import json

class UnknownPortTypeError(Exception):
  def __init__ (self, typ):
    self.typ = typ
  def __str__ (self):
    return "Unknown Port Type: %s" % (self.typ)


class VLANRange(object):
  def __init__ (self, spec):
    self.start_tag = None
    self.end_tag = None
    try:
      self.start_tag = int(spec)
      self.end_tag = int(spec)
      return
    except ValueError:
      pass

    stag,etag = spec.split("-")
    self.start_tag = int(stag)
    self.end_tag = int(etag)

  def __json__ (self):
    return {"start-tag" : self.start_tag, "end-tag" : self.end_tag}

  @classmethod
  def jconstruct (cls, obj):
    vr = cls()
    vr.start_tag = obj["start-tag"]
    vr.end_tag = obj["end-tag"]
    return vr

class ReorderInfo(object):
  def __init__ (self):
    self.percent = None
    self.correlation = None
    self.gap = None

  def __json__ (self):
    return {"percent" : self.percent, "correlation" : self.correlation, "gap" : self.gap}

  @classmethod
  def jconstruct (cls, obj):
    ri = cls()
    ri.percent = obj["percent"]
    ri.correlation = obj["correlation"]
    ri.gap = obj["gap"]
    return ri

class RateInfo(object):
  class UNITS(object):
    BITS = 0
    KBITS = 1
    MBITS = 2
    GBITS = 3

  TCMAP = {UNITS.BITS : "bit", 
           UNITS.KBITS : "kbit",
           UNITS.MBITS : "mbit",
           UNITS.GBITS : "gbit"}

  def __init__ (self):
    self.speed = None
    self._units = None

  @property
  def tcunits (self):
    return RateInfo.TCMAP[self._units]

  def __json__ (self):
    return {"speed" : self.speed, "units" : self._units, "ver" : 1}

  @classmethod
  def jconstruct (cls, obj):
    ri = cls()
    ri.speed = obj["speed"]
    ri._units = obj["units"]
    return ri

class LossInfo(object):
  def __init__ (self):
    self._percent = None

  def __json__ (self):
    return {"percent" : "%s" % (self._percent)}

  @classmethod
  def jconstruct (cls, obj):
    if obj["percent"] == 0:
      return None
    li = cls()
    li.percent = obj["percent"]
    return li

  @property
  def percent (self):
    return self._percent

  @percent.setter
  def percent (self, val):
    self._percent = decimal.Decimal(val)


class DelayInfo(object):
  def __init__ (self):
    self._time = None
    self._jitter = None
    self._correlation = None
    self.distribution = None

  def __json__ (self):
    return {"time" : self._time,
            "jitter" : self._jitter,
            "correlation" : self._correlation,
            "distribution" : self.distribution}

  @staticmethod
  def jconstruct (obj):
    di = DelayInfo()
    di._time = json.getdefault(obj, "time", 0)
    if di._time == 0:
      return None
    di._jitter = json.getdefault(obj, "jitter", None)
    di._correlation = json.getdefault(obj, "correlation", None)
    di.distribution = json.getdefault(obj, "distribution", None)
    return di

  @property
  def time (self):
    return self._time

  @time.setter
  def time (self, val):
    self._time = int(val)

  @property
  def jitter (self):
    return self._jitter

  @jitter.setter
  def jitter (self, val):
    self._jitter = int(val)

  @property
  def correlation (self):
    return self._correlation

  @correlation.setter
  def correlation (self, val):
    self._correlation = int(val)


class _PortBase(object):
  def __init__ (self, datapath, typestr):
    # Persisted
    self.client_id = None
    self.type = typestr
    self.name = None
    self._alias = None
    self._up = True
    self.driver = None

    # Ephemeral
    self.datapath = datapath
    self.needs_patch = False
    self.needs_vf_attach = False
    self._ofportnum = None
    self._pending_addresses = {}

  @property
  def up (self):
    return self._up

  @up.setter
  def up (self, val):
    self._up = val
    if self._up:
      self.driver.portUp(self)
    else:
      self.driver.portDown(self)

  @property
  def alias (self):
    if not self._alias:
      return self.name
    return self._alias

  @alias.setter
  def alias (self, val):
    self._alias = val

  @property
  def ofportnum (self):
    if not self._ofportnum:
      self.datapath.setOFPortNums()
    return self._ofportnum

  def __json__ (self):
    return {"_basever" : 3,
            "client-id" : self.client_id,
            "type" : self.type,
            "name" : self.name,
            "up" : self._up,
            "alias" : self._alias,
            "driver-name" : self.driver.__class__.__name__}

  @staticmethod
  def jconstruct (obj, datapath):
    if obj["type"] == "gre":
      p = GRE.jconstruct(obj, datapath)
    elif obj["type"] == "internal":
      p = Internal.jconstruct(obj, datapath)
    elif obj["type"] == "pglocal":
      p = PGLocal.jconstruct(obj, datapath)
    elif obj["type"] == "vf-port":
      p = VFPort.jconstruct(obj, datapath)
    elif obj["type"] == "internal-container":
      p = InternalContainer.jconstruct(obj, datapath)
    else:
      raise UnknownPortTypeError(obj["type"])

    try:
      ver = obj["_basever"]
    except:
      ver = 0

    if ver >= 0:
      p.client_id = obj["client-id"]
      p.name = obj["name"]
    if ver >= 1:
      p._alias = obj["alias"]
    if ver >= 2:
      p._up = obj["up"]
    if ver >= 3:
      if obj["driver-name"] == "NoneType" and not p.driver:
        datapath._sliver.log.warning("Missing port driver: %s" % (obj["client-id"]))
      elif obj["driver-name"] == "DockerDriver":
        from .drivers import docker
        p.driver = docker.DockerDriver(datapath._sliver.log.info)
      elif obj["driver-name"] == "HostDriver":
        from . import hostdriver
        p.driver = hostdriver.HostDriver(datapath._sliver.log.info)
      else:
        if not p.driver:
          datapath._sliver.log.warning("Unknown port driver: %s" % (obj["driver-name"]))

    return p


class GRE(_PortBase):
  def __init__ (self, datapath):
    super(GRE, self).__init__(datapath, "gre")
    self.endpoint_addr = None
    self.key = None
    self.circuit_plane = None

  @property
  def key (self):
    if self._key is None:
      self._key = self.circuit_plane.allocateGREKey(self.endpoint_addr)
    return self._key

  @key.setter
  def key (self, val):
    self._key = val

  def __json__ (self):
    d = super(GRE, self).__json__()
    d.update({"endpoint-addr" : str(self.endpoint_addr),
              "key" : self.key,
              "circuit-plane" : self.circuit_plane.label})
    return d

  @staticmethod
  def jconstruct (obj, datapath):
    from .data import DB
    p = GRE(datapath)
    p.client_id = obj["client-id"]
    p.endpoint_addr = IPy.IP(obj["endpoint-addr"])
    p.key = obj["key"]
    p.circuit_plane = DB.getCircuitPlane(obj["circuit-plane"])
    return p

  def setup (self, DB):
    self.name = DB.PNR.getName()

  def provision (self, driver, log):
    driver.attachGRE(self.datapath.id, self.name, self.endpoint_addr, self.key)

  def deprovision (self, driver, db):
    db.PNR.removeName(self.name)
    self.circuit_plane.releaseGREKey(self.endpoint_addr, self.key)


class Internal(_PortBase):
  def __init__ (self, datapath):
    super(Internal, self).__init__(datapath, "internal")
    self.target = None
    self.needs_patch = True
    self._vlan = None
    self._supported_vlans = set()
    self.delay_info = None
    self.loss_info = None
    self.reorder_info = None

  def __json__ (self):
    d = super(Internal, self).__json__()
    d["version"] = 2
    d["target"] = self.target
    d["vlan"] = self.vlan
    d["supported-vlans"] = list(self._supported_vlans)
    if self.delay_info: d["delay-info"] = self.delay_info
    if self.loss_info: d["loss-info"] = self.loss_info
    if self.reorder_info: d["reorder-info"] = self.reorder_info
    return d

  def __json_ephemeral__(self):
    obj = self.__json__()
    obj["_macaddr"] = self.macaddr
    return obj

  @staticmethod
  def jconstruct (obj, datapath):
    try:
      ver = obj["version"]
    except:
      ver = 1

    p = Internal(datapath)
    p.client_id = obj["client-id"]
    p.target = obj["target"]
    p.vlan = json.getdefault(obj, "vlan", None)

    dinfo = json.getdefault(obj, "delay-info", None)
    if dinfo:
      p.delay_info = DelayInfo.jconstruct(dinfo)

    linfo = json.getdefault(obj, "loss-info", None)
    if linfo:
      p.loss_info = LossInfo.jconstruct(linfo)

    rinfo = json.getdefault(obj, "reorder-info", None)
    if rinfo:
      p.reorder_info = ReorderInfo.jconstruct(rinfo)

    if ver >= 2:
      for tag in obj["supported-vlans"]:
        p._supported_vlans.add(tag)

    return p

  @property
  def macaddr (self):
    try:
      return self.driver.getMACAddress(self.alias)
    except AttributeError:
      return None

  @property
  def vlan (self):
    return self._vlan

  @vlan.setter
  def vlan (self, val):
    try:
      self._vlan = int(val)
    except TypeError:
      return

  def setup (self, DB):
    self.name = DB.PNR.getName()

  def attach (self):
    if self.datapath.mirror and self.datapath.mirror[0] == self.client_id:
      return
    self.driver.setPortImpairment(self.name, self.delay_info, self.loss_info, self.reorder_info)
    self.driver.attachPort(self.datapath.id, self, self.vlan)

  def provision (self, driver, log):
    self.driver = driver

  def deprovision (self, driver, db):
    driver.deleteInterface(self.name)
    db.PNR.removeName(self.name)

  def setDelayInfo (self, time = None, jitter = None, correlation = None, distribution = None):
    self.delay_info = DelayInfo()
    self.delay_info.time = time
    self.delay_info.jitter = jitter
    self.delay_info.correlation = None
    self.delay_info.distribution = None

  def resetBehaviour (self):
    self.driver.clearImpairment(self.name)
    self.driver.setPortImpairment(self.name, self.delay_info, self.loss_info, self.reorder_info)


class InternalContainer(_PortBase):
  def __init__ (self, container):
    super(InternalContainer, self).__init__(container, "internal-container")
    self.target = None
    self.needs_patch = True
    self.delay_info = None
    self.loss_info = None
    self.reorder_info = None
    self.vlan = None

  def __json__ (self):
    d = super(InternalContainer, self).__json__()
    d["target"] = self.target
    if self.delay_info: d["delay-info"] = self.delay_info
    if self.loss_info: d["loss-info"] = self.loss_info
    if self.reorder_info: d["reorder-info"] = self.reorder_info
    return d

  def __json_ephemeral__ (self):
    obj = self.__json__()
    obj["_macaddr"] = self.macaddr
    return obj

  @staticmethod
  def jconstruct (obj, container):
    p = InternalContainer(container)
    p.client_id = obj["client-id"]
    p.target = obj["target"]

    dinfo = json.getdefault(obj, "delay-info", None)
    if dinfo:
      p.delay_info = DelayInfo.jconstruct(dinfo)

    linfo = json.getdefault(obj, "loss-info", None)
    if linfo:
      p.loss_info = LossInfo.jconstruct(linfo)

    rinfo = json.getdefault(obj, "reorder-info", None)
    if rinfo:
      p.reorder_info = ReorderInfo.jconstruct(rinfo)

    return p

  def setup (self, DB):
    self.name = DB.PNR.getName()

  def attach (self):
    self.driver.attachPort(self.datapath, self)
    self.driver.setPortImpairment(self.datapath, self.alias, self.delay_info, self.loss_info, self.reorder_info)

  def provision (self, driver, log):
    self.driver = driver

  def deprovision (self, driver, db):
#    driver.deleteInterface(self.name)
    db.PNR.removeName(self.name)

  @property
  def macaddr (self):
    try:
      return self.driver.getMACAddress(self.datapath, self.alias)
    except AttributeError:
      return None

  @property
  def ipv4addrs (self):
    return self.driver.getPortIPv4Addresses(self.datapath, self.alias)

  def addIPv4Address (self, address):
    self.driver.addPortIPv4Address(self.datapath, self.alias, address)

  def setDelayInfo (self, time = None, jitter = None, correlation = None, distribution = None):
    self.delay_info = DelayInfo()
    self.delay_info.time = time
    self.delay_info.jitter = jitter
    self.delay_info.correlation = None
    self.delay_info.distribution = None

  def resetBehaviour (self):
    self.driver.clearImpairment(self.datapath, self.alias)
    self.driver.setPortImpairment(self.datapath, self.alias, self.delay_info, self.loss_info, self.reorder_info)


class PGLocal(_PortBase):
  def __init__ (self, datapath):
    super(PGLocal, self).__init__(datapath, "pglocal")
    self.circuit = None
    self.delay_info = None
    self.loss_info = None

  def __json__ (self):
    d = super(PGLocal, self).__json__()
    d["circuit"] = self.circuit
    if self.delay_info: d["delay-info"] = self.delay_info
    if self.loss_info: d["loss-info"] = self.loss_info
    return d

  @staticmethod
  def jconstruct (obj, datapath):
    p = PGLocal(datapath)
    p.circuit = obj["circuit"]
    dinfo = json.getdefault(obj, "delay-info", None)
    if dinfo: p.delay_info = DelayInfo.jconstruct(dinfo)
    linfo = json.getdefault(obj, "loss-info", None)
    if linfo: p.loss_info = LossInfo.jconstruct(linfo)
    return p

  def setup (self, DB):
    vid = DB.getVlanID(self.circuit)
    self.name = "vlan%d" % (vid)
  
  def provision (self, driver, log):
    driver.attachPort(self.datapath.id, self)
    driver.setPortImpairment(self.name, self.delay_info, self.loss_info)

  def deprovision (self, driver, db):
    db.freeLocalCircuit(self.circuit)
    driver.clearImpairment(self.name)

  def setDelayInfo (self, time = None, jitter = None, correlation = None, distribution = None):
    self.delay_info = DelayInfo()
    self.delay_info.time = time
    self.delay_info.jitter = jitter
    self.delay_info.correlation = None
    self.delay_info.distribution = None

  def resetBehaviour (self):
    self.driver.clearImpairment(self.name)
    self.driver.setPortImpairment(self.name, self.delay_info, self.loss_info)


class VFPort(_PortBase):
  def __init__ (self, datapath):
    super(VFPort, self).__init__(datapath, "vf-port")
    self.target = None
    self.needs_vf_attach = True

  def __json__ (self):
    d = super(VFPort, self).__json__()
    d["target"] = self.target
    d["version"] = 1
    return d

  @staticmethod
  def jconstruct (obj, datapath):
    version = json.getdefault(obj, "version", 1)
    p = VFPort(datapath)
    p.target = obj["target"]
    try:
      if obj["driver-name"] == "NoneType":
        from . import hostdriver
        p.driver = hostdriver.HostDriver(datapath._sliver.log.info)
    except KeyError:
      pass
    return p

  def setup (self, DB):
    pass

  def provision (self, driver, log):
    self.driver = driver

  def deprovision (self, driver, db):
    pass
