# Header info...
# Needs logging

import os
import datetime
import time
import subprocess

import hglib
import hglib.error

from ..types import UUID
from ..core.log import KeyAdapter

WORKERBIN = "/opt/foam/bin/hg-worker"
CORE_DIR = "/opt/foam/mounts/host-vols/_core-vols"
TIMEOUT = datetime.timedelta(seconds=60)

def now():
  return datetime.datetime.now()

class _ManagedHgVolume(object):
  def __init__(self, source, log):
    self.source = source
    self._log = log
    self.last_update = None # if this == None, never been provisioned
    self.volid = UUID()
    self.provisioned = False
    self.process = None
    self.branch = "default" # requires all repos to have default branch

  @staticmethod
  def build(source, log, volid):
    log.debug("Making managed volume for repo on disk: %s" % source)
    mgdvol = _ManagedHgVolume(source, log)
    mgdvol.volid = volid
    mgdvol.provisioned = True
    return mgdvol

  def getPath(self):
    return "%s/%s" % (CORE_DIR, self.volid)

  def getSource(self):
    return self.source

  def clone(self):
    if not self.provisioned: # we need to clone
      self.createVolume()
      self._log.debug("Cloning: %s" % self.source)
      arglist = [str(x) for x in [WORKERBIN, "clone", self.source, self.getPath(), self.branch]]
      #self._log.debug("Args are -- %s" % arglist)
      self.process = subprocess.Popen(arglist, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
      self.late_update = now()
    elif self.busy(): # Currently waiting for some process to finish so do nothing
      return
    elif self.stale(): # someone cloned before but too long ago
      self.pull()

  def pull(self):
    if self.busy():
      return
    if not self.provisioned: # could/should this happen? 
      self.clone() # okay to pull if not provisioned?
    else:
      if self.stale():
        self._log.debug("Pulling repo: %s" % self.source)
        arglist = [str(x) for x in [WORKERBIN, "pull", self.source, self.getPath(), self.branch]]
        self.process = subprocess.Popen(arglist, stdout = subprocess.PIPE, stderr = subprocess.PIPE) # okay to overwrite proc?
        self.late_update = now()

  def createVolume(self):
    # assuming we're root?
    self._log.debug("Creating directory on host")
    os.makedirs(self.getPath())
    self.provisioned = True

  def getHostSource(self):
    return self.getPath()

  def stale(self):
    try:
      if (now() - self.late_update) >= TIMEOUT:
        return True
      return False
    except AttributeError:
      return True

  def busy(self):
    if self.process is None:
      return False
    if self.process.poll() is None:
      return True
    else: 
      # Should we set process back to None once this is done?
      # What if we don't use this function?
      # This is to allow multiple containers to clone without calling clone more than once
      return False


class CentralVolumeManager(object):
  def __init__(self):
    self._hgvols = {}
    self._log = None

  def setLog (self, logger):
    self._log = KeyAdapter("CVM", logger)

  def getManagedSource(self, source):
    if source not in self._hgvols.keys():
      self._log.debug("Creating source: %s" % source)
      self._hgvols[source] = _ManagedHgVolume(source, self._log)
    return self._hgvols[source].getPath()
    
  def loadVolumes(self):
    self._log.debug("Loading volumes from disk")
    try:
      repos = os.listdir(CORE_DIR)
    except OSError:
      self._log.debug("Core directory not found")
      os.makedirs(CORE_DIR)
      return
    for repo in repos:
      try:
        if ".hg" in os.listdir("%s/%s" % (CORE_DIR, repo)):
          f = open("%s/%s/%s" % (CORE_DIR, repo, ".hg/hgrc"), "r").read()
          cropped = f[f.find("http"):]
          url = cropped[:cropped.find("\n")]
          self._log.debug("Repo found on disk: %s" % url)
          if url not in self._hgvols.keys():
            #need to create volume and set all the params and whatnot
            self._hgvols[url] = _ManagedHgVolume.build(url, self._log, repo)
      except OSError:
        pass

  def cloneVolume(self, source):
    try:
      self._hgvols[source].clone()
    except:
      self._log.debug("Requested repo not found: %s" % source)

  def checkStatusAndWait(self, source):
    # make sure we're waiting on what we think we are
    while (self._hgvols[source].busy()):
      time.sleep(0.5)

  def pull(self, source):
    self._hgvols[source].pull()

