#!/usr/bin/env python

# Copyright (c) 2018-2019 University of Houston

# Test file for testing central volume manager
# Should be run on the same machine as vts so it can restart foam
# Need three slices, names are hard coded

# Options:
# --test1 creates two slivers on different slices, deletes slivers; tests basic CVM functionality
# --test2 allocates two reservations on one slice, restarts foam, allocates second res on another slice,
#         restarts foam, provisions both slivers, writes ssh config file, does not delete slivers;
#         this tests the ability to persist through reboot managed volumes allocated but not provisioned
# --test3 creates sliver, waits for repo to timeout (60 sec.) then pulls using pull poa; tests pull poa
#         does not delete slivers
# --delete attempts to delete any slivers remaining in the three slices

# Example:
# "python cmvtester.py --test1 --test2 --test3"
# Use the config files test1.config & test2.config to login to hosts h0-h3 to verify repos
# "python cvmtester.py --delete" will delete all slivers

import argparse
import os
import subprocess as sp
import time
import logging

import uhgeni.ssh
import geni.util
import geni.aggregate.apis
import geni.aggregate.vts as VTSAM
import geni.rspec.vts as VTS

logging.basicConfig(level = logging.INFO)

SN1 = "test1"
SN2 = "test2"
SN3 = "test3"
AM = VTSAM.VTS("stubax-vts", "127.0.0.1")
context = geni.util.loadContext("/home/vagrant/.bssw/geni/context.json")

def parse_args ():
  parser = argparse.ArgumentParser()
  parser.add_argument("--test1", dest="test1", action="store_true", default=False)
  parser.add_argument("--test2", dest="test2", action="store_true", default=False)
  parser.add_argument("--test3", dest="test3", action="store_true", default=False)
  parser.add_argument("--delete", dest="delete", action="store_true", default=False)
  return parser.parse_args()

def restart_foam():
  logging.info("Restarting foam")
  p = sp.call('sudo service foam restart', shell=True)
  time.sleep(10)

def build_res():
  r = VTS.Request()
  d0 = r.Datapath(VTS.OVSL2Image(), "br0")
  r0 = r.Container(VTS.Image("uh.quagga"), "r0")
  c0 = r.Container(VTS.Image("uh.net-client"), "h0")
  c0.HgMount("doe-nf", "https://bitbucket.org/uh-netlab/doe-nf", "/nf")
  c1 = r.Container(VTS.Image("uh.net-client"), "h1")
  c1.HgMount("doe-nf", "https://bitbucket.org/uh-netlab/doe-nf", "/nf")
  VTS.connectInternalCircuit(d0, c0)
  VTS.connectInternalCircuit(d0, c1)
  VTS.connectInternalCircuit(d0, r0)
  return r

def build_res2():
  r2 = VTS.Request()
  d0= r2.Datapath(VTS.OVSL2Image(), "br1")
  r0= r2.Container(VTS.Image("uh.quagga"), "r1")
  c0 = r2.Container(VTS.Image("uh.net-client"), "h2")
  c0.HgMount("doe-nf", "https://bitbucket.org/uh-netlab/doe-nf", "/nf")
  c1 = r2.Container(VTS.Image("uh.net-client"), "h3")
  c1.HgMount("nfbase", "https://bitbucket.org/uh-netlab/pynf-base", "/pynf")
  VTS.connectInternalCircuit(d0, c0)
  VTS.connectInternalCircuit(d0, c1)
  VTS.connectInternalCircuit(d0, r0)
  return r2

def allocate(slicename, res):
  logging.info("Allocating sliver")
  manifest = AM.allocate(context, slicename, res)
  return manifest

def provision(slicename):
  logging.info("Provisioning sliver")
  manifest = AM.provision(context, slicename)
  return manifest

def create_sliver(slicename, res):
  try:
    logging.info("Trying to create sliver")
    manifest = AM.createsliver(context, slicename, res)
  except geni.aggregate.apis.CreateSliverError as e:
    logging.info(e)
    delete_slivers([slicename])
    manifest = AM.createsliver(context, slicename, res)
  logging.info("Sliver created")
  return manifest

def delete_slivers(slices):
  for s in slices:
    try:
      logging.info("Deleting sliver")
      AM.deletesliver(context, s)
    except Exception as e:
      logging.info(e)

if __name__ == '__main__':
  opts = parse_args()

  r1 = build_res()
  r2 = build_res2()

  if opts.test1:
    m1 = create_sliver(SN1, r1)
    m2 = create_sliver(SN2, r2)
    restart_foam()
    delete_slivers([SN1, SN2])

  if opts.test2:
    m_all = allocate(SN1, r1)
    m_all = allocate(SN1, r2)
    restart_foam()
    m_all2 = allocate(SN2, r2)
    restart_foam()
    m_pro = provision(SN1)
    m_pro2 = provision(SN2)
#    delete_slivers([SN1, SN2])
    uhgeni.ssh.writeSliceConfig (SN1, manifests = [m_pro], sliverstatus = None, pkey_path = None, path = None, user = None)
    uhgeni.ssh.writeSliceConfig (SN2, manifests = [m_pro2], sliverstatus = None, pkey_path = None, path = None, user = None)

# hgpull poa test
  if opts.test3:
    m3 = create_sliver(SN3, r2)
    time.sleep(65) # wait until CVM is stale
    for container in m3.containers:
      if container.name[0] == 'h':
        for mount in container.mounts:
          logging.info("Pulling: %s %s" % (container.name, mount.volid))
          AM.hgPull(context, SN3, [(container.name, mount.volid)])

  if opts.delete:
    delete_slivers([SN1, SN2, SN3])





# END
